Unit AtariMnu;

Interface

Uses AtariDbg,Profile,ProtMenu,TypeText;

Type
  TByteShift     = Array[0..2047] Of Byte;
  TMissileTable  = Array[0..128 * 32 - 1] Of Byte;
  TInitSound     = Procedure;
  TSoundBuf      = Array[0..1] Of Pointer;
  TColorScheme   = (csGameType,csAlphabetical,csEvery3,csEvery4,csEvery5,csEvery6);
  TDisplayScheme = (dsNameExt4,dsNameExt3,dsNameExt2,
                    dsNameOnly4,dsNameOnly3,dsNameOnly2,
                    dsRealNameLeft,dsRealNameCenter);
  ConfigRec      = Record
    MaxCols            : Word;   { HAND_TIA.ASM needs this to be first! }
    MaxRows            : Word;
    DoSound            : Boolean;
    FrameRate          : Word;
    SoundBufSize       : Word;
    PlaybackRate       : Word;
    AutoCenter         : Boolean;
    ScreenModeWidth    : Word;
    ScreenModeHeight   : Word;
    ScreenModeDepth    : Word;
    MenuColorScheme    : TColorScheme;
    MenuForeColor1     : Byte;
    MenuForeColor2     : Byte;
    MenuBackColor1     : Byte;
    MenuBackColor2     : Byte;
    MenuDisplayScheme  : TDisplayScheme;
    GameResolution     : TGameRes;
    Pitfall2DSP        : Boolean;
  End;

Procedure SaveConfiguration(Var Config: ConfigRec; Version: String);

Procedure ErrorFileNotFound(Const Path,St: String);

Function Menu(Var Config: ConfigRec; Version: String;
              Var Top,Sel: Integer;
              Var Playback_Freq: Word;
              Var Debugger,DoSound,UseMenu: Boolean;
              Var Pos200: Integer;
              Var Pos204: Integer;
              Var Pos215: Integer;
              Var Pos250: Integer;
              Var Pos305: Integer;
              Var ProfileRec: PFileType;
              Var Game: TSave): String;

Procedure InitConfig(Var Config: ConfigRec;
                     SoundBufSize,Playback_Freq: Word;
                     Version: String);

Procedure SetBankInfo(BankType: Char; Var Game: TSave);

Procedure InitMissileTable(Var MissileTable: TMissileTable);

Procedure InitByteShift(Var ByteShift: TByteShift);

Implementation

Uses Crt,Dos,HexWrite,Controls,Keyboard,Time,INI;

Const
  ConfigFile     = 'PCAE.INI';
  CRLF           = #13 + #10;
  DisplayColWidth: Array[TDisplayScheme] Of Byte = (16,15,14,12,11,10,40,40);

Var OldController: TController;

Procedure ErrorFileNotFound(Const Path,St: String);
Begin
  WriteLn('Sorry, but PCAE was unable to find file ',St,' in');
  WriteLn('directory ',FExpand(Path));
  WriteLn;
  WriteLn('You should make sure that PCAE was run from the');
  WriteLn('right directory and that ',St,' is present.');
End; { ErrorFileNotFound }


Procedure InitByteShift(Var ByteShift: TByteShift);
Var
  I,J   : Word;
  B1,B2 : Word;

Begin
  FillChar(ByteShift,SizeOf(ByteShift),#0);
  For I := 0 To 7 Do
  Begin
    For J := 0 To 71 Do
    Begin
      B2 := (I Shl 8) + J;
      B1 := B2 + 128;
      If (J <= 7) And (I In [0..4,6]) Then
      Begin
        ByteShift[(I Shl 8) + J + 128] := 1   Shl J;
        ByteShift[(I Shl 8) + J]       := $80 Shr J;
      End;
      Case I Of
        1: Case J Of
             16..23:
             Begin
               ByteShift[B1] := 1   Shl (J - 16);
               ByteShift[B2] := $80 Shr (J - 16);
             End;
           End;
        2: Case J Of
             32..39:
             Begin
               ByteShift[B1] := 1   Shl (J - 32);
               ByteShift[B2] := $80 Shr (J - 32);
             End;
           End; { Case }
        3: Case J Of
             16..23:
             Begin
               ByteShift[B1] := 1   Shl (J - 16);
               ByteShift[B2] := $80 Shr (J - 16);
             End;
             32..39:
             Begin
               ByteShift[B1] := 1   Shl (J - 32);
               ByteShift[B2] := $80 Shr (J - 32);
             End;
           End; { Case }
        4: Case J Of
             64..71:
             Begin
               ByteShift[B1] := 1   Shl (J - 64);
               ByteShift[B2] := $80 Shr (J - 64);
             End;
           End; { Case }
        5: If J <= 15 Then
           Begin
             ByteShift[B1] := 1   Shl (J Shr 1);
             ByteShift[B2] := $80 Shr (J Shr 1);
           End;
        6: Case J Of
             32..39:
             Begin
               ByteShift[B1] := 1   Shl (J - 32);
               ByteShift[B2] := $80 Shr (J - 32);
             End;
             64..71:
             Begin
               ByteShift[B1] := 1   Shl (J - 64);
               ByteShift[B2] := $80 Shr (J - 64);
             End;
           End; { Case }
        7: If J <= 31 Then
           Begin
             ByteShift[B1] := 1   Shl (J Shr 2);
             ByteShift[B2] := $80 Shr (J Shr 2);
           End;
      End; { Case }
    End; { For J }
  End; { For I }
End; { InitByteShift }


Procedure InitMissileTable(Var MissileTable: TMissileTable);
Var I,J: Word;
Begin
  FillChar(MissileTable,SizeOf(MissileTable),#0);
  For I := 0 To 31 Do
  Begin
    Case (I Shr 3) Of
      0: J := 1;
      1: J := 2;
      2: J := 4;
      3: J := 8;
    End; { Case }
    FillChar(MissileTable[(I Shl 7)],J,#255);
    Case (I And 7) Of
      1: FillChar(MissileTable[(I Shl 7) + 16],J,#255);
      2: FillChar(MissileTable[(I Shl 7) + 32],J,#255);
      3: Begin
           FillChar(MissileTable[(I Shl 7) + 16],J,#255);
           FillChar(MissileTable[(I Shl 7) + 32],J,#255);
         End;
      4: FillChar(MissileTable[(I Shl 7) + 64],J,#255);
      6: Begin
           FillChar(MissileTable[(I Shl 7) + 32],J,#255);
           FillChar(MissileTable[(I Shl 7) + 64],J,#255);
         End;
    End; { Case }
  End; { For I }
End; { InitMissileTable }


Procedure SetBankInfo(BankType: Char; Var Game: TSave);
 Var J: Integer;
Begin
  Case BankType Of
'T': Begin                                { TigerVision }
      Game.BankSwitch[$003F] := $60;
      Game.Cart              := ctTigerVision;
      Game.BankSize          := 512;
    End;
'A': Begin                                { Robot Tank/Decathlon }
      Game.BankSwitch[$01FE] := 30;
      Game.Cart              := ctActivision;
    End;
'2': Begin                                { Pitfall II }
      Game.BankSwitch[$1FF8] := $CB;
      Game.BankSwitch[$1FF9] := $CC;
      Game.Cart               := ctPitfall2;
      For J := $1040 To $107F Do Game.BankSwitch[J] := $CF;
      For J := $1840 To $187F Do Game.BankSwitch[J] := $CF;
      For J := $1000 To $103F Do Game.BankSwitch[J] := $CE;
      For J := $1800 To $183F Do Game.BankSwitch[J] := $CE;
    End;
'R': Begin                                { Superchip }
      Game.Cart := ctSuperChip;
{        BankSwitch^[$1FF6] := 1  + $40;
      BankSwitch^[$1FF7] := 5  + $40;
      BankSwitch^[$1FF8] := 9  + $40;
      BankSwitch^[$1FF9] := 13 + $40;}
      For J := $1000 To $107F Do Game.BankSwitch[J] := $5F;
      Game.BankSize := 1024 - 64{32};    { Size in dwords }
    End;
'P': Begin                                { Parker Bros. }
      For J := $1FE0 To $1FF7 Do Game.BankSwitch[J] := J And $FF;
      Game.Cart := ctParkerBros;
    End;
'M': Begin                                { M-Network }
      For J := $1FE0 To $1FEB Do Game.BankSwitch[J] := (J And $FF) - $10;
      For J := $1000 To $13FF Do Game.BankSwitch[J] := $DC;
      For J := $1800 To $18FF Do Game.BankSwitch[J] := $DD;
      Game.Cart := ctMNetwork;
      Game.MNetworkRAM := 8;
    End;
'G': Begin                               { 16-in-1 MegaBoy }
      Game.BankSwitch[$1FF0] := $80;
      Game.Cart := ctMegaBoy;
    End;
'8': Begin                               { Standard 8k }
       Game.BankSwitch[$1FF8] := 1;
       Game.BankSwitch[$1FF9] := 5;
       Game.Cart := ct8k;
     End;
'S': Begin                               { Starpath }
       For J := $1000 To $1FFF Do Game.BankSwitch[J] := $70;
       Game.BankSwitch[$1FF8] := $70;
       Game.BankSwitch[$0E00] := $70;
       Game.Cart := ctStarpath;
     End;
'C': Begin                               { CBS RAM-Plus }
       Game.BankSwitch[$1FF8] := $21;
       Game.BankSwitch[$1FF9] := $25;
       Game.BankSwitch[$1FFA] := $29;
       For J := $1000 To $10FF Do Game.BankSwitch[J] := $30;
       Game.BankSize := 1024 - 128{64};    { Size in dwords }
       Game.Cart := ctCBS;
     End;
'6': Begin                               { Standard 16k }
       Game.BankSwitch[$1FF6] := 1;
       Game.BankSwitch[$1FF7] := 5;
       Game.BankSwitch[$1FF8] := 9;
       Game.BankSwitch[$1FF9] := 13;
       Game.Cart := ct16k;
     End;
  End; { Case }
End; { SetBankInfo }


Procedure InitConfig(Var Config: ConfigRec;
                     SoundBufSize,Playback_Freq: Word;
                     Version: String);
Var
  F  : PINIFile;
  St : String;
  B  : Byte;

Begin
  Config.DoSound            := True;
  Config.FrameRate          := 60;
  Config.SoundBufSize       := SoundBufSize;
  Config.PlaybackRate       := Playback_Freq;
  Config.AutoCenter         := False;
  Config.MaxCols            := 320;
  Config.MaxRows            := 200;
  Config.ScreenModeWidth    := 80;
  Config.ScreenModeHeight   := 25;
  Config.ScreenModeDepth    := 1;
  Config.MenuColorScheme    := csGameType;
  Config.MenuForeColor1     := LightGray;
  Config.MenuForeColor2     := White;
  Config.MenuBackColor1     := Blue;
  Config.MenuBackColor2     := Blue;
  Config.MenuDisplayScheme  := dsNameExt4;
  Config.GameResolution     := gr320x200;
  Config.Pitfall2DSP        := False;

  { Initalize controller interfaces }

  InterfaceHandler^.AddInterface(New(PMouseInterface,Init));
  InterfaceHandler^.AddInterface(New(PKeyboardInterface,Init));
  InterfaceHandler^.AddInterface(New(PJoystickInterface,Init));
  InterfaceHandler^.AddInterface(New(PGrIPInterface,Init));
  InterfaceHandler^.AddInterface(New(PSideWinderInterface,Init));
{  InterfaceHandler^.AddInterface(New(PLogitechInterface,Init));}
  InterfaceHandler^.AddInterface(New(PExternalInterface,Init));

  If Exist(ConfigFile) Then
  Begin
    F := New(PINIFile,Init(ConfigFile));
    Config.MaxCols           := F^.ReadInteger('GENERAL','MAXCOLS',320);
    Config.MaxRows           := F^.ReadInteger('GENERAL','MAXROWS',200);
    Config.DoSound           := F^.ReadBool('GENERAL','DOSOUND',False);
    Config.FrameRate         := F^.ReadInteger('GENERAL','FRAMERATE',60);
    Config.SoundBufSize      := F^.ReadInteger('GENERAL','SOUNDBUFSIZE',SoundBufSize);
    Config.PlaybackRate      := F^.ReadInteger('GENERAL','PLAYBACKRATE',Playback_Freq);
    Config.AutoCenter        := F^.ReadBool('GENERAL','AUTOCENTER',False);
    Config.ScreenModeWidth   := F^.ReadInteger('GENERAL','SCREENMODEWIDTH',80);
    Config.ScreenModeHeight  := F^.ReadInteger('GENERAL','SCREENMODEHEIGHT',25);
    Config.ScreenModeDepth   := F^.ReadInteger('GENERAL','SCREENMODEDEPTH',8);
    Config.MenuColorScheme   := TColorScheme(F^.ReadInteger('GENERAL','MENUCOLORSCHEME',LongInt(csGameType)));
    Config.MenuForeColor1    := F^.ReadInteger('GENERAL','MENUFORECOLOR1',LightGray);
    Config.MenuForeColor2    := F^.ReadInteger('GENERAL','MENUFORECOLOR2',White);
    Config.MenuBackColor1    := F^.ReadInteger('GENERAL','MENUBACKCOLOR1',Blue);
    Config.MenuBackColor2    := F^.ReadInteger('GENERAL','MENUBACKCOLOR2',Blue);
    Config.MenuDisplayScheme := TDisplayScheme(F^.ReadInteger('GENERAL','MENUDISPLAYSCHEME',LongInt(dsNameExt4)));
    Config.GameResolution    := TGameRes(F^.ReadInteger('GENERAL','GAMERESOLUTION',LongInt(gr320x200)));
    Config.Pitfall2DSP       := F^.ReadBool('GENERAL','PITFALL2DSP',False);

    { Load controller interface information }

    InterfaceHandler^.LoadConfiguration(F);

    Dispose(F,Done);
  End
  Else SaveConfiguration(Config,Version);
  If Not ((Config.MaxRows - 200) In [0,4,15,50,105]) Then Config.MaxRows := 200;
End; { InitConfig }


Procedure MainMenuHelp;
Begin
{  TextMode(C80);}
  Screen.TextColor(LightGray);
  Screen.TextBackGround(Blue);
  Screen.ClrScr;
  Screen.GotoXY(3,1);  Screen.Write('Run options');

  Screen.GotoXY(5,3);  Screen.Write('<Esc> ............................ Quit');
  Screen.GotoXY(5,4);  Screen.Write('Arrow keys or Joystick ........... Move selection');
  Screen.GotoXY(5,5);  Screen.Write('<Enter> or Joystick button 1 ..... Run 2K/4K/Atari 8K/Atari 16K cart');
  Screen.GotoXY(5,6);  Screen.Write('A ................................ Run Activision 16K cart');
  Screen.GotoXY(5,7);  Screen.Write('M ................................ Run M-Network 8K RAM cart');
  Screen.GotoXY(5,8);  Screen.Write('P ................................ Run Parker Bros. 8K cart');
  Screen.GotoXY(5,9);  Screen.Write('R ................................ Run Atari super-chip cart');
  Screen.GotoXY(5,10); Screen.Write('T ................................ Run TigerVision cart');
  Screen.GotoXY(5,11); Screen.Write('2 ................................ Run Pitfall II cart');
  Screen.GotoXY(3,13); Screen.Write('Game options');
  Screen.GotoXY(5,14); Screen.Write('F1 ............................... Help/status screen');
  Screen.GotoXY(5,15); Screen.Write('F2 ............................... Reset game');
  Screen.GotoXY(5,16); Screen.Write('F3 ............................... Toggle NTSC/PAL palette');
  Screen.GotoXY(5,17); Screen.Write('F4 ............................... Select game');
  Screen.GotoXY(5,18); Screen.Write('F5 ............................... Color/BW switch');
  Screen.GotoXY(5,19); Screen.Write('F6/F7 ............................ Player 1/2 difficulty');
  Screen.GotoXY(5,20); Screen.Write('F8 ............................... Save game info to profile file');
  Screen.GotoXY(5,21); Screen.Write('F9 ............................... Capture screen to .BMP file');
  Screen.GotoXY(5,22); Screen.Write('F10 .............................. Toggle vertical autocentering');
  Screen.GotoXY(5,23); Screen.Write('F11 .............................. Save game to disk');
  Screen.GotoXY(5,24); Screen.Write('F12 .............................. Load saved game from disk');
  Screen.GotoXY(5,25); Screen.Write('=,- .............................. Shift screen up or down');
  ReadKey;
  Screen.ClrScr;
  Screen.GotoXY(3,1);  Screen.Write('Command-line options');
  Screen.GotoXY(2,2);  Screen.Write('ATARI filename [BANK|BANK16|BANKA|BANKM|BANKP|BANKC|BANKSC|BANKSP|BANKT|BANK2|');
  Screen.GotoXY(2,3);  Screen.Write('   BANKG|BANK3|PROFILE][DEBUG][DISASM][FRAMExx][KB|INDY500|VTP|BOOSTER|PADDLE]');
  Screen.GotoXY(5,4);  Screen.Write('BANK ............................. Run Atari 8K cart');
  Screen.GotoXY(5,5);  Screen.Write('BANK16 ........................... Run Atari 16K cart');
  Screen.GotoXY(5,6);  Screen.Write('BANKA ............................ Run Activision 16K cart');
  Screen.GotoXY(5,7);  Screen.Write('BANKM ............................ Run M-Network 8K RAM cart');
  Screen.GotoXY(5,8);  Screen.Write('BANKP ............................ Run Parker Bros. 8K cart');
  Screen.GotoXY(5,9);  Screen.Write('BANKC ............................ Run CBS RAM-Plus 12K cart');
  Screen.GotoXY(5,10); Screen.Write('BANKSC ........................... Run Atari super-chip cart');
  Screen.GotoXY(5,11); Screen.Write('BANKSP ........................... Run Starpath Supercharger cart');
  Screen.GotoXY(5,12); Screen.Write('BANKT ............................ Run TigerVision cart');
  Screen.GotoXY(5,13); Screen.Write('BANK2 ............................ Run Pitfall II cart');
  Screen.GotoXY(5,14); Screen.Write('BANKG ............................ Run 16-in-1 MegaBoy cart');
  Screen.GotoXY(5,15); Screen.Write('BANK3 ............................ Run Atari 32-in-1 cart');
  Screen.GotoXY(5,16); Screen.Write('32IN1NUMxx ....................... Run Atari 32-in-1 game xx (0-31)');
  Screen.GotoXY(5,17); Screen.Write('PROFILE .......................... Use game profile file');
  Screen.GotoXY(5,18); Screen.Write('DEBUG ............................ Use intregrated debugger');
  Screen.GotoXY(5,19); Screen.Write('DISASM ........................... Disassemble 4k cart to LISTING.ASM');
  Screen.GotoXY(5,20); Screen.Write('FRAMExx .......................... Limit emulation to xx frames/second');
  Screen.GotoXY(5,21); Screen.Write('KB ............................... Run Atari keyboard controller game');
  Screen.GotoXY(5,22); Screen.Write('BOOSTER .......................... Run CBS Booster-Grip controller game');
  Screen.GotoXY(5,23); Screen.Write('INDY500 .......................... Emulate Indy 500 controllers');
  Screen.GotoXY(5,24); Screen.Write('VTP .............................. Emulate Video Touch Pad');
  Screen.GotoXY(5,25); Screen.Write('PADDLE ........................... Emulate Paddle controllers');
  ReadKey;
  Screen.TextColor(LightGray);
  Screen.TextBackGround(Blue);
  Screen.ClrScr;
End; { MainMenuHelp }


Procedure SaveConfiguration(Var Config: ConfigRec; Version: String);
Var
  F  : PINIFile;
  St : String;
  F1 : File;

Begin
  If Exist(ConfigFile) Then
  Begin
    Assign(F1,ConfigFile);
    Erase(F1);
  End;
  F := New(PINIFile,Init(ConfigFile));
  F^.WriteInteger('GENERAL','MAXCOLS',Config.MaxCols);
  F^.WriteInteger('GENERAL','MAXROWS',Config.MaxRows);
  F^.WriteBool('GENERAL','DOSOUND',Config.DoSound);
  F^.WriteInteger('GENERAL','FRAMERATE',Config.FrameRate);
  F^.WriteInteger('GENERAL','SOUNDBUFSIZE',Config.SoundBufSize);
  F^.WriteInteger('GENERAL','PLAYBACKRATE',Config.PlaybackRate);
  F^.WriteBool('GENERAL','AUTOCENTER',Config.AutoCenter);
  F^.WriteInteger('GENERAL','SCREENMODEWIDTH',Config.ScreenModeWidth);
  F^.WriteInteger('GENERAL','SCREENMODEHEIGHT',Config.ScreenModeHeight);
  F^.WriteInteger('GENERAL','SCREENMODEDEPTH',Config.ScreenModeDepth);
  F^.WriteInteger('GENERAL','MENUCOLORSCHEME',LongInt(Config.MenuColorScheme));
  F^.WriteInteger('GENERAL','MENUFORECOLOR1',Config.MenuForeColor1);
  F^.WriteInteger('GENERAL','MENUFORECOLOR2',Config.MenuForeColor2);
  F^.WriteInteger('GENERAL','MENUBACKCOLOR1',Config.MenuBackColor1);
  F^.WriteInteger('GENERAL','MENUBACKCOLOR2',Config.MenuBackColor2);
  F^.WriteInteger('GENERAL','MENUDISPLAYSCHEME',LongInt(Config.MenuDisplayScheme));
  F^.WriteInteger('GENERAL','GAMERESOLUTION',LongInt(Config.GameResolution));
  F^.WriteBool('GENERAL','PITFALL2DSP',Config.Pitfall2DSP);
  InterfaceHandler^.SaveConfiguration(F);
  Dispose(F,Done);
End; { SaveConfiguration }


Function Menu(Var Config: ConfigRec; Version: String;
              Var Top,Sel: Integer;
              Var Playback_Freq: Word;
              Var Debugger,DoSound,UseMenu: Boolean;
              Var Pos200: Integer;
              Var Pos204: Integer;
              Var Pos215: Integer;
              Var Pos250: Integer;
              Var Pos305: Integer;
              Var ProfileRec: PFileType;
              Var Game: TSave): String;

Type
  PFileListType = ^TFileListType;
  TFileListType = Record
    Name : String[12];
    Size : LongInt;
    Next : PFileListType;
    Prev : PFileListType;
    Prof : PFileType;
  End;

Var
  I       : Word;
  Head    : PFileListType;
  Tail    : PFileListType;
  Work    : PFileListType;
  Work1   : PFileListType;
  Temp    : TFileListType;
  S       : SearchRec;
  Ch,C    : Char;
  Cols    : Integer;
  Rows    : Integer;
  Num     : Integer;
  Show    : Boolean;
  J       : Integer;
  JoyX1   : LongInt;
  JoyY1   : LongInt;
  JoyX2   : LongInt;
  JoyY2   : LongInt;
  CJoyX   : LongInt;
  CJoyY   : LongInt;
  JoyX1_2 : LongInt;
  JoyY1_2 : LongInt;
  JoyX2_2 : LongInt;
  JoyY2_2 : LongInt;
  JoyB    : Byte;
  F       : File;
  Path    : String[79];
  _Menu   : PMenu;
  PreMenu : PMenu;
  ScnMenu : PMenu;
  ColMenu : PMenu;
  DisMenu : PMenu;
  ResMenu : PMenu;
  ScnMode : PScreenMode;
  EnaMenu : PMenu;
  _Win    : PWindow;
  MWid    : Integer;
  St      : String;
  Code    : Integer;
  DJX,DJY : LongInt;
  Cont    : TController;
  Intf    : PInterface;
  Found   : Boolean;
  OldTop  : Integer;
  OldWork : PFileListType;
  Cap     : TCapability;
  Info    : TAtariControls;

  Procedure Display;
  Var
    I,J  : Integer;
    W    : PFileListType;
    St   : String;
    Row  : Integer;
    Col  : Integer;
    Show : Boolean;

    Procedure ShowPartial(Const St: String; Const X,Y: Integer);
    Var
      I   : Integer;
      St1 : String[1];

    Begin
      For I := 1 To Length(St) Do
      Begin
        St1 := St[I];
        If St1 <> ' ' Then Screen.FastWrite(X + I - 1,Y,Screen.TextAttr,St1);
      End; { For I }
    End; { ShowPartial }

  Begin
    I   := Rows * Cols;
    W   := Work1;
    Row := 1;
    Col := 1;


    If Cols * DisplayColWidth[Config.MenuDisplayScheme] < MaxX Then Inc(I,Rows);


    While (W <> Nil) And (I <> 0) Do
    Begin
      Show := (W = Work) Or (W = OldWork) Or (Top <> OldTop);
      If W = Work Then
      Begin
        Screen.TextColor(Black);
        Screen.TextBackGround(LightGray);
      End
      Else
      Begin
        Case Config.MenuColorScheme Of
        csGameType:
        Begin
          Case W^.Size Of
2048,4096: Screen.TextColor(LightGray);
    8192: Screen.TextColor(White);
    8448: Screen.TextColor(LightRed);
   10495: Screen.TextColor(LightGreen);
   16384: Screen.TextColor(Yellow);
          Else
            Screen.TextColor(Black);
          End; { Case }
          If (W^.Size Mod 8448) = 0 Then Screen.TextColor(LightRed);
          Screen.TextBackGround(Blue);
        End;
        csAlphabetical:
        Begin
          If (Ord(W^.Name[1]) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery3:
        Begin
          If (((Row - 1) Div 3) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery4:
        Begin
          If (((Row - 1) Div 4) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery5:
        Begin
          If (((Row - 1) Div 5) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery6:
        Begin
          If (((Row - 1) Div 6) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        Else
          Screen.TextColor(LightGray);
          Screen.TextBackground(Blue);
        End; { Case }
      End;
      If Show Then
      Begin
        Case Config.MenuDisplayScheme Of
        dsNameExt4,dsNameExt3,dsNameExt2:
        Begin
          St := W^.Name;
          J  := Pos('.',St);
          If J <> 0 Then
           While J < 9 Do
           Begin
             Insert(' ',St,J);
             Inc(J);
           End; { While }
        End;
        dsNameOnly4,dsNameOnly3,dsNameOnly2:
        Begin
          St := W^.Name;
          J  := Pos('.',St);
          If J <> 0 Then St := Copy(St,1,J - 1);
        End;
        dsRealNameLeft:
        Begin
          If (W^.Prof <> Nil) And (W^.Prof^.RealName <> '')
           Then St := W^.Prof^.RealName
           Else St := W^.Name;
        End;
        dsRealNameCenter:
        Begin
          If (W^.Prof <> Nil) And (W^.Prof^.RealName <> '')
           Then St := W^.Prof^.RealName
           Else St := W^.Name;
          While Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Do
          Begin
            St := St + ' ';
            If Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Then
             St := ' ' + St;
          End; { While }
        End;
        Else
          St := '';
        End; { Case }
        While Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Do St := St + ' ';
        If Col >= Cols Then
        Begin
          If Col * DisplayColWidth[Config.MenuDisplayScheme] > MaxX Then
           St := Copy(St,1,MaxX - (Col - 1) * DisplayColWidth[Config.MenuDisplayScheme]);
        End;
        Screen.FastWrite((Col - 1) * DisplayColWidth[Config.MenuDisplayScheme] + 1,Row + 1,Screen.TextAttr,St);
      End;
      Inc(Row);
      If Row >= Rows + 1 Then
      Begin
        Row := 1;
        Inc(Col);
      End;
      Dec(I);
      W := W^.Next;
    End; { While }

    { Erase any remaining entries }

    If Top <> OldTop Then
    Begin
      St := '';
      While Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Do St := St + ' ';
      Screen.TextColor(LightGray);
      Screen.TextBackground(Blue);
      While I > 0 Do
      Begin
        Case Config.MenuColorScheme Of
        csGameType:
        Begin
          Screen.TextColor(LightGray);
          Screen.TextBackground(Blue);
        End;
        csAlphabetical:
        Begin
          Screen.TextColor(Config.MenuForeColor1);
          Screen.TextBackground(Config.MenuBackColor1);
        End;
        csEvery3:
        Begin
          If (((Row - 1) Div 3) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery4:
        Begin
          If (((Row - 1) Div 4) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery5:
        Begin
          If (((Row - 1) Div 5) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery6:
        Begin
          If (((Row - 1) Div 6) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        Else
          Screen.TextColor(LightGray);
          Screen.TextBackground(Blue);
        End; { Case }
        If Col >= Cols Then
        Begin
          If Col * DisplayColWidth[Config.MenuDisplayScheme] > MaxX Then
           St := Copy(St,1,MaxX - (Col - 1) * DisplayColWidth[Config.MenuDisplayScheme]);
        End;
        Screen.FastWrite((Col - 1) * DisplayColWidth[Config.MenuDisplayScheme] + 1,Row + 1,Screen.TextAttr,St);
        Inc(Row);
        If Row >= Rows + 1 Then
        Begin
          Row := 1;
          Inc(Col);
        End;
        Dec(I);
      End; { While }
    End;

    { Display the top status line }

    Screen.TextColor(Magenta);
    Screen.TextBackGround(LightGray);
    Screen.GotoXY(1,1);
    Screen.ClrEol;

    St := 'PCAE ' + Version{ + '  ' + IntToStr(MemAvail)};
    Screen.FastWrite(2,1,Screen.TextAttr,St);

    Screen.TextColor(Black);

    St := Work^.Name;
    Screen.FastWrite(22,1,Screen.TextAttr,St);
    Str(Work^.Size,St);
    Screen.FastWrite(40,1,Screen.TextAttr,St);

    Screen.TextColor(Red);

    St := 'F1       F10       Esc     ';
    ShowPartial(St,52,1);
    Screen.TextColor(Black);
    St := '   Help      Menu      Quit';
    ShowPartial(St,52,1);

    { Display the bottom status line }

    If Top <> OldTop Then
    Begin
      Screen.GotoXY(1,MaxY);
      Screen.ClrEol;

      Str(Num,St);
      St := St + ' Files';
      Screen.FastWrite(2,MaxY,Screen.TextAttr,St);
      St := '32In1 MeGaBoy Pitfall2 ParkBrs ActVsn MNetwork SupRchp TigrVsn';
      Screen.FastWrite(17,MaxY,Screen.TextAttr,St);

      Screen.TextColor(Red);

      St := '3       G            2 P       A      M           R    T      ';
      ShowPartial(St,17,MaxY);
    End;
    OldTop  := Top;
    OldWork := Work;
  End; { Display }

  Procedure FindTop;
  Var I: Integer;
  Begin
    I     := Top;
    Work1 := Head;
    While (Work1 <> Nil) And (I > 0) Do
    Begin
      Work1 := Work1^.Next;
      Dec(I);
    End; { While }
  End;

  Procedure FindSel;
  Var I: Integer;
  Begin
    I     := Sel;
    Work  := Head;
    While (Work <> Nil) And (I > 0) Do
    Begin
      Work := Work^.Next;
      Dec(I);
    End; { While }
  End; { FindSel }

  Procedure _Up;
  Begin
    If Sel > 0 Then
    Begin
      If Sel = Top Then
      Begin
        Dec(Top);
        Work1   := Work^.Prev;
      End;
      Dec(Sel);
      Work := Work^.Prev;
      Show := True;
    End;
  End; { _Up }

  Procedure _Down;
  Begin
    If Sel < Num - 1 Then
    Begin
      If Sel = Top + Cols * Rows - 1 Then
      Begin
        Inc(Top);
        Work1   := Work1^.Next;
      End;
      Inc(Sel);
      Work := Work^.Next;
      Show := True;
    End;
  End; { _Down }

  Procedure _Left;
  Begin
    If Sel >= Rows Then
    Begin
      Show := True;
      Dec(Sel,Rows);
      If Sel < Top Then Dec(Top,Rows);
      If Top < 0 Then Top := 0;
      FindTop;
      FindSel;
    End;
  End; { _Left }

  Procedure _Right;
  Begin
    If Sel < Num - Rows Then
    Begin
      Show := True;
      Inc(Sel,Rows);
      If Sel >= Top + Cols * Rows Then Inc(Top,Rows);
      FindTop;
      FindSel;
    End;
  End; { _Right }

  Procedure Popup(St: String);
  Var
    W       : PWindow;
    I,J,X,Y : Integer;

  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St);
    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    W := New(PWindow,Init(X,Y - 2,X + I + 3,Y + 2,
           Black,LightGray,Black,LightGray,Black,LightGray,
           DoubleBorder,'','',True));
    W^.Display;
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 2,Screen.TextAttr,St);
    Sound(880);
    Delay(10);
    NoSound;
    Delay(1000);
    W^.Remove;
    Dispose(W,Done);
  End; { Popup }

  Procedure Popup4(St1,St2,St3,St4: String);
  Var
    W       : PWindow;
    I,J,X,Y : Integer;

  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St1);
    If I < Length(St2) Then I := Length(St2);
    If I < Length(St3) Then I := Length(St3);
    If I < Length(St4) Then I := Length(St4);
    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    W := New(PWindow,Init(X,Y - 2,X + I + 3,Y + 3,
           Black,LightGray,Black,LightGray,Black,LightGray,
           DoubleBorder,'','',True));
    W^.Display;
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 1,Screen.TextAttr,St1);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 2,Screen.TextAttr,St2);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 3,Screen.TextAttr,St3);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 4,Screen.TextAttr,St4);
    Sound(880);
    Delay(10);
    NoSound;
    Delay(1000);
    W^.Remove;
    Dispose(W,Done);
  End; { Popup4 }

  Procedure StayPopup(St: String);
  Var I,J,X,Y : Integer;
  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St);
    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    _Win := New(PWindow,Init(X,Y - 2,X + I + 3,Y + 2,
              Black,LightGray,Black,LightGray,Black,LightGray,
              DoubleBorder,'','',True));
    _Win^.Display;
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 2,Screen.TextAttr,St);
  End; { StayPopup }

  Procedure StayPopup4(St1,St2,St3,St4: String);
  Var I,J,X,Y : Integer;
  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St1);
    If Length(St2) > I Then I := Length(St2);
    If Length(St3) > I Then I := Length(St3);
    If Length(St4) > I Then I := Length(St4);

    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    _Win := New(PWindow,Init(X,Y - 2,X + I + 3,Y + 3,
              Black,LightGray,Black,LightGray,Black,LightGray,
              DoubleBorder,'','',True));
    _Win^.Display;
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 1,Screen.TextAttr,St1);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 2,Screen.TextAttr,St2);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 3,Screen.TextAttr,St3);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 4,Screen.TextAttr,St4);
  End; { StayPopup4 }

  Procedure SetScreenMode;
  Var
    I       : Integer;
    Sel     : PSelection;
    MaxX0   : Integer;
    MaxY0   : Integer;
    DX,DY   : Integer;
    OldMode : Word;
    T1      : LongInt;
    T2      : LongInt;
    Ch      : Char;

    Procedure GotoMode(Mode: Word);
    Begin
      MaxX0 := MaxX;
      MaxY0 := MaxY;
      Screen.SetMode(Mode);
      Screen.ShowCursor(False);
      Screen.TextColor(LightGray);
      Screen.TextBackground(Blue);
      Screen.ClrScr;
      Cols   := MaxX Div DisplayColWidth[Config.MenuDisplayScheme];
      Rows   := MaxY - 2;
      OldTop := -1;      { Invalidate to force full redisplay }
      Display;
      Config.ScreenModeWidth  := Screen.Width;
      Config.ScreenModeHeight := Screen.Height;
      Config.ScreenModeDepth  := Screen.BPP;
      SaveConfiguration(Config,Version);
      DX := (MaxX - MaxX0) Div 2;
      DY := (MaxY - MaxY0) Div 2;
      _Menu^.MoveWindow(DX,DY);
      DisMenu^.MoveWindow(DX,DY);
      ColMenu^.MoveWindow(DX,DY);
      PreMenu^.MoveWindow(DX,DY);
      ScnMenu^.MoveWindow(DX,DY);
      EnaMenu^.MoveWindow(DX,DY);
    End; { GotoMode }

  Begin
    Repeat
      OldMode := Screen.CurrentMode;
      I := 0;
      While I < ScnMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(ScnMenu^.SelList^.At(I));
        Sel^.Checked := (I = Screen.CurrentMode);
        Inc(I);
      End; { While }
      ScnMenu^.Execute;
      If ScnMenu^.CurSel <> 0 Then
      Begin
        StayPopup4('PCAE will now resize your desktop.  If the screen does not',
                   'display correctly, wait ten seconds, and the screen will',
                   'revert to this screen mode.',
                   '                                         Hit a key to test');
        ReadKey;
        _Win^.Remove;
        Dispose(_Win,Done);
        GotoMode(ScnMenu^.CurSel - 1);
        T1 := Clock;
        StayPopup4('Did the screen display correctly?','',
                   '         (Y)es  or (N)o','');
        Ch := #0;
        Repeat
          If KeyPressed Then Ch := UpCase(ReadKey);
          T2 := Clock;
        Until (T2 < T1 - 1000) Or (T2 >= T1 + 10000) Or (Ch In ['N','Y']);
        _Win^.Remove;
        Dispose(_Win,Done);
        If (Ch = 'N') Or (T2 >= T1 + 10000) Then GotoMode(OldMode);
      End;
    Until ScnMenu^.CurSel = 0;
  End; { SetScreenMode }

  Procedure SetDisplayScheme;
  Var
    Sel : PSelection;
    I   : Integer;

  Begin
    Repeat
      I := 0;
      While I < DisMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(DisMenu^.SelList^.At(I));
        Sel^.Checked := (I = Byte(Config.MenuDisplayScheme));
        Inc(I);
      End; { While }
      DisMenu^.Execute;
      If DisMenu^.CurSel > 0 Then
      Begin
        Case DisMenu^.CurSel Of
          1: Config.MenuDisplayScheme := dsNameExt4;
          2: Config.MenuDisplayScheme := dsNameExt3;
          3: Config.MenuDisplayScheme := dsNameExt2;
          4: Config.MenuDisplayScheme := dsNameOnly4;
          5: Config.MenuDisplayScheme := dsNameOnly3;
          6: Config.MenuDisplayScheme := dsNameOnly2;
          7: Config.MenuDisplayScheme := dsRealNameLeft;
          8: Config.MenuDisplayScheme := dsRealNameCenter;
        End; { Case }
        SaveConfiguration(Config,Version);
        Cols   := MaxX Div DisplayColWidth[Config.MenuDisplayScheme];
        OldTop := -1;      { Invalidate to force full redisplay }
        Display;
      End;
    Until DisMenu^.CurSel = 0;
  End; { SetDisplayScheme }

  Procedure SetColorScheme;
  Var
    Sel : PSelection;
    I   : Integer;

    Procedure ChooseColors;
    Var
      Win : PWindow;
      C   : Array[1..4] Of Byte;
      Ch  : Char;
      Col : Byte;
      I   : Byte;

    Begin
      C[1] := Config.MenuForeColor1;
      C[2] := Config.MenuBackColor1;
      C[3] := Config.MenuForeColor2;
      C[4] := Config.MenuBackColor2;
      Col  := 1;
      Win  := New(PWindow,InitWH(50,20,
                 Black,LightGray,Black,LightGray,Black,LightGray,
                 DoubleBorder,' Choose colors ',' Arrow keys to move, <Esc> exits',True));
      Win^.Display;
      Screen.TextColor(Black);
      Screen.TextBackground(LightGray);
      Screen.ClrScr;
      Repeat
        Screen.TextBackground(LightGray);
        For I := 0 To 15 Do
        Begin
          Screen.TextColor(I);
          Screen.GotoXY(10,I + 3); Screen.Write('  ');
          Screen.GotoXY(20,I + 3); Screen.Write('  ');
          Screen.GotoXY(30,I + 3); Screen.Write('  ');
          Screen.GotoXY(40,I + 3); Screen.Write('  ');
        End; { For I }
        For I := 1 To 4 Do
        Begin
          Screen.TextBackground(LightGray);
          If Col = I Then Screen.TextColor(LightRed) Else Screen.TextColor(Black);
          Screen.GotoXY(I * 10 + 2,C[I] + 3); Screen.Write(#17);
        End; { For I }
        Screen.TextColor(Black);
        Screen.GotoXY(1,1);
        Screen.Write('        Fore      Back      Fore      Back');

        Ch := ReadKey;
        If Ch = #0 Then
        Begin
          Ch := ReadKey;
          Case Ch Of
            #71: Begin   { Home }
                   If C[Col] <> 0 Then C[Col] := 0;
                 End;
            #72: Begin   { Up }
                   If C[Col] > 0 Then Dec(C[Col]);
                 End;
            #75: Begin   { Left }
                   If Col > 1 Then Dec(Col);
                 End;
            #77: Begin   { Right }
                   If Col < 4 Then Inc(Col);
                 End;
            #79: Begin   { End }
                   If C[Col] <> 15 Then C[Col] := 15;
                 End;
            #80: Begin   { Down }
                   If C[Col] < 15 Then Inc(C[Col]);
                 End;
          End; { Case }
          Ch := #0;
        End;
      Until Ch = #27;
      Win^.Remove;
      Dispose(Win,Done);
      Config.MenuForeColor1 := C[1];
      Config.MenuBackColor1 := C[2];
      Config.MenuForeColor2 := C[3];
      Config.MenuBackColor2 := C[4];
    End; { ChooseColors }

  Begin
    Repeat
      I := 0;
      While I < ColMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(ColMenu^.SelList^.At(I));
        Sel^.Checked := (I = Byte(Config.MenuColorScheme));
        Inc(I);
      End; { While }
      ColMenu^.Execute;
      If ColMenu^.CurSel > 0 Then
      Begin
        Case ColMenu^.CurSel Of
          1: Config.MenuColorScheme := csGameType;
          2: Begin
               Config.MenuColorScheme := csAlphabetical;
               ChooseColors;
             End;
          3: Begin
               Config.MenuColorScheme := csEvery3;
               ChooseColors;
             End;
          4: Begin
               Config.MenuColorScheme := csEvery4;
               ChooseColors;
             End;
          5: Begin
               Config.MenuColorScheme := csEvery5;
               ChooseColors;
             End;
          6: Begin
               Config.MenuColorScheme := csEvery6;
               ChooseColors;
             End;
        End; { Case }
        SaveConfiguration(Config,Version);
        OldTop := -1;      { Invalidate to force full redisplay }
        Display;
      End;
    Until ColMenu^.CurSel = 0;
  End; { SetColorScheme }

  Procedure SetGameResolution;
  Var
    Sel : PSelection;
    I   : Integer;

  Begin
    Repeat
      I := 0;
      While I < ResMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(ResMenu^.SelList^.At(I));
        Sel^.Checked := (Integer(Config.GameResolution) = I);
        Inc(I);
      End; { While }
      ResMenu^.Execute;
      Case ResMenu^.CurSel Of
        1: Begin
             Config.MaxCols        := 320;
             Config.MaxRows        := 200;
             Config.GameResolution := gr320x200;
           End;
        2: Begin
             Config.MaxCols        := 320;
             Config.MaxRows        := 204;
             Config.GameResolution := gr320x204;
           End;
        3: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 200;
             Config.GameResolution := gr160x200A;
           End;
        4: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 204;
             Config.GameResolution := gr160x204A;
           End;
        5: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 215;
             Config.GameResolution := gr160x215A;
           End;
        6: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 250;
             Config.GameResolution := gr160x250A;
           End;
        7: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 305;
             Config.GameResolution := gr160x305A;
           End;
        8: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 200;
             Config.GameResolution := gr160x200B;
           End;
        9: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 204;
             Config.GameResolution := gr160x204B;
           End;
       10: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 215;
             Config.GameResolution := gr160x215B;
           End;
       11: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 250;
             Config.GameResolution := gr160x250B;
           End;
       12: Begin
             Config.MaxCols        := 160;
             Config.MaxRows        := 305;
             Config.GameResolution := gr160x305B;
           End;
      End; { Case }
      SaveConfiguration(Config,Version);
    Until ResMenu^.CurSel = 0;
  End; { SetGameResolution }

  Procedure SetFramesPerSecond;
  Var
    _Win : PWindow;
    St   : String;

  Begin
    _Win := New(PWindow,InitWH(60,5,
     Black,LightGray,Black,LightGray,Black,LightGray,
     DoubleBorder,'','',True));
    _Win^.Display;
    Screen.GotoXY(2,2);
    Screen.Write('Enter frames-per-second limit (or 0 for no limit): ');
    Str(Config.FrameRate,St);
    St := _BufLen(53,2,St,Yellow,Blue,3,4);
    Screen.ShowCursor(False);
    Dispose(_Win,Done);
    Val(St,Config.FrameRate,Code);
    If Config.FrameRate < 0 Then Config.FrameRate := 0;
    SaveConfiguration(Config,Version);
    Screen.TextColor(LightGray);
    Screen.TextBackGround(Blue);
  End; { SetFramesPerSecond }

  Procedure SetSoundOptions;
  Var
    _Win : PWindow;
    C    : Char;

  Begin
    _Win := New(PWindow,Init(6,(MaxY Div 2) - 6,MaxX - 5,MaxY Div 2 + 6,
     Black,LightGray,Black,LightGray,Black,LightGray,
     DoubleBorder,'','',True));
    _Win^.Display;
    Screen.GotoXY(2,2);
    Screen.Write('Sound playback rate adjustment');
    Screen.GotoXY(2,3);
    Screen.Write('(The largest value gives the best sound, but might not work on all sound cards)');
    Screen.GotoXY(2,5);
    Screen.Write('Playback rate: [' + IntToStr(Playback_Freq) + ']: ');
    Screen.GotoXY(2,6);
    Screen.Write('  1.............11025 Hz');
    Screen.GotoXY(2,7);
    Screen.Write('  2.............15700 Hz');
    Screen.GotoXY(2,8);
    Screen.Write('  3.............22050 Hz');
    Screen.GotoXY(2,9);
    Screen.Write('  4.............31400 Hz');
    Screen.GotoXY(2,11);
    Screen.Write('Choice (1-4): ');
    Repeat
      C := ReadKey;
    Until C In ['1'..'4'];
    Dispose(_Win,Done);
    Case C Of
      '1': Playback_Freq := 11025;
      '2': Playback_Freq := 15700;
      '3': Playback_Freq := 22050;
      '4': Playback_Freq := 31400;
    End; { Case }
    Config.PlaybackRate := Playback_Freq;
    SaveConfiguration(Config,Version);
    Screen.TextColor(LightGray);
    Screen.TextBackGround(Blue);
  End; { SetSoundOptions }

  Procedure SetSoundBufferSize;
  Var I: Integer;
  Begin
    _Win := New(PWindow,InitWH(50,5,
     Black,LightGray,Black,LightGray,Black,LightGray,
     DoubleBorder,'','',True));
    _Win^.Display;
    Screen.GotoXY(2,2);
    Screen.Write('Enter sound buffer size (64-8192): ');
    Str(Config.SoundBufSize,St);
    St := _BufLen(37,2,St,Yellow,Blue,4,5);
    Screen.ShowCursor(False);
    Dispose(_Win,Done);
    Val(St,I,Code);
    If (Code = 0) And (I >= 64) And (I <= 8192) Then { Force multiple of 4 }
     Config.SoundBufSize := I And $FFFC;
    SaveConfiguration(Config,Version);
    Screen.TextColor(LightGray);
    Screen.TextBackGround(Blue);
  End; { SetSoundBufferSize }

  Procedure SetPreferences;
  Var Sel: PSelection;
  Begin
    Repeat
      Sel := PreMenu^.GetSelection(1);
      Sel^.Checked := Config.AutoCenter;
      Sel := PreMenu^.GetSelection(9);
      Sel^.Checked := Config.Pitfall2DSP;
      PreMenu^.Execute;
      Case PreMenu^.CurSel Of
        1: Begin
             Sel := PreMenu^.GetSelection(1);
             Sel^.Checked := Not Sel^.Checked;
             Config.AutoCenter := Sel^.Checked;
             SaveConfiguration(Config,Version);
           End;
        2: SetScreenMode;
        3: SetColorScheme;
        4: SetDisplayScheme;
        5: SetGameResolution;
        6: SetFramesPerSecond;
        7: SetSoundOptions;
{        8: SetSoundBufferSize;}
        8: Begin
             Sel := PreMenu^.GetSelection(9);
             Sel^.Checked := Not Sel^.Checked;
             Config.Pitfall2DSP := Sel^.Checked;
             SaveConfiguration(Config,Version);
           End;
      End; { Case }
    Until PreMenu^.CurSel = 0;
  End; { SetPreferences }

  Procedure InitControllerInfo;
  Begin
    Info.PortA   := $FF;
    Info.Console := 0;
    Info.Pad[1]  := 0;
    Info.Pad[2]  := 0;
    Info.Pad[3]  := 0;
    Info.Pad[4]  := 0;
    Info.Input4  := False;
    Info.Input5  := False;
    Info.Driv[1] := 0;
    Info.Driv[2] := 0;
    Info.CBS1    := False;
    Info.CBS2    := False;
    Info.KB      := 0;
  End; { InitControllerInfo }

Begin
  OldTop  := -1;
  OldWork := Nil;

  { Try to set to the mode saved in the configuration, otherwise }
  { use mode 0                                                   }

  I     := 0;
  Found := False;
  While (I < Screen.ScreenModes^.Count) And Not Found Do
  Begin
    ScnMode := PScreenMode(Screen.ScreenModes^.At(I));
    If (ScnMode^.Width  = Config.ScreenModeWidth) And
       (ScnMode^.Height = Config.ScreenModeHeight) And
       (ScnMode^.BPP    = Config.ScreenModeDepth)
     Then Found := True Else Inc(I);
  End; { While }
  If Not Found Then I := 0;
  Screen.SetMode(I);
  Screen.ShowCursor(False);
  Game.Controller := OldController;
  Screen.TextColor(LightGray);
  Screen.TextBackGround(Blue);
  Screen.ClrScr;

  { Locate all the .BIN and .PAL files }

  Head := Nil;
  Tail := Nil;
  Num  := 0;
  If ParamCount > 0 Then
  Begin
    Path := ParamStr(1);
    If (Path <> '') And Not (Path[Length(Path)] In [':','\']) Then Path := Path + '\';
  End
  Else Path := '';

  { Locate all .BIN and .PAL files }

  FindFirst(Path + '*.BIN',Archive,S);
  While DosError = 0 Do
  Begin
    If Head = Nil Then
    Begin
      New(Head);
      Tail := Head;
      Tail^.Prev := Nil;
    End
    Else
    Begin
      New(Tail^.Next);
      Tail^.Next^.Prev := Tail;
      Tail := Tail^.Next;
    End;
    Inc(Num);
    Tail^.Name := S.Name;
    Tail^.Size := S.Size;
    Tail^.Next := Nil;
    FindProfile(S.Name,Cont,Pos200,Pos204,Pos215,Pos250,Pos305,Tail^.Prof);
    FindNext(S);
  End; { While }

  FindFirst(Path + '*.PAL',Archive,S);
  While DosError = 0 Do
  Begin
    If Head = Nil Then
    Begin
      New(Head);
      Tail := Head;
      Tail^.Prev := Nil;
    End
    Else
    Begin
      New(Tail^.Next);
      Tail^.Next^.Prev := Tail;
      Tail := Tail^.Next;
    End;
    Inc(Num);
    Tail^.Name := S.Name;
    Tail^.Size := S.Size;
    Tail^.Next := Nil;
    FindProfile(S.Name,Cont,Pos200,Pos204,Pos215,Pos250,Pos305,Tail^.Prof);
    FindNext(S);
  End; { While }

  { Sort the file list in ascending alphabetical order }

  Work := Head;
  While Work <> Nil Do
  Begin
    Work1 := Work^.Next;
    While Work1 <> Nil Do
    Begin
      If Work1^.Name < Work^.Name Then
      Begin
        Temp.Name   := Work1^.Name;
        Temp.Size   := Work1^.Size;
        Temp.Prof   := Work1^.Prof;
        Work1^.Name := Work^.Name;
        Work1^.Size := Work^.Size;
        Work1^.Prof := Work^.Prof;
        Work^.Name  := Temp.Name;
        Work^.Size  := Temp.Size;
        Work^.Prof  := Temp.Prof;
      End;
      Work1 := Work1^.Next;
    End; { Work1 }
    Work := Work^.Next;
  End; { While }

  { Main loop }

  If Head = Nil Then
  Begin
    WriteLn('Sorry, but PCAE was unable to find any .BIN or .PAL files in');
    WriteLn('directory ',FExpand(Path));
    WriteLn;
    WriteLn('You should make sure that this is the right');
    WriteLn('path to specify and that there are .BIN or .PAL files there.');

    Menu := '';
    Exit;
  End;
  FindTop;
  FindSel;

  Cols  := MaxX Div DisplayColWidth[Config.MenuDisplayScheme];
  Rows  := MaxY - 2;
  Show  := True;

  { Main menu }

  _Menu := New(PMenu,InitWH(31,17,
             Black,LightGray,Black,LightGray,Blue,LightGray,
             Red,LightGray,Black,Green,Red,Green,
             DarkGray,LightGray,DarkGray,Green,
             SingleBorder,' PCAE ' + Version + ' Main Menu ',' (C) 1999 John Dullea ',True));
  _Menu^.AddSelection(1, 1,29,7,False,False,'Build game profile');

  _Menu^.AddSelection(1, 3,29,5,False,False,'Use joystick controllers');
  _Menu^.AddSelection(1, 4,29,6,False,False,'Use paddle controllers');
  _Menu^.AddSelection(1, 5,29,8,False,False,'Use keyboard controllers');
  _Menu^.AddSelection(1, 6,29,5,False,False,'Use Indy 500 controllers');
  _Menu^.AddSelection(1, 7,29,5,False,False,'Use Video Touch Pad');

  _Menu^.AddSelection(1, 9,29,1,False,False,'Configure controllers');
  _Menu^.AddSelection(1,10,29,1,False,False,'Enable interfaces');

  _Menu^.AddSelection(1,12,29,1,False,False,'Preferences');

  _Menu^.AddSelection(1,14,29,8,False,False,'Toggle debugger');
  _Menu^.AddSelection(1,15,29,8,False,False,'Toggle sound');

  EnaMenu := New(PMenu,InitWH(36,InterfaceHandler^.Stack^.Count + 2,
             Black,LightGray,Black,LightGray,Blue,LightGray,
             Red,LightGray,Black,Green,Red,Green,
             DarkGray,LightGray,DarkGray,Green,
             SingleBorder,' Enable interfaces ',' <Enter> toggles, <Esc> exits ',True));
  I := 0;
  While I < InterfaceHandler^.Stack^.Count Do
  Begin
    EnaMenu^.AddSelection(1,I + 1,34,1,False,False,'  ' +
                         PInterface(InterfaceHandler^.Stack^.At(I))^.Name);
    Inc(I);
  End; { While }

  { Preferences menu }

  PreMenu := New(PMenu,InitWH(36,10,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Preferences ',' <Enter> selects, <Esc> exits ',True));
  PreMenu^.AddSelection(1,1,34, 3,False,False,'  Auto-centering');
  PreMenu^.AddSelection(1,2,34, 3,False,False,'  Screen mode');
  PreMenu^.AddSelection(1,3,34,13,False,False,'  File list color scheme');
  PreMenu^.AddSelection(1,4,34,13,False,False,'  File list display scheme');
  PreMenu^.AddSelection(1,5,34, 8,False,False,'  Game Resolution');
  PreMenu^.AddSelection(1,6,34, 7,False,False,'  Set frames per second');
  PreMenu^.AddSelection(1,7,34,13,False,False,'  Set sound options');
{  PreMenu^.AddSelection(1,8,34,13,False,False,'  Set sound buffer size');}
  PreMenu^.AddSelection(1,8,34,13,False,False,'  Realistic Pitfall II music');

  { Screen mode menu }

  ScnMenu := New(PMenu,InitWH(36,Screen.ScreenModes^.Count + 2,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Choose screen mode ',' <Enter> selects, <Esc> exits ',True));
  I := 0;
  While I < Screen.ScreenModes^.Count Do
  Begin
    ScnMode := PScreenMode(Screen.ScreenModes^.At(I));
    St      := '  ' + IntToStr(ScnMode^.Width) + 'x' + IntToStr(ScnMode^.Height);
    Case ScnMode^.DisplayType Of
      dtText: St := St + ' text';
  dtVESAText: St := St + ' VESA text';
   dtVESA256: St := St + 'x256 VESA';
  dtVESATrue: St := St + 'x' + IntToStr(ScnMode^.BPP) + '-bit VESA';
    End; { Case }
    ScnMenu^.AddSelection(1,I + 1,34,1,False,I = Screen.CurrentMode,St);
    Inc(I);
  End; { While }

  { Color scheme menu }

  ColMenu := New(PMenu,InitWH(36,8,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Color schemes ',' <Enter> selects, <Esc> exits ',True));
  ColMenu^.AddSelection(1,1,34, 3,False,False,'  Game type');
  ColMenu^.AddSelection(1,2,34,13,False,False,'  Alternate alphabetically');
  ColMenu^.AddSelection(1,3,34,19,False,False,'  Alternate every 3 rows');
  ColMenu^.AddSelection(1,4,34,19,False,False,'  Alternate every 4 rows');
  ColMenu^.AddSelection(1,5,34,19,False,False,'  Alternate every 5 rows');
  ColMenu^.AddSelection(1,6,34,19,False,False,'  Alternate every 6 rows');

  { Display scheme menu }

  DisMenu := New(PMenu,InitWH(36,10,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Display schemes ',' <Enter> selects, <Esc> exits ',True));
  DisMenu^.AddSelection(1,1,34,12,False,False,'  Name and extension w/4 spaces');
  DisMenu^.AddSelection(1,2,34,12,False,False,'  Name and extension w/3 spaces');
  DisMenu^.AddSelection(1,3,34,12,False,False,'  Name and extension w/2 spaces');
  DisMenu^.AddSelection(1,4,34, 3,False,False,'  Name only w/4 spaces');
  DisMenu^.AddSelection(1,5,34, 3,False,False,'  Name only w/3 spaces');
  DisMenu^.AddSelection(1,6,34, 3,False,False,'  Name only w/2 spaces');
  DisMenu^.AddSelection(1,7,34, 3,False,False,'  Real name (left justified)');
  DisMenu^.AddSelection(1,8,34, 3,False,False,'  Real name (centered)');

  { Game resolution menu }

  ResMenu := New(PMenu,InitWH(36,14,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Game resolutions ',' <Enter> selects, <Esc> exits ',True));
  ResMenu^.AddSelection(1, 1,34,1,False,False,'  320x200 (standard VGA)');
  ResMenu^.AddSelection(1, 2,34,1,False,False,'  320x204');
  ResMenu^.AddSelection(1, 3,34,1,False,False,'  160x200 (faster)');
  ResMenu^.AddSelection(1, 4,34,1,False,False,'  160x204');
  ResMenu^.AddSelection(1, 5,34,1,False,False,'  160x215');
  ResMenu^.AddSelection(1, 6,34,1,False,False,'  160x250');
  ResMenu^.AddSelection(1, 7,34,1,False,False,'  160x305');
  ResMenu^.AddSelection(1, 8,34,1,False,False,'  160x200 (alternate)');
  ResMenu^.AddSelection(1, 9,34,1,False,False,'  160x204');
  ResMenu^.AddSelection(1,10,34,1,False,False,'  160x215');
  ResMenu^.AddSelection(1,11,34,1,False,False,'  160x250');
  ResMenu^.AddSelection(1,12,34,1,False,False,'  160x305');
  InterfaceHandler^.ActivateInterfaces;
  Repeat
    If Show Then
    Begin
      Display;
      Show := False;
    End;

    Ch := #0;

    { See if the controllers can move the cursor }

    Cap.A := imJoysticks;
    Cap.B := 0;

    { Initialize the info structure }

    InitControllerInfo;

    { We're purposely NOT executing InitKeyboard because we don't want }
    { the keyboard interface responding                                }

    FillChar(Key,SizeOf(Key),#0); { Clear the key array just to be sure }

    InterfaceHandler^.GetControllerInfo(Cap,Info);

    If (Info.PortA And $10) = 0 Then
    Begin
      Delay(1);
      InitControllerInfo;
      InterfaceHandler^.GetControllerInfo(Cap,Info);
      If (Info.PortA And $10) = 0 Then
      Begin
        _Up;
        Delay(30);
      End;
    End;
    If (Info.PortA And $20) = 0 Then
    Begin
      Delay(1);
      InitControllerInfo;
      InterfaceHandler^.GetControllerInfo(Cap,Info);
      If (Info.PortA And $20) = 0 Then
      Begin
        _Down;
        Delay(30);
      End;
    End;
    If (Info.PortA And $40) = 0 Then
    Begin
      Delay(1);
      InitControllerInfo;
      InterfaceHandler^.GetControllerInfo(Cap,Info);
      If (Info.PortA And $40) = 0 Then
      Begin
        _Left;
        Delay(30);
      End;
    End;
    If (Info.PortA And $80) = 0 Then
    Begin
      Delay(1);
      InitControllerInfo;
      InterfaceHandler^.GetControllerInfo(Cap,Info);
      If (Info.PortA And $80) = 0 Then
      Begin
        _Right;
        Delay(30);
      End;
    End;
    If Info.Input4 Then
    Begin
      Delay(1);
      InitControllerInfo;
      InterfaceHandler^.GetControllerInfo(Cap,Info);
      If Info.Input4 Then
      Begin
        Ch := #13;

        { Wait until button is released }

        Repeat
          Cap.A := imJoysticks;
          Cap.B := 0;

          { Initialize the info structure }

          InitControllerInfo;

          InterfaceHandler^.GetControllerInfo(Cap,Info);
        Until Not Info.Input4;
      End;
    End;
    If KeyPressed Then
    Begin
      Ch := UpCase(ReadKey);
      If Ch = #0 Then
      Begin
        C := ReadKey;
        Case C Of
          #59: Begin                     { F1 }
                 MainMenuHelp;
                 OldTop := -1;  { Invalidate to force full redisplay }
                 Show := True;
               End;
          #68: Begin                     { F10 }
                 Repeat
                   _Menu^.Execute;
                   Case  _Menu^.CurSel Of
                     1: Begin
                          _Win := New(PWindow,Init(6,(MaxY Div 2) - 5,MaxX - 5,MaxY Div 2 + 5,
                           Black,LightGray,Black,LightGray,Black,LightGray,
                           DoubleBorder,'','',True));
                          _Win^.Display;
                          Screen.GotoXY(7,2);
                          Screen.Write('Rebuilding PCAE.PRO cartridge profile file');
                          Screen.GotoXY(2,4);
                          MakeProfile(Path);
                          Dispose(_Win,Done);
                        End;
                     2: Begin
                          Game.Controller := crJoystick;
                          Popup('Using standard controllers');
                        End;
                     3: Begin
                          Game.Controller := crPaddle;
                          Popup('Using paddle controllers');
                        End;
                     4: Begin
                          Game.Controller := crKeyboard;
                          Popup('Keyboard controllers ACTIVE');
                        End;
                     5: Begin
                          Game.Controller := crDriving;
                          Popup('Indy 500 controllers ACTIVE');
                        End;
                     6: Begin
                          Game.Controller := crVTP;
                          Popup('Video Touch Pad ACTIVE');
                        End;
                     7: Begin
                          InterfaceHandler^.DeactivateInterfaces;
                          InterfaceHandler^.Configure;
                          SaveConfiguration(Config,Version);
                          InterfaceHandler^.ActivateInterfaces;
                        End;
                     8: Begin
                          If InterfaceHandler^.Stack^.Count > 0 Then
                          Begin

                            { Check/uncheck various interfaces as appropriate }

                            I := 0;
                            While I < InterfaceHandler^.Stack^.Count Do
                            Begin
                              EnaMenu^.GetSelection(I + 1)^.Checked :=
                               PInterface(InterfaceHandler^.Stack^.At(I))^.Enabled;
                              Inc(I);
                            End; { While }
                            Repeat
                              EnaMenu^.Execute;
                              If EnaMenu^.CurSel > 0 Then
                              Begin
                                InterfaceHandler^.DeactivateInterfaces;
                                Intf := PInterface(InterfaceHandler^.Stack^.At(EnaMenu^.CurSel - 1));
                                Intf^.Enabled := Not Intf^.Enabled;
                                EnaMenu^.GetSelection(EnaMenu^.CurSel)^.Checked := Intf^.Enabled;
                                InterfaceHandler^.ActivateInterfaces;
                              End;
                            Until EnaMenu^.CurSel = 0;
                            SaveConfiguration(Config,Version);
                          End;
                        End;
                     9: SetPreferences;
                    10: Begin
                          Debugger := Not Debugger;
                          If Debugger
                           Then Popup('Debugger is ON')
                           Else Popup('Debugger is OFF');
                        End;
                    11: Begin
                          Config.DoSound := Not Config.DoSound;
                          DoSound        := Config.DoSound;
                          SaveConfiguration(Config,Version);
                          If DoSound
                           Then Popup('Sound is ON')
                           Else Popup('Sound is OFF');
                        End;
                   End; { Case }
                 Until _Menu^.CurSel = 0;
               End;
          #71: Begin                     { Home }
                 If Sel <> 0 Then Show := True;
                 Top   := 0;
                 Sel   := 0;
                 Work1 := Head;
                 Work  := Head;
               End;
          #72: _Up;                      { Up }
          #73: Begin                     { PgUp }
                 If Sel >= Cols * Rows Then
                 Begin
                   Show := True;
                   Dec(Sel,Cols * Rows);
                   Dec(Top,Cols * Rows);
                   If Top < 0 Then Top := 0;
                   FindTop;
                   FindSel;
                 End;
               End;
          #75: _Left;                    { Left }
          #77: _Right;                   { Right }
          #79: Begin                     { End }
                 If Sel <> Num - 1 Then Show := True;
                 Sel := Num - 1;
                 I   := Top;
                 Top := Sel - Cols * Rows + 1;
                 If Top < 0 Then Top := 0;
                 If I <> Top Then Show := True;
                 Work  := Tail;
                 FindTop;
               End;
          #80: _Down;                    { Down }
          #81: Begin                     { PgDn }
                 If Sel < Num - Cols * Rows Then
                 Begin
                   Show := True;
                   Inc(Sel,Cols * Rows);
                   Inc(Top,Cols * Rows);
                   FindTop;
                   FindSel;
                 End;
               End;
        End; { Case }
      End;
    End;
  Until Ch In [#13,#27,'A','G','R','P','M','T','2','3'];

  InterfaceHandler^.DeactivateInterfaces;

  If Ch In [#13,'A','G','R','P','M','T','2','3'] Then
  Begin
    UseMenu := True;
    Menu    := Path + Work^.Name;
    If Ch = #13 Then
    Begin
      OldController := Game.Controller;
      Ch := MenuKey[FindProfile(Work^.Name,Cont,Pos200,Pos204,Pos215,Pos250,Pos305,ProfileRec)];
      If Cont <> crUnknown Then Game.Controller := Cont;
    End;
    If Ch In ['T','A','2','R','P','M','G'] Then SetBankInfo(Ch,Game)
    Else
    Begin
      Case Ch Of
  '3': Begin                               { 32-in-1 }
  {      BankSwitch^[$1FF0] := $80;}
        Game.Cart := ct32In1;
        _Win := New(PWindow,Init(6,(MaxY Div 2) - 2,MaxX - 5,MaxY Div 2 + 2,
         Black,LightGray,Black,LightGray,Black,LightGray,
         DoubleBorder,'','',True));
        _Win^.Display;
        Screen.GotoXY(2,2);
        Screen.Write('Enter game number to play (0-31):');
        Str(Game.MegaBoyNum,St);
        St := _BufLen(34,2,St,Yellow,Blue,3,3);
        Screen.ShowCursor(False);
        Val(St,J,Code);
        If (J >= 0) And (J <= 31) Then Game.MegaBoyNum := J;
        Dispose(_Win,Done);
      End;
      Else
        Game.Cart := ct4k;
        If Work^.Size = 8192  Then SetBankInfo('8',Game); { Standard 8k }
        If Work^.Size = 10495 Then SetBankInfo('2',Game); { Pitfall II }
        If (Work^.Size Mod 8448) = 0 Then SetBankInfo('S',Game); { Starpath }
        If Work^.Size = 12288 Then SetBankInfo('C',Game); { CBS RAM-Plus }
        If Work^.Size = 16384 Then SetBankInfo('6',Game); { Standard 16k }
      End; { Case }
    End;
  End
  Else
  Begin
    Menu    := '';
    UseMenu := False;
  End;

  { Kill the file list }

  Work := Head;
  While Work <> Nil Do
  Begin
    Work1 := Work^.Next;
    Dispose(Work);
    Work := Work1;
  End; { While }
  Screen.TextColor(LightGray);
  Screen.TextBackGround(Black);
  Screen.ClrScr;
  Dispose(ResMenu,Done);
  Dispose(DisMenu,Done);
  Dispose(ColMenu,Done);
  Dispose(ScnMenu,Done);
  Dispose(PreMenu,Done);
  Dispose(EnaMenu,Done);
  Dispose(_Menu,Done);
End; { Menu }

Begin
  OldController := crJoystick;
End.