Unit GameStat;

Interface

Uses Keyboard,G320,AtariDbg;

Type
  RGB = Record
    R,G,B: Byte;
  End;
  PPalette   = ^TPalette;
  TPalette   = Array[0..127] Of RGB;

Var
  OldKey    : Array[0..$FF] Of Byte;

Procedure GameStatus(Width,Height: Word; Var Difficulty1,Difficulty2: LongInt; Var AutoCenter: Boolean);
Procedure ScreenCapture(Width,Height,Palette: Word; Pal0,Pal1: PPalette);
Function  EnterStringPopup(Width,Height: Word; Title,OldStr: String): String;
Function  CreateFile(Width,Height: Word; Var F: File; St: String): Boolean;
Function  CanOpenFile(Width,Height: Word; St: String): Boolean;
Procedure SetPalette(Var Pal: TPalette);

Implementation

Uses Time;

Const
  KeyBufferSize  = 256;
  BI_RGB         = 0;

Type
  BITMAPFILEHEADER = Record
    bfType      : Array[0..1] Of Char;
    bfSize      : LongInt;
    bfReserved1 : Word;
    bfReserved2 : Word;
    bfOffBits   : LongInt;
  End;
  BITMAPINFOHEADER = Record
    biSize          : LongInt;
    biWidth         : LongInt;
    biHeight        : LongInt;
    biPlanes        : Word;
    biBitCount      : Word;
    biCompression   : LongInt;
    bisizeImage     : LongInt;
    biXPelsPerMeter : LongInt;
    biYPelsPerMeter : LongInt;
    biClrUsed       : LongInt;
    biClrImportant  : LongInt;
  End;
  RGBQuad = Record
    B,G,R,Reserved : Byte;
  End;

Var
  KeyBuffer : Array[0..KeyBufferSize - 1] Of Byte;
  BufPos    : Integer;
  BufStart  : Integer;
  TM0,TM1   : LongInt;
  TM2,TM3   : LongInt;
  LastKey   : Integer;
  Wait      : Integer;

Procedure SetPalette(Var Pal: TPalette);
Var I: Integer;
Begin
  Port[$3C8] := 0;
  For I := 0 To 127 Do
  Begin
    Port[$3C9] := Pal[I].B Shr 2;
    Port[$3C9] := Pal[I].G Shr 2;
    Port[$3C9] := Pal[I].R Shr 2;

    { Duplicate the colors here to avoid having to worry about bit 0 }

    Port[$3C9] := Pal[I].B Shr 2;
    Port[$3C9] := Pal[I].G Shr 2;
    Port[$3C9] := Pal[I].R Shr 2;
  End; { For I }
End; { SetPalette }

Procedure CheckNewKeyPressed;
Var
  I : Integer;
  B : Boolean;

  Procedure AddToBuffer(Ch: Byte);
  Begin
    If Not ((BufPos = BufStart - 1) Or
            ((BufPos = KeyBufferSize - 1) And (BufStart = 0))) Then
    Begin
      LastKey := Ch;
      KeyBuffer[BufPos] := Ch;
      Inc(BufPos);
      If BufPos >= KeyBufferSize Then BufPos := 0;
    End;
  End; { AddToBuffer }

Begin

  { We only want to check the keyboard every 20 ms }

  TM3 := Clock;
  If (TM3 < (TM2 - 1000)) Or (TM3 >= TM2 + 20) Then
  Begin

    { Compare the Key buffer to the OldKey buffer.  If a key has }
    { been pressed that wasn't pressed before, set B to true and }
    { save the key number in I.                                  }

    Asm
      MOV   AX,DS
      MOV   ES,AX
      MOV   AX,0FFh
      MOV   SI,OFFSET Key
      MOV   DI,OFFSET OldKey
      MOV   CX,256
  @L1:
      TEST  AL,[SI]
      JZ    @L2
      MOV   BL,[SI]
      CMP   BL,[DI]
      JE    @L2
      INC   AH
      JMP   @L3
  @L2:
      INC   SI
      INC   DI
      DEC   CX
      JNZ   @L1
  @L3:
      MOV   BYTE PTR B,AH
      MOV   DX,256
      SUB   DX,CX
      MOV   WORD PTR I,DX
    End; { Asm }
    If B Then
    Begin
      AddToBuffer(I);
      TM0 := Clock;
    End
    Else
    Begin
      If LastKey >= 0 Then
      Begin
        TM1 := Clock;
        If (Key[LastKey] <> 0) And
           ((TM1 < TM0) Or
            (TM1 >= TM0 + Wait)) Then
        Begin
          AddToBuffer(LastKey);
          Wait := 33;
          TM0  := TM1;
        End;
        If Key[LastKey] = 0 Then
        Begin
          Wait    := 250;
          LastKey := -1;
        End;
      End;
    End;

    { Copy the Key buffer to the OldKey buffer }

    Asm
      MOV   AX,DS
      MOV   ES,AX
      MOV   SI,OFFSET Key
      MOV   DI,OFFSET OldKey
      CLD
      MOV   CX,64
      DB    _32BIT_OPERAND
      REP   MOVSW
    End; { Asm }

    { Remember the last time the keyboard was checked }

    TM2 := TM3;
  End;
End; { CheckNewKeyPressed }

Function KeyPressed: Boolean;
Begin
  KeyPressed := (BufStart <> BufPos);
End; { KeyPressed }

Function ReadKey: Char;
Begin
  If BufPos = BufStart Then
  Begin
    Repeat
      CheckNewKeyPressed;
    Until KeyPressed;
  End;
  ReadKey := Char(KeyBuffer[BufStart]);
  Inc(BufStart);
  If BufStart >= KeyBufferSize Then BufStart := 0;
End; { ReadKey }

Procedure SaveScreen(Width,BufSize: Word; Buf: Pointer);
Begin
  If Width = 160 Then Move(Mem[$A000:0],Buf^,BufSize)
  Else
  Begin
    Asm
      CLD
      PUSH  DS
      MOV   CX,WORD PTR BufSize
      LES   DI,DWORD PTR Buf
      MOV   AX,0A000h
      MOV   DS,AX
      SUB   SI,SI
@L1:
      MOV   AL,BYTE PTR [SI]
      MOV   BYTE PTR ES:[DI],AL
      ADD   SI,2
      INC   DI
      DEC   CX
      JNZ   @L1
      POP   DS
    End; { Asm }
  End;
End; { SaveScreen }


Procedure RestoreScreen(Width,BufSize: Word; Buf: Pointer);
Begin
  If Width = 160 Then Move(Buf^,Mem[$A000:0],BufSize)
  Else
  Begin
    Asm
      CLD
      PUSH  DS
      MOV   CX,WORD PTR BufSize
      MOV   AX,0A000h
      MOV   ES,AX
      SUB   DI,DI
      LDS   SI,DWORD PTR Buf
@L1:
      MOV   AL,BYTE PTR [SI]
      MOV   AH,AL
      MOV   WORD PTR ES:[DI],AX
      INC   SI
      ADD   DI,2
      DEC   CX
      JNZ   @L1
      POP   DS
    End; { Asm }
  End;
End; { RestoreScreen }

Procedure Dim(BufSize: Word); Assembler;
Asm
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   CX,WORD PTR BufSize
  SHR   CX,1
  SUB   DI,DI
@L1:
  MOV   DX,WORD PTR [ES:DI]
  MOV   BX,DX
  AND   BX,0E0Eh
  AND   DX,0F0F0h

  OR    BL,BL
  JZ    @L2
  SHR   BL,2
  CMP   BL,2
  JGE   @L2
  MOV   BL,2
@L2:
  OR    BH,BH
  JZ    @L3
  SHR   BH,2
  CMP   BH,2
  JGE   @L3
  MOV   BH,2
@L3:
  OR    DX,BX
  MOV   WORD PTR [ES:DI],DX
  ADD   DI,2
  LOOP  @L1
End; { Dim }

Procedure DrawWindow(X1,Y1,X2,Y2: Word; BK,C1,C2: Byte; Width: Word);
Const BorderWidth = 5;
Var
  X,Y   : Word;
  Addr1 : Word;
  Addr2 : Word;
  CInc  : Integer;

Begin
  BK := BK Shl 1;
  C1 := C1 Shl 1;
  C2 := C2 Shl 1;

  { Draw top and bottom borders }

  For Y := 0 To BorderWidth - 1 Do
  Begin
    Addr1 := Width * (Y1 + Y);
    Addr2 := Width * (Y2 + Y - BorderWidth + 1);
    CInc  := 2 * ((BorderWidth Div 2) - Abs((BorderWidth Div 2) - Integer(Y)));
    For X := X1 + 1 To X2 - 1 Do
    Begin
      If ((X + Y + Y1) And 2) <> 0
       Then Mem[$A000:Addr1 + X] := C1 + CInc
       Else Mem[$A000:Addr1 + X] := C2 + CInc;
      If ((X + Y + Y2 - BorderWidth + 1) And 2) <> 0
       Then Mem[$A000:Addr2 + X] := C1 + CInc
       Else Mem[$A000:Addr2 + X] := C2 + CInc;
    End; { For X }
  End; { For Y }

  { Draw left and right borders }

  For Y := Y1 + 1 To Y2 - 1 Do
  Begin
    Addr1 := Width * Y;
    For X := X1 To X1 + BorderWidth - 1 Do
    Begin
      CInc  := 2 * ((BorderWidth Div 2) - Abs((BorderWidth Div 2) - Integer(X - X1)));
      If ((X + Y) And 2) <> 0
       Then Mem[$A000:Addr1 + X] := C1 + CInc
       Else Mem[$A000:Addr1 + X] := C2 + CInc;
    End; { For X }
    For X := X2 - BorderWidth + 1 To X2 Do
    Begin
      CInc  := 2 * ((BorderWidth Div 2) - Abs((BorderWidth Div 2) - Integer(X - (X2 - BorderWidth + 1))));
      If ((X + Y) And 2) <> 0
       Then Mem[$A000:Addr1 + X] := C1 + CInc
       Else Mem[$A000:Addr1 + X] := C2 + CInc;
    End; { For X }
  End; { For Y }

  { Draw center area }

  For Y := Y1 + BorderWidth To Y2 - BorderWidth Do
   FillChar(Mem[$A000:Width * Y + X1 + BorderWidth],X2 - X1 - 2 * BorderWidth + 1,BK);
End; { DrawWindow }

Procedure PutChar(X,Y: Word; C: Char; Color,Width: Word); Assembler;
Asm
  MOV   AX,WORD PTR Y
{
  SUB   BX,BX
  MOV   BH,AL
  SHL   AX,6
  ADD   BX,AX
  MOV   DI,BX
}
  MUL   WORD PTR Width
  MOV   DI,AX
  ADD   DI,WORD PTR X
  PUSH  DS
  LDS   SI,DWORD PTR FontPtr
  MOV   AL,BYTE PTR C
  SUB   AH,AH
  SHL   AX,3
  ADD   SI,AX
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   AL,8
  ADD   DI,7
  MOV   DX,WORD PTR Color
@Loop:
  MOV   BH,BYTE PTR [SI]
  MOV   CX,8
  MOV   BL,1
@L1:
  TEST  BH,BL
  JZ    @L2
  MOV   BYTE PTR [ES:DI],DL
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,WORD PTR Width
  INC   SI
  ADD   DI,8
  DEC   AL
  JNZ   @Loop
  POP   DS
  JMP   @L3
@L2:
  MOV   BYTE PTR [ES:DI],DH
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,WORD PTR Width
  INC   SI
  ADD   DI,8
  DEC   AL
  JNZ   @Loop
  POP   DS
@L3:
End; { PutChar }

Procedure WriteOut(X,Y: Word; St: String; Color,Width: Word);
Var I: Byte;
Begin
  If St = '' Then Exit;
  For I := 0 To Length(St) - 1 Do PutChar(X + (I * 8),Y,St[I + 1],Color,Width);
End; { WriteOut }

Procedure PutSmChar(X,Y: Word; C: Char; Color,Width: Word); Assembler;
Asm
  MOV   AX,WORD PTR Y
  MUL   WORD PTR Width
  MOV   DI,AX
  ADD   DI,WORD PTR X
  PUSH  DS
  LDS   SI,DWORD PTR FontSmPtr
  MOV   AL,BYTE PTR C
  SUB   AH,AH
  SHL   AX,3
  ADD   SI,AX
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   AL,8
  ADD   DI,7
  MOV   DX,WORD PTR Color
@Loop:
  MOV   BH,BYTE PTR [SI]
  MOV   CX,6
  MOV   BL,4
@L1:
  TEST  BH,BL
  JZ    @L2
  MOV   BYTE PTR [ES:DI],DL
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,WORD PTR Width
  INC   SI
  ADD   DI,6
  DEC   AL
  JNZ   @Loop
  POP   DS
  JMP   @L3
@L2:
  MOV   BYTE PTR [ES:DI],DH
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,WORD PTR Width
  INC   SI
  ADD   DI,6
  DEC   AL
  JNZ   @Loop
  POP   DS
@L3:
End; { PutSmChar }

Procedure WriteSmOut(X,Y: Word; St: String; Color,Width: Word);
Var I: Byte;
Begin
  If St = '' Then Exit;
  For I := 0 To Length(St) - 1 Do PutSmChar(X + (I * 6),Y,St[I + 1],Color,Width);
End; { WriteSmOut }

Procedure FBox(X1,Y1,X2,Y2,Color,Width: Word);
Var Y: Word;
Begin
  For Y := Y1 To Y2 Do FillChar(Mem[$A000:Width * Y + X1],X2 - X1 + 1,Color);
End; { FBox }

Function GetText(X1,Y1,X2,Y2: Integer; InitString: String; Fore,Back: Byte;
                 MaxLen,WindowLen: Integer; Width: Word): String;
{ ---------------------------------------------------------------------- }
{ Allows the user to type in a text string in graphics mode              }
{    X1 ................. Upper left X coord of text input box           }
{    Y1 ................. Upper left Y coord of text input box           }
{    X2 ................. Lower right X coord of text input box          }
{    Y2 ................. Lower right Y coord of text input box          }
{    InitString ......... Initial text string                            }
{    Fore ............... Text foreground color                          }
{    Back ............... Text background color                          }
{    MaxLen ............. Maximum string length                          }
{    WindowLen .......... Displayed string length, for scrollable window }
{ ---------------------------------------------------------------------- }
Const
  TextWidth  = 8;
  TextHeight = 8;

Var
  Work  : String;
  Ch    : Char;
  Ch1   : Char;
  Top   : Integer;
  X     : Integer;
  Ins   : Boolean;
  TPtr  : Pointer;
  On    : Boolean;
  Pen   : Word;
  Brush : Word;

  Procedure GetImage(X1,Y1,X2,Y2: Word; Var P; Width: Word);
  Var X,Y,Addr,Dest: Word;
  Begin
    For Y := Y1 To Y2 Do
    Begin
      Addr := Y * Width + X1;
      Dest := (Y - Y1) * (X2 - X1 + 1);
      Move(Mem[$A000:Addr],Mem[Seg(P):Ofs(P) + Dest],X2 - X1 + 1);
    End; { For Y }
  End; { GetImage }


  Procedure PutImage(X1,Y1,W,H: Word; Var P; Width: Word);
  Var X,Y,Addr,Source: Word;
  Begin
    For Y := Y1 To Y1 + H - 1 Do
    Begin
      Addr   := Y * Width + X1;
      Source := (Y - Y1) * W;
      Move(Mem[Seg(P):Ofs(P) + Source],Mem[$A000:Addr],W);
    End; { For Y }
  End; { PutImage }

  Function ImageSize(X1,Y1,X2,Y2: Word): Word;
  Begin
    ImageSize := (X2 - X1 + 1) * (Y2 - Y1 + 1);
  End; { ImageSize }


  Function Max(A,B: Integer): Integer;
  Begin
    If A >= B Then Max := A Else Max := B;
  End; { Max }


  Function Min(A,B: Integer): Integer;
  Begin
    If A <= B Then Min := A Else Min := B;
  End; { Min }


  Procedure Cursor;
  { --------------------------------- }
  { Toggle to display/hide the cursor }
  { --------------------------------- }
  Begin
    If On Then
    Begin
      Pen   := Back;
      Brush := Back;
    End
    Else
    Begin
      Pen   := Fore;
      Brush := Fore;
    End;
    If Ins
     Then FBox(X1 + (X - Top)     * TextWidth + 1,Y1 + 4,
               X1 + (X - Top + 1) * TextWidth,Y1 + TextHeight,Brush,Width)
     Else FBox(X1 + (X - Top)     * TextWidth + 1,Y1 + TextHeight,
               X1 + (X - Top + 1) * TextWidth,Y1 + TextHeight,Pen,Width);
    On := Not On;
    If Not On Then
    Begin
      Pen   := Fore;
      Brush := Back;
      WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,
       Copy(Work,X,1),Pen + (Brush Shl 8),Width);
    End;
    Pen   := Fore;
    Brush := Fore;
  End; { Cursor }


  Procedure ScrollLeft;
  { --------------------------- }
  { Scroll the text window left }
  { --------------------------- }
  Var W,H,S: Word;
  Begin
    Inc(Top);
    S := ImageSize(X1 + TextWidth + 1,Y1,X2,Y2);
    GetMem(TPtr,S);
    GetImage(X1 + TextWidth + 1,Y1,X2,Y2,TPtr^,Width);
    W     := X2 - (X1 + TextWidth + 1) + 1;
    H     := Y2 - Y1 + 1;
    Pen   := Back;
    Brush := Back;
    FBox(X1 + (WindowLen - 1) * TextWidth + 1,Y1,X2,Y2,Brush,Width);
    Pen   := Fore;
    PutImage(X1 + 1,Y1,W,H,TPtr^,Width);
    WriteOut(X1 + (WindowLen) * TextWidth + 1,Y1 + 1,
     Copy(Work,Top + WindowLen,1),Pen + (Brush Shl 8),Width);
    FreeMem(TPtr,S);
  End; { ScrollLeft }


  Procedure ScrollRight;
  { ---------------------------- }
  { Scroll the text window right }
  { ---------------------------- }
  Var W,H,S: Word;
  Begin
    Dec(Top);
    S := ImageSize(X1 + TextWidth + 1,Y1,X2 + (WindowLen - 1) * TextWidth,Y2);
    GetMem(TPtr,S);
    GetImage(X1 + 1,Y1,X2 - TextWidth,Y2,TPtr^,Width);
    W     := (X2 - TextWidth) - (X1 + 1) + 1;
    H     := Y2 - Y1 + 1;
    Pen   := Back;
    Brush := Back;
    FBox(X1,Y1,X1 + TextWidth,Y2,Brush,Width);
    Pen   := Fore;
    PutImage(X1 + TextWidth + 1,Y1,W,H,TPtr^,Width);
    WriteOut(X1 + 1,Y1 + 1,Copy(Work,Top,1),Pen + (Brush Shl 8),Width);
    FreeMem(TPtr,S);
  End; { ScrollRight }

Begin
  TM0 := Clock;
  TM2 := 0;
  BufPos   := 0;
  BufStart := 0;
  FillChar(OldKey,SizeOf(OldKey),#0);
  Work := InitString;
  Top  := 1;
  X    := 1;
  Ins  := False;
  On   := False;

  { Blank the window }

  Pen   := Back;
  Brush := Back;
  FBox(X1,Y1,X2,Y2,Brush,Width);
  Pen   := Fore;

  { Display the string }

  WriteOut(X1 + 1,Y1 + 1,Copy(Work,1,WindowLen + 1),Pen + (Brush Shl 8),Width);

  Cursor;
  Ch := ' ';
  Move(Key,OldKey,SizeOf(OldKey));
  While KeyPressed Do ReadKey;
  Wait := 250;
  LastKey := -1;
  While Not (Byte(Ch) In [kENTER,KEENTER,kESC]) Do
  Begin
    CheckNewKeyPressed;
    If KeyPressed Then
    Begin
      Ch := ReadKey;
      Case Byte(Ch) Of
  kENTER,kEENTER,kESC: Begin
            End;
  kBACKSPC: Begin                                   { Backspace }
              If X > 1 Then
              Begin
                Cursor;
                Delete(Work,X - 1,1);
                If X > Top Then
                Begin
                  Pen   := Back;
                  Brush := Back;
                  FBox(X1 + (X - Top - 1) * TextWidth + 1,Y1 + 1,X2 - 1,Y2 - 1,Brush,Width);
                  Pen   := Fore;
                  WriteOut(X1 + (X - Top - 1) * TextWidth + 1,Y1 + 1,
                           Copy(Work,X - 1,WindowLen - (X - Top) + 1),Pen + (Brush Shl 8),Width);
                End;
                Dec(X);
                If X < Top Then ScrollRight;
                Cursor;
              End;
            End;
 kHOME,kP7: If X > 1 Then                  { Home }
            Begin
              Cursor;
              If Top <> 1 Then
              Begin
                Pen   := Back;
                Brush := Back;
                FBox(X1,Y1,X2,Y2,Brush,Width);
                Pen := Fore;
                Top := 1;
                WriteOut(X1 + 1,Y1 + 1,Copy(Work,1,WindowLen + 1),Pen + (Brush Shl 8),Width);
              End;
              X := 1;
              Cursor;
            End;
   kLF,kP4: If X > 1 Then                  { Cursor Left }
            Begin
              Cursor;
              Dec(X);
              If X < Top Then ScrollRight;
              Cursor;
            End;
   kRT,kP6: If X <= Length(Work) Then      { Cursor Right }
            Begin
              Cursor;
              Inc(X);
              If X - Top = WindowLen Then ScrollLeft;
              Cursor;
            End;
  kEND,kP1: If X <= Length(Work) Then      { End }
            Begin
              Cursor;
              If Top < Length(Work) - WindowLen + 2 Then
              Begin
                Pen   := Back;
                Brush := Back;
                FBox(X1,Y1,X2,Y2,Brush,Width);
                Pen   := Fore;
                Top   := Length(Work) - WindowLen + 2;
                WriteOut(X1 + 1,Y1 + 1,Copy(Work,Top,WindowLen),Pen + (Brush Shl 8),Width);
              End;
              X := Length(Work) + 1;
              Cursor;
            End;
      kINS: Begin                          { Insert }
              Cursor;
              Ins := Not Ins;
              Cursor;
            End;
      kDEL: Begin                          { Delete }
              If ((Work <> '') And (X <= Length(Work))) Then
              Begin
                Cursor;
                Delete(Work,X,1);
                Pen   := Back;
                Brush := Back;
                FBox(X1 + (X - Top) * TextWidth + 1,Y1 + 1,X2 - 1,Y2 - 1,Brush,Width);
                Pen   := Fore;
                WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,
                 Copy(Work,X,WindowLen - (X - Top) + 1),Pen + (Brush Shl 8),Width);
                Cursor;
              End;
            End;
      Else
        If ProduceChar[Byte(Ch)] <> 0 Then
        Begin
          Cursor;
          If (X > Length(Work)) Or Ins Then
          Begin
            If (Length(Work) < MaxLen) Then
            Begin
              If (Key[kLSHIFT] <> 0) Or (Key[kRSHIFT] <> 0)
               Then Insert(ASCIITranslationShift[Byte(Ch)],Work,X)
               Else Insert(ASCIITranslation[Byte(Ch)],Work,X);
              Pen   := Back;
              Brush := Back;
              FBox(X1 + (X - Top) * TextWidth + 1,Y1 + 1,X2 - 1,Y2 - 1,Brush,Width);
              Pen   := Fore;
              WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,
               Copy(Work,X,WindowLen - (X - Top) + 1),Pen + (Brush Shl 8),Width);
            End;
          End
          Else
          Begin
            If (Key[kLSHIFT] <> 0) Or (Key[kRSHIFT] <> 0)
             Then Work[X] := ASCIITranslationShift[Byte(Ch)]
             Else Work[X] := ASCIITranslation[Byte(Ch)];
            Pen   := Back;
            Brush := Back;
            FBox(X1 + (X - Top)     * TextWidth + 1,Y1 + 1,
                 X1 + (X - Top + 1) * TextWidth    ,Y2 - 1,Brush,Width);
            Pen   := Fore;
            WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,Copy(Work,X,1),Pen + (Brush Shl 8),Width);
          End;
          If Not ((X > Length(Work)) And (Length(Work) = MaxLen)) Then
          Begin
            Inc(X);
             { Check to see if I must scroll the text left }
             If X - Top = WindowLen Then ScrollLeft;
          End;
          Cursor;
        End;
      End; { Case }
    End;
  End; { While }
  If Byte(Ch) = kESC Then
  Begin
    Work := InitString;
    Repeat
    Until Key[kEsc] = 0;
  End;
  Pen   := Back;
  Brush := Back;
  FBox(X1,Y1,X2,Y2,Brush,Width);
  Pen   := Fore;
  WriteOut(X1 + 1,Y1 + 1,Copy(Work,1,WindowLen),Pen + (Brush Shl 8),Width);
  GetText := Work;
End; { GetText }

Procedure GameStatus(Width,Height: Word; Var Difficulty1,Difficulty2: LongInt; Var AutoCenter: Boolean);
Var
  Buf     : Pointer;
  BufSize : LongInt;

  Procedure DrawDiff(X,Y: Word; Diff: Boolean; Width: Word);
  Begin
    If Diff Then WriteOut(X,Y,'A',$089F,Width) Else WriteOut(X,Y,'B',$089F,Width);
  End; { DrawDiff }

  Procedure DrawYN(X,Y: Word; Diff: Boolean; Width: Word);
  Begin
    If Diff Then WriteOut(X,Y,'Y',$089F,Width) Else WriteOut(X,Y,'N',$089F,Width);
  End; { DrawDiff }

Begin
  BufSize := Width * Height;
  GetMem(Buf,BufSize);
  SaveScreen(Width,BufSize,Buf);
  Key[kEsc] := 0;
  Dim(BufSize);                { Dim the screen }
  If Width = 320 Then
  Begin
    DrawWindow(100,15,220,185,4,5,28,Width);
    WriteOut(108,24,'PCAE 2.4',$081E,Width);
    WriteSmOut(108,40, 'Esc - Exit game',$0800,Width);
    WriteSmOut(108,50, 'F1  - Help',$0800,Width);
    WriteSmOut(108,60, 'F2  - Reset',$0800,Width);
    WriteSmOut(108,70, 'F3  - Toggle PAL',$0800,Width);
    WriteSmOut(108,80, 'F4  - Select',$0800,Width);
    WriteSmOut(108,90, 'F5  - Color/BW',$0800,Width);
    WriteSmOut(108,100,'F6  - P1 diff.',$0886,Width);
    WriteSmOut(108,110,'F7  - P2 diff.',$0886,Width);
    WriteSmOut(108,120,'F8  - Save info',$0800,Width);
    WriteSmOut(108,130,'F9  - Sc.capture',$0800,Width);
    WriteSmOut(108,140,'F10 - A.Center',$0886,Width);
    WriteSmOut(108,150,'F11 - Save game',$0800,Width);
    WriteSmOut(108,160,'F12 - Load game',$0800,Width);
    WriteSmOut(108,170,'-/= - Shift dn/up',$0800,Width);
    DrawDiff(200,100,((Difficulty1 And $40) <> 0),Width);
    DrawDiff(200,110,((Difficulty2 And $80) <> 0),Width);
    DrawYN(200,140,AutoCenter,Width);
  End
  Else
  Begin
    DrawWindow(20,15,140,185,4,5,28,Width);
    WriteOut(28,24,'PCAE 2.4',$081E,Width);
    WriteSmOut(28,40, 'Esc - Exit game',$0800,Width);
    WriteSmOut(28,50, 'F1  - Help',$0800,Width);
    WriteSmOut(28,60, 'F2  - Reset',$0800,Width);
    WriteSmOut(28,70, 'F3  - Toggle PAL',$0800,Width);
    WriteSmOut(28,80, 'F4  - Select',$0800,Width);
    WriteSmOut(28,90, 'F5  - Color/BW',$0800,Width);
    WriteSmOut(28,100,'F6  - P1 diff.',$0886,Width);
    WriteSmOut(28,110,'F7  - P2 diff.',$0886,Width);
    WriteSmOut(28,120,'F8  - Save info',$0800,Width);
    WriteSmOut(28,130,'F9  - Sc.Capture',$0800,Width);
    WriteSmOut(28,140,'F10 - A.Center',$0886,Width);
    WriteSmOut(28,150,'F11 - Save game',$0800,Width);
    WriteSmOut(28,160,'F12 - Load game',$0800,Width);
    WriteSmOut(28,170,'-/= - Shift dn/up',$0800,Width);
    DrawDiff(120,100,((Difficulty1 And $40) <> 0),Width);
    DrawDiff(120,110,((Difficulty2 And $80) <> 0),Width);
    DrawYN(120,140,AutoCenter,Width);
  End;
  Repeat
    If Key[kF6] <> 0 Then
    Begin
      Difficulty1 := Difficulty1 Xor $40;
      If Width = 320
       Then DrawDiff(200,100,((Difficulty1 And $40) <> 0),Width)
       Else DrawDiff(120,100,((Difficulty1 And $40) <> 0),Width);
      Key[kF6] := 0;
    End;
    If Key[kF7] <> 0 Then
    Begin
      Difficulty2 := Difficulty2 Xor $80;
      If Width = 320
       Then DrawDiff(200,110,((Difficulty2 And $80) <> 0),Width)
       Else DrawDiff(120,110,((Difficulty2 And $80) <> 0),Width);
      Key[kF7] := 0;
    End;
    If Key[kF10] <> 0 Then
    Begin
      AutoCenter := Not AutoCenter;
      If Width = 320
       Then DrawYN(200,140,AutoCenter,Width)
       Else DrawYN(120,140,AutoCenter,Width);
      Key[kF10] := 0;
    End;
  Until Key[kEsc] <> 0;
  Repeat
  Until Key[kEsc] = 0;
  RestoreScreen(Width,BufSize,Buf);
  FreeMem(Buf,BufSize);
End; { GameStatus }


Function EnterStringPopup(Width,Height: Word; Title,OldStr: String): String;
Var
  Buf     : Pointer;
  BufSize : LongInt;
  St      : String;
  Error   : Boolean;
  F       : File;
  Indent  : Integer;

Begin
  BufSize := Width * Height;
  GetMem(Buf,BufSize);
  SaveScreen(Width,BufSize,Buf);
  Key[kEsc] := 0;
  Dim(BufSize);                { Dim the screen }
  Error := False;

  If Width = 320 Then
  Begin
    Indent := 20 + Round(Length(Title) / 16);
    If Indent < 0 Then Indent := 0;
    Title := Copy(Title,1,33);

    DrawWindow(20,70,300,130,4,5,28,Width);
    WriteOut(20 + Indent,79,Title,$081E,Width);
    EnterStringPopup := Trim(GetText(28,96,285,105,OldStr,7,0,60,30,Width));
  End
  Else
  Begin
    Indent := 10 + Round(Length(Title) / 16);
    If Indent < 0 Then Indent := 0;
    Title := Copy(Title,1,15);

    DrawWindow(10,70,150,130,4,5,28,Width);
    WriteOut(10 + Indent,79,Title,$081E,Width);
    EnterStringPopup := Trim(GetText(16,96,144,105,OldStr,7,0,60,15,Width));
  End;
  RestoreScreen(Width,BufSize,Buf);
  FreeMem(Buf,BufSize);
End; { EnterStringPopup }


Function CanOpenFile(Width,Height: Word; St: String): Boolean;
Var
  Buf     : Pointer;
  BufSize : LongInt;
  Error   : Boolean;

Begin
  Error   := False;
  BufSize := Width * Height;
  If St = '' Then Error := True;
  If Not Exist(St) Then Error := True;
  If Error Then
  Begin
    Repeat
    Until Key[kEsc] = 0;
    If St <> '' Then
    Begin
      GetMem(Buf,BufSize);
      SaveScreen(Width,BufSize,Buf);
      Dim(BufSize);
      If Width = 320 Then
      Begin
        DrawWindow(40,70,280,130,4,5,28,Width);
        WriteOut(94,92, 'Error opening file',$081E,Width);
        WriteOut(94,100,'Hit Esc to return',$081E,Width);
      End
      Else
      Begin
        DrawWindow(20,70,140,130,4,5,28,Width);
        WriteOut(16,92, 'Error openng file',$081E,Width);
        WriteOut(16,100,'Hit Esc to return',$081E,Width);
      End;
      Repeat
      Until Key[kEsc] = 0;
      Repeat
      Until Key[kEsc] <> 0;
      Repeat
      Until Key[kEsc] = 0;
      RestoreScreen(Width,BufSize,Buf);
      FreeMem(Buf,BufSize);
    End;
    CanOpenFile := False;
  End
  Else CanOpenFile := True;
End; { CanOpenFile }


Function CreateFile(Width,Height: Word; Var F: File; St: String): Boolean;
Var
  Buf      : Pointer;
  BufSize  : LongInt;
  Error    : Boolean;
  Continue : Boolean;

Begin
  Error   := False;
  BufSize := Width * Height;
  If St = '' Then Error := True;

  Continue := True;
  If (Not Error) And Exist(St) Then
  Begin
    Key[kY] := 0;
    Key[kN] := 0;
    GetMem(Buf,BufSize);
    SaveScreen(Width,BufSize,Buf);
    Dim(BufSize);
    If Width = 320 Then
    Begin
      DrawWindow(40,70,280,130,4,5,28,Width);
      WriteOut(110,92, 'File exists!',$081E,Width);
      WriteOut(94,100,'Overwrite? (Y/N)',$081E,Width);
    End
    Else
    Begin
      DrawWindow(20,70,140,130,4,5,28,Width);
      WriteOut(32,92, 'File exists!',$081E,Width);
      WriteOut(24,100,'Overwrite? Y/N',$081E,Width);
    End;
    Repeat
    Until (Key[kY] = 0) And (Key[kN] = 0);
    Repeat
    Until (Key[kY] <> 0) Or (Key[kN] <> 0);
    Continue := (Key[kY] <> 0);
    Repeat
    Until (Key[kY] = 0) And (Key[kN] = 0);
    RestoreScreen(Width,BufSize,Buf);
    FreeMem(Buf,BufSize);
  End;
  If Error Then Continue := False;

  If Continue Then
  Begin
    Assign(F,St);
    {$I-}
    ReWrite(F,1);
    {$I+}
    If IOResult <> 0 Then Error := True;
    If Error Then
    Begin
      Repeat
      Until Key[kEsc] = 0;
      If St <> '' Then
      Begin
        GetMem(Buf,BufSize);
        SaveScreen(Width,BufSize,Buf);
        Dim(BufSize);
        If Width = 320 Then
        Begin
          DrawWindow(40,70,280,130,4,5,28,Width);
          WriteOut(86,92, 'Error creating file',$081E,Width);
          WriteOut(94,100,'Hit Esc to return',$081E,Width);
        End
        Else
        Begin
          DrawWindow(20,70,140,130,4,5,28,Width);
          WriteOut(16,92, 'Error writing file',$081E,Width);
          WriteOut(16,100,'Hit Esc to return',$081E,Width);
        End;
        Repeat
        Until Key[kEsc] = 0;
        Repeat
        Until Key[kEsc] <> 0;
        Repeat
        Until Key[kEsc] = 0;
        RestoreScreen(Width,BufSize,Buf);
        FreeMem(Buf,BufSize);
      End;
      CreateFile := False;
    End
    Else CreateFile := True;
  End
  Else CreateFile := False;
End; { CreateFile }


Procedure ScreenCapture(Width,Height,Palette: Word; Pal0,Pal1: PPalette);
Var
  Buf     : Pointer;
  BufSize : LongInt;
  St      : String;
  Error   : Boolean;
  F       : File;

  Procedure MakeBMPFile(Var F: File);
  Var
    BF      : BITMAPFILEHEADER;
    BI      : BITMAPINFOHEADER;
    I       : Integer;
    Q       : RGBQuad;

  Begin
    FillChar(BF,SizeOf(BF),#0);
    FillChar(BI,SizeOf(BI),#0);

    BF.bfType[0] := 'B';
    BF.bfType[1] := 'M';
    BF.bfSize    := SizeOf(BF) + SizeOf(BI) + 256 * SizeOf(RGBQuad) + BufSize;
    BF.bfOffBits := SizeOf(BF) + SizeOf(BI) + 256 * SizeOf(RGBQuad);

    BI.biSize          := 40;
    BI.biWidth         := Width;
    BI.biHeight        := Height;  { Top-down bitmap }
    BI.biPlanes        := 1;
    BI.biBitCount      := 8;
    BI.biCompression   := BI_RGB;
    BI.biSizeImage     := 0;
    BI.biXPelsPerMeter := 2000;
    BI.biYPelsPerMeter := 2000;
    BI.biClrUsed       := 256;
    BI.biClrImportant  := 256;

    BlockWrite(F,BF,SizeOf(BF));
    BlockWrite(F,BI,SizeOf(BI));

    If Palette = 0 Then
    Begin
      For I := 0 To 255 Do
      Begin
        Q.R := Pal0^[I Shr 1].B;
        Q.G := Pal0^[I Shr 1].G;
        Q.B := Pal0^[I Shr 1].R;
        Q.Reserved := 0;
        BlockWrite(F,Q,SizeOf(Q));
      End; { For I }
    End
    Else
    Begin
      For I := 0 To 255 Do
      Begin
        Q.R := Pal1^[I Shr 1].B;
        Q.G := Pal1^[I Shr 1].G;
        Q.B := Pal1^[I Shr 1].R;
        Q.Reserved := 0;
        BlockWrite(F,Q,SizeOf(Q));
      End; { For I }
    End;
    For I := 0 To Height - 1 Do
     BlockWrite(F,Mem[Seg(Buf^):Ofs(Buf^) + (Height - I - 1) * Width],Width);
  End; { MakeBMPFile }

Begin
  Key[kEsc] := 0;
  BufSize := Width * Height;
  GetMem(Buf,BufSize);
  Move(Mem[$A000:0],Buf^,BufSize);
  Dim(BufSize);
  Error := False;
  If Width = 320 Then
  Begin
    St := EnterStringPopup(Width,Height,'Save screen as: (Windows BMP)','');
    If CreateFile(Width,Height,F,St) Then
    Begin
      MakeBMPFile(F);
      Close(F);
    End;
  End
  Else
  Begin
    St := EnterStringPopup(Width,Height,'Save screen as:','');
    If CreateFile(Width,Height,F,St) Then
    Begin
      MakeBMPFile(F);
      Close(F);
    End;
  End;
  Move(Buf^,Mem[$A000:0],BufSize);
  FreeMem(Buf,BufSize);
End; { ScreenCapture }

End.
