Unit ProtMenu;
{$F+}
{ ---------------------------------------------------------------------- }
{ This unit provides basic support for pop-up windows and menus in DOS   }
{ protected mode.                                                        }
{ ---------------------------------------------------------------------- }

Interface

Uses Crt,Objects,VESA;

Type
  Rect = Record
    Active      : Boolean;
    X1,Y1,X2,Y2 : Integer;
    AUL,AUR,ABL,ABR,AHZ,AVT : String[1];
    PUL,PUR,PBL,PBR,PHZ,PVT : String[1];
    PFore,PBack : Byte;
    AFore,ABack : Byte;
  End;

Const
  ScreenWidth = 200;
{ ------------------------ Keys ------------------------ }
  Key_Null    = #0;
  Key_Tab     = #9;
  Key_Enter   = #13;
  Key_Esc     = #27;
  Key_F1      = #59;
  Key_Alt_F5  = #108;
  Key_Alt_F10 = #113;
  Key_Up      = #72;
  Key_Down    = #80;
  Key_Left    = #75;
  Key_Right   = #77;
  Key_PgUp    = #73;
  Key_PgDn    = #81;
  Key_Home    = #71;
  Key_End     = #79;
{ -------------------------------------------------------- }
{ FastScroll speeds up character drawing for VESA graphics }
{ modes by cutting corners: it causes characters to be     }
{ written only if the character or attribute has changed.  }
{ This is OK for text applications, but could cause        }
{ strange results on screen if graphics is drawn on the    }
{ screen.                                                  }
{ -------------------------------------------------------- }
  FastWriteChar : Boolean = True;

Type
  TDisplayType = (dtText,dtVESAText,dtVESA256,dtVESATrue);
  PScreenMode  = ^TScreenMode;
  TScreenMode  = Object(TObject)
    DisplayType : TDisplayType;
    Width       : Word;
    Height      : Word;
    ModeNumber  : Word;
    BPP         : Word;
  End;
  TFontChar = Array[0..15] Of Byte;
  TScreen = Object(TObject)
  Private
    VESAInfo     : VESAInfoRec;
    VESAModeInfo : VESAModeInfoRec;
    VESAWinAddr  : Word;
    VESAInc      : Word;
    CharBuf      : Pointer;
    CharBufSize  : Word;
    Procedure   TextWrite(P: Pointer; X,Y: Integer; Color: Byte; Var S: String);
    Procedure   TextShade(P: Pointer; X,Y: Integer; Len: Byte);
    Procedure   VESAWrite256(X,Y: Integer; Color: Byte; Var S: String);
    Procedure   VESAWriteTrue(X,Y: Integer; Color: Byte; Var S: String);
    Procedure   VESAShade(X,Y: Integer; Len: Byte);
    Procedure   VESAMove(Source: Pointer; Dest,NumBytes: Word);
    Function    GetNumModes: Integer;
    Function    SetVESAWindow(Addr: LongInt): Word;
    Procedure   VESAScrollWindowUp;
    Procedure   VESAShowCursor;
    Procedure   VESAHideCursor;
  Public
    DisplayType : TDisplayType;
    Width       : Word;
    Height      : Word;
    NumModes    : Word;
    ScreenModes : PCollection;
    CurrentMode : Word;
    TextAttr    : Word;
    WhereX      : Word;
    WhereY      : Word;
    ScreenPtr   : Pointer;
    WindMin     : Word;
    WindMax     : Word;
    BPP         : Word;
    CursorStart : Byte;
    CursorEnd   : Byte;
    CursorOn    : Boolean;
    Constructor Init;
    Destructor  Done; Virtual;
    Procedure   FastWrite(X,Y: Integer; Color: Byte; S: String);
    Procedure   FastShade(X,Y: Integer; Len: Byte);
    Procedure   SetMode(Mode: Integer);
    Procedure   ClrScr;
    Procedure   ClrEol;
    Procedure   TextColor(Color: Byte);
    Procedure   TextBackground(Color: Byte);
    Procedure   GotoXY(X,Y: Byte);
    Procedure   Write(St: String);
    Procedure   WriteLn(St: String);
    Procedure   FastMove(Source: Pointer; Addr,NumBytes: Word);
    Procedure   Window(X1,Y1,X2,Y2: Byte);
    Procedure   SetTextAttr(Attr: Byte);
    Procedure   ShowCursor(Show: Boolean);
    Procedure   SetCursor(MinC,MaxC: Byte);
  End;
  BorderType = (SingleBorder,DoubleBorder);
  ScreenStr  = String[ScreenWidth];
  PSelection = ^DSelection;
  DSelection = Object(TObject)
    X,Y            : Integer;
    Width          : Integer;
    SelStr         : ScreenStr;
    HotChar        : Byte;
    Protected      : Boolean;
    Checked        : Boolean;
    Asterisk       : Boolean;
    Bar            : Integer;
    WindowAttr     : Byte;
    HotCharAttr    : Byte;
    SelAttr        : Byte;
    SelHotCharAttr : Byte;
    ProtectedAttr  : Byte;
    SelProtAttr    : Byte;
    Constructor Init(_X,_Y,_Width: Integer;
                     Const _SelStr: ScreenStr;
                     _HotChar: Byte;
                     _Protected: Boolean;
                     _Checked: Boolean;
                     _WindowAttr: Byte;
                     _HotCharAttr: Byte;
                     _SelAttr: Byte;
                     _SelHotCharAttr: Byte;
                     _ProtectedAttr: Byte;
                     _SelProtAttr: Byte);
    Destructor  Done; Virtual;
    Procedure   Show(Selected: Boolean; Top: Integer);
  End;
  PWindow = ^DWindow;
  DWindow = Object(TObject)
    OX1,OY1     : Integer;
    OX2,OY2     : Integer;
    X1,Y1,X2,Y2 : Integer;
    Size        : Word;
    WindowAttr  : Byte;
    BorderAttr  : Byte;
    CaptionAttr : Byte;
    CaptionTop  : ScreenStr;
    CaptionBot  : ScreenStr;
    Border      : BorderType;
    WindowPtr   : Pointer;
    Visible     : Boolean;
    Shadow      : Boolean;
    LastWindMin : Word;
    LastWindMax : Word;
    Constructor Init(_X1,_Y1,_X2,_Y2: Integer;
                     WFore,WBack,BFore,BBack,CFore,CBack: Byte;
                     BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
    Constructor InitWH(Width,Height: Integer;
                     WFore,WBack,BFore,BBack,CFore,CBack: Byte;
                     BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
    Procedure   Display; Virtual;
    Procedure   Remove;  Virtual;
    Destructor  Done;    Virtual;
    Procedure   MoveWindow(DeltaX,DeltaY: Integer);
    Procedure   WriteFrame(Var R: Rect);
    Procedure   ShadeAt(X,Y: Integer; Len: Byte);
    Procedure   WriteArrows(Up,Down: Boolean);
  End;
  DMenuCallback = Procedure;
  PMenu = ^DMenu;
  DMenu = Object(DWindow)
    SelList        : PCollection;
    CurSel         : Integer;
    OldSel         : Integer;
    HotCharAttr    : Byte;
    SelAttr        : Byte;
    SelHotCharAttr : Byte;
    ProtectedAttr  : Byte;
    SelProtAttr    : Byte;
    Callback       : DMenuCallback;
    Top            : Integer;
    MaxSelY        : Integer;
  Private
    Procedure   CheckTop;
  Public
    Constructor Init(_X1,_Y1,_X2,_Y2: Integer;
                     WFore,WBack,BFore,BBack,CFore,CBack,HCFore,HCBack,
                     SFore,SBack,SHCFore,SHCBack,PFore,PBack,SPFore,SPBack: Byte;
                     BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
    Constructor InitWH(Width,Height: Integer;
                       WFore,WBack,BFore,BBack,CFore,CBack,HCFore,HCBack,
                       SFore,SBack,SHCFore,SHCBack,PFore,PBack,SPFore,SPBack: Byte;
                       BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
    Procedure   Display; Virtual;
    Procedure   AddSelection(_X,_Y,SelWidth: Integer; HotCh: Byte;
                             Protect,Checked: Boolean; St: ScreenStr);
    Procedure   SetSelection(Sel: Integer);
    Procedure   Execute;
    Procedure   Remove;  Virtual;
    Destructor  Done;    Virtual;
    Function    GetSelection(Sel: Integer): PSelection;
    Procedure   DisplaySelection(Sel: Integer);
  End;

Const
  WindowRect: Rect = (Active : False;
                      X1     : 1;
                      Y1     : 1;
                      X2     : 1;
                      Y2     : 1;
                      AUL    : '';
                      AUR    : '';
                      ABL    : '';
                      ABR    : '';
                      AHZ    : '';
                      AVT    : '';
                      PUL    : '';
                      PUR    : '';
                      PBL    : '';
                      PBR    : '';
                      PHZ    : '';
                      PVT    : '';
                      PFore  : LightGray;
                      PBack  : Black;
                      AFore  : LightGray;
                      ABack  : Black);

Var
  MaxX   : Integer;
  MaxY   : Integer;
  Screen : TScreen;

Implementation

Uses Dos,HexWrite,ScrnFont;

Const
  TextPalette : Array[0..16 * 3 - 1] Of Byte =
   ($00,$00,$00, $00,$00,$AA, $00,$AA,$00, $00,$AA,$AA,
    $AA,$00,$00, $AA,$00,$AA, $AA,$55,$00, $AA,$AA,$AA,
    $55,$55,$55, $55,$55,$FF, $55,$FF,$55, $55,$FF,$FF,
    $FF,$55,$55, $FF,$55,$FF, $FF,$FF,$55, $FF,$FF,$FF);
  _32BIT_OPERAND = $66;
  CopyLastPixel : Array[0..255] Of Byte =
   (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 00 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 10 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 20 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 30 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 40 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 50 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 60 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 70 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 80 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 90 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { A0 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { B0 }
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,   { C0 }
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,   { D0 }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,   { 0E }
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);  { F0 }

Type
  TRGBQuad = Record
    B,G,R,A: Byte;
  End;

Var
  OldExitProc : Pointer;

{ ------------------------------ }
{ TScreen                        }
{ ------------------------------ }

Constructor TScreen.Init;
Begin
  Inherited Init;
  NumModes    := 1;
  CurrentMode := 0;
  DisplayType := dtText;
  CharBuf     := Nil;
  CharBufSize := 0;
  VESAWinAddr := 0;
  Width       := Mem[$40:$4A];
  Height      := Mem[$40:$84] + 1;
  WhereX      := 1;
  WhereY      := 1;
  ScreenPtr   := Ptr($B800,0);
  If Height = 1 Then Height := 25;
  MaxX := Width;
  MaxY := Height;
  ScreenModes := New(PCollection,Init(10,10));
  VESAInc     := 0;
  GetNumModes;
  CursorStart := 6;
  CursorEnd   := 7;
  CursorOn    := True;

  SetMode(0);
End; { TScreen.Init }

Destructor TScreen.Done;
Var I: Integer;
Begin
  Dispose(ScreenModes,Done);
  Inherited Done;
  TextMode(C80);
End; { TScreen.Done }

Procedure TScreen.TextWrite(P: Pointer; X,Y: Integer; Color: Byte;
                            Var S: String);
Var Addr: Word;
Begin
  Addr := ((Y - 1) * MaxX + X - 1) * 2 + Ofs(P^);
  Asm
    CLI
    CLD
    MOV   AX,WORD PTR P[2]{0B800h}
    MOV   ES,AX
    MOV   DI,WORD PTR Addr
    MOV   AH,BYTE PTR Color
    PUSH  DS
    LDS   SI,DWORD PTR S
    MOV   CL,BYTE PTR [SI]
    SUB   CH,CH
    INC   SI
@L1:
    LODSB
    STOSW
    LOOP  @L1
    POP   DS
    STI
  End; { Asm }
End; { TScreen.TextWrite }


Procedure TScreen.TextShade(P: Pointer; X,Y: Integer; Len: Byte);
Var Addr: Word;
Begin
  Addr := ((Y - 1) * MaxX + X - 1) * 2 + Ofs(P^);
  Asm
    CLI
    CLD
    MOV    DI,WORD PTR Addr
    MOV    AX,WORD PTR P[2]{0B800h}
    MOV    ES,AX
    SUB    CX,CX
    MOV    CL,BYTE PTR Len
    MOV    AL,8
@L1:
    INC    DI
    STOSB
    LOOP   @L1
    STI
  End; { Asm }
End; { TScreen.TextShade }


Function TScreen.SetVESAWindow(Addr: LongInt): Word;
Var
  A       : LongInt;
  WG      : LongInt;
  WinAddr : Word;
  PosFn   : Pointer;

Begin
  PosFn   := VESAModeInfo.PosFunction;
  WG      := VESAModeInfo.WindowGranularity;
  A       := Addr Div (WG * 1024);
  Dec(Addr,A * WG * 1024);
  VESAInc := 64 Div WG;
  WinAddr := A;
  If WinAddr <> VESAWinAddr Then
  Begin
    Asm
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
    End; { Asm }
    VESAWinAddr := WinAddr;
  End;
  SetVESAWindow := Addr And $FFFF;
End; { TScreen.SetVESAWindow }


Procedure TScreen.VESAWrite256(X,Y: Integer; Color: Byte; Var S: String);
Var
  LX,LY   : LongInt;
  LW      : LongInt;
  Addr    : LongInt;
  WinAddr : Word;
  DX,DY   : Word;
  Mask    : Byte;
  Index   : Word;
  B       : Byte;
  C       : Byte;
  Ch,Ch1  : Word;
  AddrW   : Word;
  AddrW0  : Word;
  Fore    : Byte;
  Back    : Byte;
  PosFn   : Pointer;
  VInc    : Word;
  CopyLP  : Byte;
  AddInc  : Word;
  P,P0    : ^Word;

Begin
  PosFn   := VESAModeInfo.PosFunction;

  LX      := X - 1;
  LY      := Y - 1;
  LW      := Width;
  Addr    := LY * LW * 16 + LX * 9;
  VInc    := VESAInc;

  P  := Ptr(Seg(CharBuf^),Ofs(CharBuf^) + ((LY * MaxX) + LX) * 2);
  P0 := P;

  AddrW   := SetVESAWindow(Addr);
  WinAddr := VESAWinAddr;

  { Write the string }

  Fore   := Color And $F;
  Back   := Color Shr 4;
  AddInc := Width - Length(S) * 9;
  For DY := 0 To 15 Do
  Begin
    P := P0;
    For Index := 1 To Length(S) Do
    Begin
      Ch     := Word(S[Index]);
      CopyLP := CopyLastPixel[Ch];
      Ch1    := P^;
      Inc(LongInt(P),2);
      If (Ch <> Lo(Ch1)) Or (Color <> Hi(Ch1)) Or Not FastWriteChar Then
      Begin
        B := Font9x16[Ch * 16 + DY];
        Asm
          MOV   AX,0A000h
          MOV   ES,AX
          MOV   DI,WORD PTR AddrW
          MOV   CX,8
          MOV   AH,BYTE PTR B
          MOV   AL,80h
          MOV   DL,BYTE PTR Back
          MOV   DH,BYTE PTR Fore
          CLD
  @CharLoop:
          TEST  AH,AL
          JZ    @BackDot

          MOV   BYTE PTR [ES:DI],DH  { Write foreground color }
          SHR   AL,1
          INC   DI
          JNZ   @NoCarryFore
          PUSH  AX
          PUSH  CX
          PUSH  DX
          PUSH  ES
          PUSH  DI
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn

          POP   DI
          POP   ES
          POP   DX
          POP   CX
          POP   AX
  @NoCarryFore:
          DEC   CX
          JNZ   @CharLoop

          { Copy the last pixel if necessary }

          TEST  BYTE PTR CopyLP,0FFh
          JZ    @CopyBack
          MOV   BYTE PTR [ES:DI],DH
          JMP   @LastPixel
  @BackDot:
          MOV   BYTE PTR [ES:DI],DL  { Write background color }
          SHR   AL,1
          INC   DI
          JNZ   @NoCarryBack
          PUSH  AX
          PUSH  CX
          PUSH  DX
          PUSH  ES
          PUSH  DI
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   DI
          POP   ES
          POP   DX
          POP   CX
          POP   AX
  @NoCarryBack:
          DEC   CX
          JNZ   @CharLoop
  @CopyBack:

          { Copy the last pixel if necessary }

          MOV   BYTE PTR [ES:DI],DL
  @LastPixel:
          INC   DI
          JNZ   @NoCarry1
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          PUSHA
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POPA
  @NoCarry1:
          MOV   WORD PTR AddrW,DI
        End; { Asm }
        VESAWinAddr := WinAddr;
      End
      Else
      Begin
        AddrW0 := AddrW;
        Inc(AddrW,9);
        If AddrW < AddrW0 Then
        Begin
          Inc(WinAddr,VInc);
          Asm
            SUB   BX,BX
            MOV   DX,WORD PTR WinAddr
            CALL  DWORD PTR PosFn
          End; { Asm }
          VESAWinAddr := WinAddr;
        End;
      End;
    End; { For Index }
    AddrW0 := AddrW;
    Inc(AddrW,AddInc);
    If AddrW < AddrW0 Then
    Begin
      Inc(WinAddr,VInc);
      Asm
        SUB   BX,BX
        MOV   DX,WORD PTR WinAddr
        CALL  DWORD PTR PosFn
      End; { Asm }
      VESAWinAddr := WinAddr;
    End;
  End; { For DY }
  TextWrite(CharBuf,X,Y,Color,S);
End; { TScreen.VESAWrite256 }


Procedure TScreen.VESAWriteTrue(X,Y: Integer; Color: Byte; Var S: String);
Var
  LX,LY   : LongInt;
  LW      : LongInt;
  Addr    : LongInt;
  WinAddr : Word;
  DX,DY   : Word;
  Mask    : Byte;
  Index   : Word;
  B       : Byte;
  C       : Byte;
  Ch,Ch1  : Word;
  AddrW   : Word;
  AddrW0  : Word;
  Fore    : LongInt;
  Back    : LongInt;
  PosFn   : Pointer;
  VInc    : Word;
  CopyLP  : Byte;
  AddInc  : Word;
  BytesPerPixel : LongInt;
  FR,FG,FB      : Word;
  BR,BG,BB      : Word;
  P,P0    : ^Word;

Begin
  BytesPerPixel := BPP Div 8;

  PosFn   := VESAModeInfo.PosFunction;

  LX      := X - 1;
  LY      := Y - 1;
  LW      := Width;
  Addr    := (LY * LW * 16 + LX * 9) * BytesPerPixel;
  VInc    := VESAInc;

  P  := Ptr(Seg(CharBuf^),Ofs(CharBuf^) + ((LY * MaxX) + LX) * 2);
  P0 := P;

  AddrW   := SetVESAWindow(Addr);
  WinAddr := VESAWinAddr;

  { Write the string }

  FR := TextPalette[(Color And $F) * 3 + 0];
  FG := TextPalette[(Color And $F) * 3 + 1];
  FB := TextPalette[(Color And $F) * 3 + 2];
  BR := TextPalette[(Color Shr 4)  * 3 + 0];
  BG := TextPalette[(Color Shr 4)  * 3 + 1];
  BB := TextPalette[(Color Shr 4)  * 3 + 2];
  Case BPP Of
    16: Begin
      Fore := ((FR Shr (8 - VESAModeInfo.RMaskSize)) Shl VESAModeInfo.RFieldPos) +
              ((FG Shr (8 - VESAModeInfo.GMaskSize)) Shl VESAModeInfo.GFieldPos) +
              ((FB Shr (8 - VESAModeInfo.BMaskSize)) Shl VESAModeInfo.BFieldPos);
      Back := ((BR Shr (8 - VESAModeInfo.RMaskSize)) Shl VESAModeInfo.RFieldPos) +
              ((BG Shr (8 - VESAModeInfo.GMaskSize)) Shl VESAModeInfo.GFieldPos) +
              ((BB Shr (8 - VESAModeInfo.BMaskSize)) Shl VESAModeInfo.BFieldPos);
        End;
 24,32: Begin
          TRGBQuad(Fore).R := FR;
          TRGBQuad(Fore).G := FG;
          TRGBQuad(Fore).B := FB;
          TRGBQuad(Fore).A := 0;
          TRGBQuad(Back).R := BR;
          TRGBQuad(Back).G := BG;
          TRGBQuad(Back).B := BB;
          TRGBQuad(Back).A := 0;
        End;
  End; { Case }
  AddInc := (Width - Length(S) * 9) * BytesPerPixel;
  For DY := 0 To 15 Do
  Begin
    P := P0;
    For Index := 1 To Length(S) Do
    Begin
      Ch     := Word(S[Index]);
      B      := Font9x16[Ch * 16 + DY];
      CopyLP := CopyLastPixel[Ch];
      Ch1    := P^;
      Inc(LongInt(P),2);
      If (Ch <> Lo(Ch1)) Or (Color <> Hi(Ch1)) Or Not FastWriteChar Then
      Begin
        Asm
          MOV   AX,0A000h
          MOV   ES,AX
          MOV   DI,WORD PTR AddrW
          MOV   SI,8
          MOV   AH,BYTE PTR B
          MOV   AL,80h
          DB    _32BIT_OPERAND
          MOV   DX,WORD PTR Fore
          DB    _32BIT_OPERAND
          MOV   CX,WORD PTR Back
          CLD
  @CharLoop:
          TEST  AH,AL
          JZ    @BackDot

          { Write foreground color }

          CMP   WORD PTR BytesPerPixel,2
          JNE   @Not16Bit
          MOV   WORD PTR [ES:DI],DX
          JMP   @WroteFore
  @Not16Bit:
          CMP   WORD PTR BytesPerPixel,3
          JNE   @Not24Bit
          MOV   BYTE PTR [ES:DI],DL
          INC   DI
          JNZ   @ForeByte2
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByte2:
          MOV   BYTE PTR [ES:DI],DH
          INC   DI
          JNZ   @ForeByte3
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByte3:
          DB    _32BIT_OPERAND
          ROR   DX,16
          MOV   BYTE PTR [ES:DI],DL
          DB    _32BIT_OPERAND
          ROR   DX,16
          INC   DI
          JNZ   @ForeByteDone
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByteDone:
          SHR   AL,1
          JMP   @NoCarryFore
  @Not24Bit:
          CMP   WORD PTR BytesPerPixel,4
          JNE   @WroteFore
          DB    _32BIT_OPERAND
          MOV   WORD PTR [ES:DI],DX
  @WroteFore:
          SHR   AL,1
          ADD   DI,WORD PTR BytesPerPixel
          JNZ   @NoCarryFore

          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA

  @NoCarryFore:
          DEC   SI
          JNZ   @CharLoop

          { Copy the last pixel if necessary }

          TEST  BYTE PTR CopyLP,0FFh
          JZ    @CopyBack

          CMP   WORD PTR BytesPerPixel,2
          JNE   @Not16Bit1
          MOV   WORD PTR [ES:DI],DX
          JMP   @WroteFore1
  @Not16Bit1:
          CMP   WORD PTR BytesPerPixel,3
          JNE   @Not24Bit1
          MOV   BYTE PTR [ES:DI],DL
          INC   DI
          JNZ   @ForeByte2A
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByte2A:
          MOV   BYTE PTR [ES:DI],DH
          INC   DI
          JNZ   @ForeByte3A
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByte3A:
          DB    _32BIT_OPERAND
          ROR   DX,16
          MOV   BYTE PTR [ES:DI],DL
          DB    _32BIT_OPERAND
          ROR   DX,16
          INC   DI
          JNZ   @ForeByteDoneA
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByteDoneA:
          JMP   @NotNextPage


  @Not24Bit1:
          CMP   WORD PTR BytesPerPixel,4
          JNE   @WroteFore1
          DB    _32BIT_OPERAND
          MOV   WORD PTR [ES:DI],DX
  @WroteFore1:
          JMP   @LastPixel

          { Write background color }

  @BackDot:

          CMP   WORD PTR BytesPerPixel,2
          JNE   @Not16Bit2
          MOV   WORD PTR [ES:DI],CX
          JMP   @WroteBack
  @Not16Bit2:
          CMP   WORD PTR BytesPerPixel,3
          JNE   @Not24Bit2
          MOV   BYTE PTR [ES:DI],CL
          INC   DI
          JNZ   @BackByte2
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByte2:
          MOV   BYTE PTR [ES:DI],CH
          INC   DI
          JNZ   @BackByte3
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByte3:
          DB    _32BIT_OPERAND
          ROR   CX,16
          MOV   BYTE PTR [ES:DI],CL
          DB    _32BIT_OPERAND
          ROR   CX,16
          INC   DI
          JNZ   @BackByteDone
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByteDone:
          SHR   AL,1
          JMP   @NoCarryBack
  @Not24Bit2:
          CMP   WORD PTR BytesPerPixel,4
          JNE   @WroteBack
          DB    _32BIT_OPERAND
          MOV   WORD PTR [ES:DI],CX
  @WroteBack:
          SHR   AL,1
          ADD   DI,WORD PTR BytesPerPixel
          JNZ   @NoCarryBack

          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA

  @NoCarryBack:
          DEC   SI
          JNZ   @CharLoop
  @CopyBack:

          { Copy the last pixel if necessary }

          CMP   WORD PTR BytesPerPixel,2
          JNE   @Not16Bit3
          MOV   WORD PTR [ES:DI],CX  { Write background color }
          JMP   @WroteBack1
  @Not16Bit3:
          CMP   WORD PTR BytesPerPixel,3
          JNE   @Not24Bit3
          MOV   BYTE PTR [ES:DI],CL
          INC   DI
          JNZ   @BackByte2A
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByte2A:
          MOV   BYTE PTR [ES:DI],CH
          INC   DI
          JNZ   @BackByte3A
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByte3A:
          DB    _32BIT_OPERAND
          ROR   CX,16
          MOV   BYTE PTR [ES:DI],CL
          DB    _32BIT_OPERAND
          ROR   CX,16
          INC   DI
          JNZ   @BackByteDoneA
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByteDoneA:
          JMP   @NotNextPage
  @Not24Bit3:
          CMP   WORD PTR BytesPerPixel,4
          JNE   @WroteBack1
          DB    _32BIT_OPERAND
          MOV   WORD PTR [ES:DI],CX  { Write background color }
  @WroteBack1:

          { Move to the next pixel }

  @LastPixel:
          ADD   DI,WORD PTR BytesPerPixel
          JNZ   @NotNextPage

          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX

          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA

  @NotNextPage:
          MOV   WORD PTR AddrW,DI
        End; { Asm }
        VESAWinAddr := WinAddr;
      End
      Else
      Begin
        AddrW0 := AddrW;
        Inc(AddrW,9 * BytesPerPixel);
        If AddrW < AddrW0 Then
        Begin
          Inc(WinAddr,VInc);
          Asm
            SUB   BX,BX
            MOV   DX,WORD PTR WinAddr
            CALL  DWORD PTR PosFn
          End; { Asm }
          VESAWinAddr := WinAddr;
        End;
      End;
    End; { For Index }
    AddrW0 := AddrW;
    Inc(AddrW,AddInc);
    If AddrW < AddrW0 Then
    Begin
      Inc(WinAddr);
      Asm
        SUB   BX,BX
        MOV   DX,WORD PTR WinAddr
        CALL  DWORD PTR PosFn
      End; { Asm }
      VESAWinAddr := WinAddr;
    End;
  End; { For DY }
  TextWrite(CharBuf,X,Y,Color,S);
End; { TScreen.VESAWriteTrue }


Procedure TScreen.VESAMove(Source: Pointer; Dest,NumBytes: Word);
Var
  LX,LY    : LongInt;
  LW       : LongInt;
  Addr     : LongInt;
  WinAddr  : Word;
  DX,DY    : Word;
{  Mask     : Byte;}
  Index    : Word;
  B        : Byte;
  CB       : Byte;
  CW       : Word;
  CLR      : LongInt;
  Ch,Col   : Byte;
  Ch0,Col0 : Byte;
  AddrW    : Word;
  AddrW0   : Word;
  Fore     : LongInt;
  Back     : LongInt;
  PosFn    : Pointer;
  VInc     : Word;
  P,P1     : ^Byte;
  BytesPerPixel : LongInt;
  FR,FG,FB      : Byte;
  BR,BG,BB      : Byte;

Begin
  PosFn   := VESAModeInfo.PosFunction;

  BytesPerPixel := BPP Div 8;

  VInc    := VESAInc;

  Addr    := Dest Div 2;
  LY      := Addr Div MaxX;
  LX      := Addr - (LY * MaxX);

  LW      := Width;
  Addr    := (LY * LW * 16 + LX * 9) * BytesPerPixel;

  AddrW   := SetVESAWindow(Addr);
  WinAddr := VESAWinAddr;

  { Write the string }

  P  := Source;
  P1 := Ptr(Seg(ScreenPtr^),Ofs(ScreenPtr^) + Dest);
  For Index := 1 To NumBytes Shr 1 Do
  Begin
    AddrW   := SetVESAWindow(Addr);
    WinAddr := VESAWinAddr;

    Ch := P^;
    Inc(LongInt(P));
    Col := P^;
    Inc(LongInt(P));

    Ch0 := P1^;
    Inc(LongInt(P1));
    Col0 := P1^;
    Inc(LongInt(P1));

    If (Ch <> Ch0) Or (Col <> Col0) Or Not FastWriteChar Then
    Begin

      Case BytesPerPixel Of
         1: Begin
              Fore := Col And $F;
              Back := Col Shr 4;
            End;
         2: Begin
              FR := TextPalette[(Col And $F) * 3 + 0];
              FG := TextPalette[(Col And $F) * 3 + 1];
              FB := TextPalette[(Col And $F) * 3 + 2];
              BR := TextPalette[(Col Shr 4)  * 3 + 0];
              BG := TextPalette[(Col Shr 4)  * 3 + 1];
              BB := TextPalette[(Col Shr 4)  * 3 + 2];
              Fore := ((FR Shr (8 - VESAModeInfo.RMaskSize)) Shl VESAModeInfo.RFieldPos) +
                      ((FG Shr (8 - VESAModeInfo.GMaskSize)) Shl VESAModeInfo.GFieldPos) +
                      ((FB Shr (8 - VESAModeInfo.BMaskSize)) Shl VESAModeInfo.BFieldPos);
              Back := ((BR Shr (8 - VESAModeInfo.RMaskSize)) Shl VESAModeInfo.RFieldPos) +
                      ((BG Shr (8 - VESAModeInfo.GMaskSize)) Shl VESAModeInfo.GFieldPos) +
                      ((BB Shr (8 - VESAModeInfo.BMaskSize)) Shl VESAModeInfo.BFieldPos);
            End;
       3,4: Begin
              FR := TextPalette[(Col And $F) * 3 + 0];
              FG := TextPalette[(Col And $F) * 3 + 1];
              FB := TextPalette[(Col And $F) * 3 + 2];
              BR := TextPalette[(Col Shr 4)  * 3 + 0];
              BG := TextPalette[(Col Shr 4)  * 3 + 1];
              BB := TextPalette[(Col Shr 4)  * 3 + 2];
              TRGBQuad(Fore).R := FR;
              TRGBQuad(Fore).G := FG;
              TRGBQuad(Fore).B := FB;
              TRGBQuad(Fore).A := 0;
              TRGBQuad(Back).R := BR;
              TRGBQuad(Back).G := BG;
              TRGBQuad(Back).B := BB;
              TRGBQuad(Back).A := 0;
            End;
      End; { Case }
      For DY := 0 To 15 Do
      Begin
        B    := Font9x16[Ch * 16 + DY];
{        Mask := $80;}
        Asm
          MOV   AX,0A000h
          MOV   ES,AX
          MOV   AL,80h{BYTE PTR Mask}
          MOV   AH,BYTE PTR B
          DB    _32BIT_OPERAND
          MOV   DX,WORD PTR Fore
          DB    _32BIT_OPERAND
          MOV   CX,WORD PTR Back
          MOV   SI,WORD PTR BytesPerPixel
          MOV   DI,WORD PTR AddrW
          MOV   BX,8
  @PixelLoop:
          TEST  AH,AL
          JZ    @WriteBack
          DB    _32BIT_OPERAND
          MOV   WORD PTR CLR,DX

          { Write foreground pixel }

          CMP   SI,1
          JNE   @Not8Bit
          MOV   BYTE PTR [ES:DI],DL
          JMP   @DoneWrite
  @Not8Bit:
          CMP   SI,2
          JNE   @Not16Bit
          MOV   WORD PTR [ES:DI],DX
          JMP   @DoneWrite
  @Not16Bit:
          CMP   SI,3
          JNE   @Not24Bit
          MOV   BYTE PTR [ES:DI],DL
          INC   DI
          JNZ   @ForeByte2
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByte2:
          MOV   BYTE PTR [ES:DI],DH
          INC   DI
          JNZ   @ForeByte3
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByte3:
          DB    _32BIT_OPERAND
          ROR   DX,16
          MOV   BYTE PTR [ES:DI],DL
          DB    _32BIT_OPERAND
          ROR   DX,16
          INC   DI
          JNZ   @ForeByteDone
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @ForeByteDone:
          JMP   @NotNextPage
  @Not24Bit:
          CMP   SI,4
          JNE   @Not32Bit
          DB    _32BIT_OPERAND
          MOV   WORD PTR [ES:DI],DX
  @Not32Bit:
          JMP   @DoneWrite
  @WriteBack:
          DB    _32BIT_OPERAND
          MOV   WORD PTR CLR,CX

          { Write background pixel }

          CMP   SI,1
          JNE   @Not8BitA
          MOV   BYTE PTR [ES:DI],CL
          JMP   @DoneWrite
  @Not8BitA:
          CMP   SI,2
          JNE   @Not16BitA
          MOV   WORD PTR [ES:DI],CX
          JMP   @DoneWrite
  @Not16BitA:
          CMP   SI,3
          JNE   @Not24BitA
          MOV   BYTE PTR [ES:DI],CL
          INC   DI
          JNZ   @BackByte2
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByte2:
          MOV   BYTE PTR [ES:DI],CH
          INC   DI
          JNZ   @BackByte3
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByte3:
          DB    _32BIT_OPERAND
          ROR   CX,16
          MOV   BYTE PTR [ES:DI],CL
          DB    _32BIT_OPERAND
          ROR   CX,16
          INC   DI
          JNZ   @BackByteDone
          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @BackByteDone:
          JMP   @NotNextPage
  @Not24BitA:
          CMP   SI,4
          JNE   @Not32BitA
          DB    _32BIT_OPERAND
          MOV   WORD PTR [ES:DI],CX
  @Not32BitA:
  {        JMP   @DoneWrite}
  @DoneWrite:
          ADD   DI,SI
          JNZ   @NotNextPage

          DB    _32BIT_OPERAND
          PUSHA
          PUSH  ES
          MOV   DX,WORD PTR VInc
          ADD   WORD PTR WinAddr,DX
          SUB   BX,BX
          MOV   DX,WORD PTR WinAddr
          CALL  DWORD PTR PosFn
          POP   ES
          DB    _32BIT_OPERAND
          POPA
  @NotNextPage:
          SHR   AL,1
          DEC   BX
          JNZ   @PixelLoop
          MOV   WORD PTR AddrW,DI
        End; { Asm }
        VESAWinAddr := WinAddr;
  (*
        For DX := 0 To 7 Do
        Begin
          If (B And Mask) <> 0 Then C := Fore Else C := Back;
          Case BytesPerPixel Of
            1: Mem[$A000:AddrW]  := TRGBQuad(C).B;
            2: MemW[$A000:AddrW] := C And $FFFF;
            3: Begin
                 MemW[$A000:AddrW]     := C And $FFFF;
                 MemW[$A000:AddrW + 2] := TRGBQuad(C).R;
               End;
            4: MemL[$A000:AddrW] := C;
          End; { Case }
          Inc(AddrW,BytesPerPixel);
          If AddrW = 0 Then
          Begin
            Inc(WinAddr);
            Asm
              SUB   BX,BX
              MOV   DX,WORD PTR WinAddr
              CALL  DWORD PTR PosFn
            End; { Asm }
            VESAWinAddr := WinAddr;
          End;
          Mask := Mask Shr 1;
        End; { For DX }
  *)
        { Copy last pixel if necessary }

        If CopyLastPixel[Ch] = 0 Then CLR := Back;
        Case BytesPerPixel Of
          1: Mem[$A000:AddrW]  := TRGBQuad(CLR).B;
          2: MemW[$A000:AddrW] := CLR And $FFFF;
          3: Begin
               Mem[$A000:AddrW] := TRGBQuad(CLR).B;
               AddrW0 := AddrW;
               Inc(AddrW);
               If AddrW < AddrW0 Then
               Begin
                 Inc(WinAddr);
                 Asm
                   SUB   BX,BX
                   MOV   DX,WORD PTR WinAddr
                   CALL  DWORD PTR PosFn
                 End; { Asm }
                 VESAWinAddr := WinAddr;
               End;

               Mem[$A000:AddrW] := TRGBQuad(CLR).G;
               AddrW0 := AddrW;
               Inc(AddrW);
               If AddrW < AddrW0 Then
               Begin
                 Inc(WinAddr);
                 Asm
                   SUB   BX,BX
                   MOV   DX,WORD PTR WinAddr
                   CALL  DWORD PTR PosFn
                 End; { Asm }
                 VESAWinAddr := WinAddr;
               End;

               Mem[$A000:AddrW] := TRGBQuad(CLR).R;
               AddrW0 := AddrW;
               Inc(AddrW);
               If AddrW < AddrW0 Then
               Begin
                 Inc(WinAddr);
                 Asm
                   SUB   BX,BX
                   MOV   DX,WORD PTR WinAddr
                   CALL  DWORD PTR PosFn
                 End; { Asm }
                 VESAWinAddr := WinAddr;
               End;

  {             MemW[$A000:AddrW]     := CL And $FFFF;
               MemW[$A000:AddrW + 2] := TRGBQuad(CL).R;}
             End;
          4: MemL[$A000:AddrW] := CLR;
        End; { Case }
        If BytesPerPixel <> 3 Then
        Begin
          AddrW0 := AddrW;
          Inc(AddrW,(Width - 8) * BytesPerPixel);
          If AddrW < AddrW0 Then
          Begin
            Inc(WinAddr);
            Asm
              SUB   BX,BX
              MOV   DX,WORD PTR WinAddr
              CALL  DWORD PTR PosFn
            End; { Asm }
            VESAWinAddr := WinAddr;
          End;
        End
        Else
        Begin
          AddrW0 := AddrW;
          Inc(AddrW,(Width - 9) * BytesPerPixel);
          If AddrW < AddrW0 Then
          Begin
            Inc(WinAddr);
            Asm
              SUB   BX,BX
              MOV   DX,WORD PTR WinAddr
              CALL  DWORD PTR PosFn
            End; { Asm }
            VESAWinAddr := WinAddr;
          End;
        End;
      End; { For DY }
    End;
    Inc(Addr,9 * BytesPerPixel);
  End; { For Index }
  Move(Source^,Mem[Seg(ScreenPtr^):Ofs(ScreenPtr^) + Dest],NumBytes);
End; { TScreen.VESAMove }


Procedure TScreen.VESAShowCursor;
Var
  CStart        : Integer;
  CEnd          : Integer;
  P             : ^Byte;
  BytesPerPixel : LongInt;
  FR,FG,FB      : Byte;
  BR,BG,BB      : Byte;
  AddrW         : Word;
  AddrW0        : Word;
  Fore          : LongInt;
  Back          : LongInt;
  PosFn         : Pointer;
  LX,LY         : LongInt;
  LW            : LongInt;
  Addr          : LongInt;
  WinAddr       : Word;
  DX,DY         : Word;
  VInc          : Word;
  Ch,Ch1        : Byte;
  B             : Byte;

Begin

  BytesPerPixel := BPP Div 8;

  PosFn   := VESAModeInfo.PosFunction;

  CStart := CursorStart;
  CEnd   := CursorEnd;
  Case CStart Of
    0: CStart := 0;
    6: CStart := 13;
    7: CStart := 14;
  Else
    CStart := CStart * 2;
  End; { Case }
  Case CEnd Of
    0: CEnd := 0;
    6: CEnd := 13;
    7: CEnd := 14;
  Else
    CEnd := CEnd * 2;
  End; { Case }

  LX      := Lo(WindMin) + WhereX - 1;
  LY      := Hi(WindMin) + WhereY - 1;
  LW      := Width;
  Addr    := ((LY * 16 + CStart) * LW + LX * 9) * BytesPerPixel;
  VInc    := VESAInc;

  P       := Ptr(Seg(CharBuf^),Ofs(CharBuf^) + ((LY * MaxX) + LX) * 2);
  Ch      := P^;
  Inc(LongInt(P));
  Ch1     := P^;

  AddrW   := SetVESAWindow(Addr);
  WinAddr := VESAWinAddr;

  { Show the cursor }

  FR := TextPalette[(Ch1 And $F) * 3 + 0];
  FG := TextPalette[(Ch1 And $F) * 3 + 1];
  FB := TextPalette[(Ch1 And $F) * 3 + 2];
  BR := TextPalette[(Ch1 Shr 4)  * 3 + 0];
  BG := TextPalette[(Ch1 Shr 4)  * 3 + 1];
  BB := TextPalette[(Ch1 Shr 4)  * 3 + 2];
  Case BPP Of
     8: Begin
          Fore := Ch1 And $F;
          Back := Ch1 Shr 4;
        End;
    16: Begin
      Fore := ((FR Shr (8 - VESAModeInfo.RMaskSize)) Shl VESAModeInfo.RFieldPos) +
              ((FG Shr (8 - VESAModeInfo.GMaskSize)) Shl VESAModeInfo.GFieldPos) +
              ((FB Shr (8 - VESAModeInfo.BMaskSize)) Shl VESAModeInfo.BFieldPos);
      Back := ((BR Shr (8 - VESAModeInfo.RMaskSize)) Shl VESAModeInfo.RFieldPos) +
              ((BG Shr (8 - VESAModeInfo.GMaskSize)) Shl VESAModeInfo.GFieldPos) +
              ((BB Shr (8 - VESAModeInfo.BMaskSize)) Shl VESAModeInfo.BFieldPos);
        End;
 24,32: Begin
          TRGBQuad(Fore).R := FR;
          TRGBQuad(Fore).G := FG;
          TRGBQuad(Fore).B := FB;
          TRGBQuad(Fore).A := 0;
          TRGBQuad(Back).R := BR;
          TRGBQuad(Back).G := BG;
          TRGBQuad(Back).B := BB;
          TRGBQuad(Back).A := 0;
        End;
  End; { Case }

  For DY := CStart To CEnd Do
  Begin
    B := Font9x16[Ch * 16 + DY];
    Asm
      MOV   AX,0A000h
      MOV   ES,AX
      MOV   AL,80h{BYTE PTR Mask}
      MOV   AH,BYTE PTR B
      DB    _32BIT_OPERAND
      MOV   DX,WORD PTR Fore
      DB    _32BIT_OPERAND
      MOV   CX,WORD PTR Back
      MOV   SI,WORD PTR BytesPerPixel
      MOV   DI,WORD PTR AddrW
      MOV   BX,8
@PixelLoop:
      TEST  AH,AL
      JNZ   @WriteBack   { Note the difference! }

      { Write foreground pixel }

      CMP   SI,1
      JNE   @Not8Bit
      MOV   BYTE PTR [ES:DI],DL
      JMP   @DoneWrite
@Not8Bit:
      CMP   SI,2
      JNE   @Not16Bit
      MOV   WORD PTR [ES:DI],DX
      JMP   @DoneWrite
@Not16Bit:
      CMP   SI,3
      JNE   @Not24Bit
      MOV   BYTE PTR [ES:DI],DL
      INC   DI
      JNZ   @ForeByte2
      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@ForeByte2:
      MOV   BYTE PTR [ES:DI],DH
      INC   DI
      JNZ   @ForeByte3
      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@ForeByte3:
      DB    _32BIT_OPERAND
      ROR   DX,16
      MOV   BYTE PTR [ES:DI],DL
      DB    _32BIT_OPERAND
      ROR   DX,16
      INC   DI
      JNZ   @ForeByteDone
      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@ForeByteDone:
      JMP   @NotNextPage
@Not24Bit:
      CMP   SI,4
      JNE   @Not32Bit
      DB    _32BIT_OPERAND
      MOV   WORD PTR [ES:DI],DX
@Not32Bit:
      JMP   @DoneWrite
@WriteBack:
      { Write background pixel }

      CMP   SI,1
      JNE   @Not8BitA
      MOV   BYTE PTR [ES:DI],CL
      JMP   @DoneWrite
@Not8BitA:
      CMP   SI,2
      JNE   @Not16BitA
      MOV   WORD PTR [ES:DI],CX
      JMP   @DoneWrite
@Not16BitA:
      CMP   SI,3
      JNE   @Not24BitA
      MOV   BYTE PTR [ES:DI],CL
      INC   DI
      JNZ   @BackByte2
      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@BackByte2:
      MOV   BYTE PTR [ES:DI],CH
      INC   DI
      JNZ   @BackByte3
      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@BackByte3:
      DB    _32BIT_OPERAND
      ROR   CX,16
      MOV   BYTE PTR [ES:DI],CL
      DB    _32BIT_OPERAND
      ROR   CX,16
      INC   DI
      JNZ   @BackByteDone
      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@BackByteDone:
      JMP   @NotNextPage
@Not24BitA:
      CMP   SI,4
      JNE   @Not32BitA
      DB    _32BIT_OPERAND
      MOV   WORD PTR [ES:DI],CX
@Not32BitA:
{        JMP   @DoneWrite}
@DoneWrite:
      ADD   DI,SI
      JNZ   @NotNextPage

      DB    _32BIT_OPERAND
      PUSHA
      PUSH  ES
      MOV   DX,WORD PTR VInc
      ADD   WORD PTR WinAddr,DX
      SUB   BX,BX
      MOV   DX,WORD PTR WinAddr
      CALL  DWORD PTR PosFn
      POP   ES
      DB    _32BIT_OPERAND
      POPA
@NotNextPage:
      SHR   AL,1
      DEC   BX
      JNZ   @PixelLoop
      MOV   WORD PTR AddrW,DI
    End; { Asm }
    AddrW0 := AddrW;
    Inc(AddrW,(Width - 8) * BytesPerPixel);
    If AddrW < AddrW0 Then
    Begin
      Inc(WinAddr);
      Asm
        SUB   BX,BX
        MOV   DX,WORD PTR WinAddr
        CALL  DWORD PTR PosFn
      End; { Asm }
      VESAWinAddr := WinAddr;
    End;
  End; { For DY }
End; { TScreen.VESAShowCursor }


Procedure TScreen.VESAHideCursor;
Var
  Addr : Word;
  Old  : Boolean;

Begin
  Old := FastWriteChar;
  FastWriteChar := False;         { Force a write }
  Addr := ((Hi(WindMin) + WhereY - 1) * MaxX + Lo(WindMin) + WhereX - 1) * 2;
  VESAMove(Ptr(Seg(CharBuf^),Ofs(CharBuf^) + Addr),Addr,2);
  FastWriteChar := Old;
End; { TScreen.VESAHideCursor }


Procedure TScreen.VESAShade(X,Y: Integer; Len: Byte);
Var
  Addr : Word;
  Old  : Boolean;

Begin
  Old := FastWriteChar;
  FastWriteChar := False;         { Force a write }
  TextShade(CharBuf,X,Y,Len);
  Addr := ((Y - 1) * MaxX + X - 1) * 2;
  VESAMove(Ptr(Seg(CharBuf^),Ofs(CharBuf^) + Addr),Addr,Len * 2);
  FastWriteChar := Old;
End; { TScreen.VESAShade }


Procedure TScreen.FastWrite(X,Y: Integer; Color: Byte; S: String);
Begin
  If S <> '' Then
  Begin
    Case DisplayType Of
    dtText,dtVESAText: TextWrite(Ptr($B800,0),X,Y,Color,S);
            dtVESA256: VESAWrite256(X,Y,Color,S);
           dtVESATrue: VESAWriteTrue(X,Y,Color,S);
    End; { Case }
  End;
End; { TScreen.FastWrite }


Procedure TScreen.FastShade(X,Y: Integer; Len: Byte);
Begin
  If Len > 0 Then
  Begin
    Case DisplayType Of
    dtText,dtVESAText: TextShade(Ptr($B800,0),X,Y,Len);
 dtVESA256,dtVESATrue: VESAShade(X,Y,Len);
    End; { Case }
  End;
End; { TScreen.FastShade }


Function TScreen.GetNumModes: Integer;
Type WPtr = ^Word;
Var
  P          : WPtr;
  Mode       : Word;
  ScreenMode : PScreenMode;

Begin
  NumModes := 1;  { Normal text mode }
  ScreenMode := New(PScreenMode,Init);
  ScreenMode^.DisplayType := dtText;
  ScreenMode^.Width       := MaxX;
  ScreenMode^.Height      := MaxY;
  ScreenMode^.ModeNumber  := 3;
  ScreenMode^.BPP         := 8;
  ScreenModes^.Insert(ScreenMode);

  { If the initial mode is 80x25, we can likely set 80x30 }

  If MaxY = 25 Then
  Begin
    Inc(NumModes);
    ScreenMode := New(PScreenMode,Init);
    ScreenMode^.DisplayType := dtText;
    ScreenMode^.Width       := MaxX;
    ScreenMode^.Height      := 30;
    ScreenMode^.ModeNumber  := 3;
    ScreenMode^.BPP         := 8;
    ScreenModes^.Insert(ScreenMode);
  End;

  { Get VESA modes }

  If GetSuperVGAInfo(VESAInfo) Then
  Begin
    P := VESAInfo.ModeListPtr;
    While P^ <> $FFFF Do
    Begin
      Mode := P^;
{
      System.WriteLn(HexWord(Mode));
      ReadKey;
}
      If GetSuperVGAModeInfo(Mode,VESAModeInfo) Then
      Begin
{
        If VESAModeInfo.Width = 800 Then
        Begin
          System.WriteLn('ModeAttr          = ',HexWord(VESAModeInfo.ModeAttr));
          System.WriteLn('Width             = ',VESAModeInfo.Width);
          System.WriteLn('Height            = ',VESAModeInfo.Height);
          System.WriteLn('MemModelType      = ',VESAModeInfo.MemModelType);
          System.WriteLn('WindowGranularity = ',VESAModeInfo.WindowGranularity);
          System.WriteLn('BytesPerLine      = ',VESAModeInfo.BytesPerLine);
          System.WriteLn('PosFunction       = ',HexLongInt(LongInt(VESAModeInfo.PosFunction)));
          ReadKey;
        End;
}
        If ((VESAModeInfo.ModeAttr And $B) = $B) And
           ((((VESAModeInfo.ModeAttr And $10) <> 0) And
             (VESAModeInfo.PosFunction <> Nil)) Or
            ((VESAModeInfo.ModeAttr And $10) = 0)) Then
{           (VESAModeInfo.PosFunction <> Nil)     Then}
        Begin
          If ((VESAModeInfo.MemModelType = 0) And
              (VESAModeInfo.Width >= 80))     Or
             ((VESAModeInfo.MemModelType = 4) And
              (VESAModeInfo.Width >= 720))    Or
             ((VESAModeInfo.MemModelType = 6) And
              (VESAModeInfo.Width >= 720))    Then


{          If (VESAModeInfo.MemModelType In [0,4]) And
             (VESAModeInfo.Width >= 720)          Then}
          Begin
            ScreenMode := New(PScreenMode,Init);
            Case VESAModeInfo.MemModelType Of
              0: ScreenMode^.DisplayType := dtVESAText;
              4: ScreenMode^.DisplayType := dtVESA256;
              6: ScreenMode^.DisplayType := dtVESATrue;
            End; { Case }
            Case ScreenMode^.DisplayType Of
  dtVESAText: Begin

                { The VESA standard says that VESAModeInfo.Width and }
                { VESAModeInfo.Height should be reported in pixels,  }
                { but the 3dfx voodoo3 3500 TV incorrectly reports   }
                { them in characters.  Therefore we have to try to   }
                { detect both schemes.                               }

                If VESAModeInfo.Width >= 300 Then
                Begin
                  ScreenMode^.Width  := VESAModeInfo.Width  Div VESAModeInfo.CellWidth;
                  ScreenMode^.Height := VESAModeInfo.Height Div VESAModeInfo.CellHeight;
                End
                Else
                Begin
                  ScreenMode^.Width  := VESAModeInfo.Width;
                  ScreenMode^.Height := VESAModeInfo.Height;
                End;
                ScreenMode^.BPP := 8;
              End;
   dtVESA256,
  dtVESATrue: Begin
                ScreenMode^.Width  := VESAModeInfo.Width;
                ScreenMode^.Height := VESAModeInfo.Height;
                ScreenMode^.BPP    := Trunc(8 * (VESAModeInfo.BytesPerLine / VESAModeInfo.Width));
              End;
            End; { Case }

            ScreenMode^.ModeNumber := Mode;
            ScreenModes^.Insert(ScreenMode);
            Inc(NumModes);
          End;
        End;
      End;
      Inc(LongInt(P),2);
    End; { While }
  End;
End; { TScreen.GetNumModes }

Procedure TScreen.SetMode(Mode: Integer);
Const
  VGA_SEGMENT   = $A000;
  SC_INDEX      = $3C4;   { Sequence Controller Index register   }
  GC_INDEX      = $3CE;   { Graphics Controller Index register   }
  CRTC_INDEX    = $3D4;   { CRT Controller Index register        }
  MISC_OUTPUT   = $3C2;   { Miscellaneous Output Register        }
  MISCELLANEOUS = 6;      { Moscellaneous register index in GC   }
  MEMORY_MODE   = 4;      { Memory Mode register index in SC     }
  GRAPHICS_MODE = 5;      { Graphics Mode register index in GC   }
  MAX_SCAN_LINE = 9;      { Maximum Scan Line reg index in CRTC  }
  UNDERLINE     = $14;    { Underline Location reg index in CRTC }
  MODE_CONTROL  = $17;    { Mode Control register index in CRTC  }
  MAP_MASK      = 2;      { Map Mask register index in SC        }

  HORZ_TOTAL         = 0;
  HORZ_DISPLAY_END   = 1;
  START_HORZ_BLANK   = 2;
  END_HORZ_BLANK     = 3;
  START_HORZ_RETRACE = 4;
  END_HORZ_RETRACE   = 5;
  VERT_TOTAL         = 6;
  OVERFLOW           = 7;
  START_VERT_RETRACE = $10;
  END_VERT_RETRACE   = $11;
  VERT_DISPLAY_END   = $12;
  CRTC_OFFSET        = $13;
  START_VERT_BLANK   = $15;
  END_VERT_BLANK     = $16;
  LINE_COMPARE       = $18;

Var
  Sel : PScreenMode;
  I   : Integer;

  Procedure UnprotectCRTC; Assembler;
  Asm

    { Un-protect the CRTC registers }

    MOV   DX,CRTC_INDEX
    MOV   AL,END_VERT_RETRACE
    OUT   DX,AL
    INC   DX
    IN    AL,DX
    AND   AL,7Fh
    OUT   DX,AL
    DEC   DX
    MOV   AL,END_HORZ_BLANK
    OUT   DX,AL
    INC   DX
    IN    AL,DX
    OR    AL,80h
    OUT   DX,AL
  End; { UnprotectCRTC }


  Procedure ProtectCRTC; Assembler;
  Asm

    { Re-protect the CRTC registers }

    MOV   DX,CRTC_INDEX
    MOV   AL,END_HORZ_BLANK
    OUT   DX,AL
    INC   DX
    IN    AL,DX
    AND   AL,7Fh
    OUT   DX,AL
    DEC   DX
    MOV   AL,END_VERT_RETRACE
    OUT   DX,AL
    INC   DX
    IN    AL,DX
    OR    AL,80h
    OUT   DX,AL
  End; { ProtectCRTC }

  Procedure SetRows(Rows: Word; VPositive,HPositive: Boolean; Crystal: Byte);
  Var
    VT,VRS,VRE,VDE,VBS,VBE,LC                 : Word;
    _VT,_OV,_MSL,_VRS,_VRE,_VDE,_VBS,_VBE,_LC : Word;

  Begin

    { Important to initialize these since they are words }

    _VT  := 0;
    _OV  := 0;
    _MSL := 0;
    _VRS := 0;
    _VRE := 0;
    _VDE := 0;
    _VBS := 0;
    _VBE := 0;
    _LC  := 0;

    UnprotectCRTC;

    { Get the current register values }

    Asm
      MOV   DX,CRTC_INDEX

      MOV   AL,VERT_TOTAL
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _VT,AL

      MOV   AL,OVERFLOW
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _OV,AL

      MOV   AL,MAX_SCAN_LINE
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _MSL,AL

      MOV   AL,START_VERT_RETRACE
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _VRS,AL

      MOV   AL,END_VERT_RETRACE
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _VRE,AL

      MOV   AL,VERT_DISPLAY_END
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _VDE,AL

      MOV   AL,START_VERT_BLANK
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _VBS,AL

      MOV   AL,END_VERT_BLANK
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _VBE,AL

      MOV   AL,LINE_COMPARE
      OUT   DX,AL
      INC   DX
      IN    AL,DX
      DEC   DX
      MOV   BYTE PTR _LC,AL
    End; { Asm }

    { Retrieve the full values }

    VT  := ((_OV And $20) Shr 4) + (_OV And 1);
    VT  := (VT Shl 8) + _VT;

    VRS := ((_OV And $80) Shr 6) + ((_OV And 4) Shr 2);
    VRS := (VRS Shl 8) + _VRS;

    VRE := (VRS And $FFF0) + (_VRE And $F);
    If VRE < VRS Then Inc(VRE,$10);

    VDE := ((_OV And $40) Shr 5) + ((_OV And 2) Shr 1);
    VDE := (VDE Shl 8) + _VDE;

    VBS := ((_MSL And $20) Shr 4) + ((_OV And 8) Shr 3);
    VBS := (VBS Shl 8) + _VBS;

    VBE := (VBS And $FF00) + (_VBE And $FF);{ Not really standard VGA... }
    While VBE < VBS Do Inc(VBE,$80);        { Correct if necessary       }

    LC  := ((_MSL And $40) Shr 5) + ((_OV And $10) Shr 4);
    LC  := (LC Shl 8) + _LC;

    { If the 200-to-400 bit is set on the Max Scan Line register, multiply }
    { the number of rows by 2                                              }

    If (_MSL And $80) <> 0 Then Rows := Rows * 2;

    { Only do something if the rows actually changed }

    If Rows <> VDE + 1 Then
    Begin

      { Recalculate the values }

      Inc(VT ,Rows - (VDE + 1));
      Inc(VRS,Rows - (VDE + 1));
      Inc(VRE,Rows - (VDE + 1));
      Inc(VBS,Rows - (VDE + 1));
      Inc(VBE,Rows - (VDE + 1));
      Inc(LC ,Rows - (VDE + 1));
      Inc(VDE,Rows - (VDE + 1)); { Must do last }

      { Bounds-checking }

      If VT  > 1023 Then VT  := 1023;
      If VRS > 1023 Then VRS := 1023;
      If VRE > 1023 Then VRE := 1023;
      If VDE > 1023 Then VDE := 1023;
      If VBS > 1023 Then VBS := 1023;
      If VBE > 1023 Then VBE := 1023;
      If LC  > 1023 Then LC  := 1023;

      { Break the new values into the registers }

      _VT  := VT  And $FF;
      _MSL := (_MSL And $9F) + ((LC Shr 3) And $40) + ((VBS Shr 4) And $20);
      _VRS := VRS And $FF;
      _VRE := (_VRE And $F0) + (VRE And $F);
      _VDE := VDE And $FF;
      _VBS := VBS And $FF;
      _VBE := VBE And $FF;           { The VGA standard says that VBE is  }
      _LC  := LC  And $FF;           { only seven bits, but my ATI 3D Pro }
      _OV  := ((VRS Shr 2) And $80) +{ Turbo doesn't handle this properly }
              ((VDE Shr 3) And $40) +{ unless I fill in all 8 bits.  I    }
              ((VT  Shr 4) And $20) +{ figure that cards that follow the  }
              ((LC  Shr 4) And $10) +{ standard properly shouldn't be     }
              ((VBS Shr 5) And $08) +{ bothered by the extra bit...       }
              ((VRS Shr 6) And $04) +
              ((VDE Shr 7) And $02) +
              ((VT  Shr 8) And $01);

      { Change to new rows by changing the vertical registers }

      Asm
        MOV   DX,CRTC_INDEX

        MOV   AL,VERT_TOTAL
        MOV   AH,BYTE PTR _VT
        OUT   DX,AX

        MOV   AL,OVERFLOW
        MOV   AH,BYTE PTR _OV
        OUT   DX,AX

        MOV   AL,MAX_SCAN_LINE
        MOV   AH,BYTE PTR _MSL
        OUT   DX,AX

        MOV   AL,START_VERT_RETRACE
        MOV   AH,BYTE PTR _VRS
        OUT   DX,AX

        MOV   AL,END_VERT_RETRACE
        MOV   AH,BYTE PTR _VRE
        OUT   DX,AX

        MOV   AL,VERT_DISPLAY_END
        MOV   AH,BYTE PTR _VDE
        OUT   DX,AX

        MOV   AL,START_VERT_BLANK
        MOV   AH,BYTE PTR _VBS
        OUT   DX,AX

        MOV   AL,END_VERT_BLANK
        MOV   AH,BYTE PTR _VBE
        OUT   DX,AX

        MOV   AL,LINE_COMPARE
        MOV   AH,BYTE PTR _LC
        OUT   DX,AX
      End; { Asm }

      { Set the vertical and horizontal sync polarity and crystal }

      Asm
        MOV   DX,03CCh
        IN    AL,DX
        AND   AL,33h
        TEST  BYTE PTR VPositive,0FFh
        JNZ   @VPos
        OR    AL,80h
@VPos:
        TEST  BYTE PTR HPositive,0FFh
        JNZ   @HPos
        OR    AL,40h
@HPos:
        MOV   BL,BYTE PTR Crystal
        AND   BL,3
        SHL   BL,2
        OR    AL,BL
        MOV   DX,03C2h
        OUT   DX,AL
      End; { Asm }

    End;
    ProtectCRTC;
  End; { SetRows }

Begin
  If (Mode >= 0) And (Mode < ScreenModes^.Count) Then
  Begin

    { Deallocate the character info array (only used for graphics modes) }

    If (CharBuf <> Nil) And (CharBufSize <> 0) Then
    Begin
      FreeMem(CharBuf,CharBufSize);
      CharBuf     := Nil;
      CharBufSize := 0;
    End;

    { Set the mode }

    Sel := ScreenModes^.At(Mode);
    If Sel^.DisplayType = dtText Then
    Begin
      TextMode(C80);
      If Sel^.Height = 30 Then SetRows(Sel^.Height * 16,False,False,1);
    End
    Else
    Begin
      SetSuperVGAMode(Sel^.ModeNumber);
      GetSuperVGAModeInfo(Sel^.ModeNumber,VESAModeInfo);
      VESAWinAddr := 0;

      { Set the palette if necessary }

      If Sel^.DisplayType = dtVESA256 Then
      Begin
        Port[$3C8] := 0;
        For I := 0 To 16 * 3 - 1 Do Port[$3C9] := TextPalette[I] Shr 2;
      End;
    End;
    DisplayType := Sel^.DisplayType;
    Width       := Sel^.Width;
    Height      := Sel^.Height;
    BPP         := Sel^.BPP;
    Case DisplayType Of
      dtText,dtVESAText: Begin
                           MaxX := Width;
                           MaxY := Height;
                           ScreenPtr := Ptr($B800,0);
                           Mem[$40:$4A] := MaxX;
                           Mem[$40:$84] := MaxY - 1;
                           Crt.WindMin := 0;
                           Crt.WindMax := (MaxX - 1) + ((MaxY - 1) Shl 8);
                           WhereX  := Crt.WhereX;
                           WhereY  := Crt.WhereY;
                           WindMin := Crt.WindMin;
                           WindMax := Crt.WindMax;

                           { Disable blinking/enable background intensity }

                           Asm
                             CLI
                             MOV   DX,03DAh  { Reset flip-flop }
                             IN    AL,DX

                             MOV   DX,03C0h
                             MOV   AL,10h
                             OUT   DX,AL
                             INC   DX
                             IN    AL,DX
                             MOV   AH,AL
                             AND   AH,0F7h

                             MOV   DX,03DAh  { Reset flip-flop }
                             IN    AL,DX

                             MOV   DX,03C0h
                             MOV   AL,10h
                             OUT   DX,AL
                             MOV   AL,AH
                             OUT   DX,AL

                             MOV   DX,03DAh  { Reset flip-flop }
                             IN    AL,DX

                             MOV   DX,03C0h
                             MOV   AL,30h
                             OUT   DX,AL
                             STI
                           End; { Asm }

                         End;
              dtVESA256,
             dtVESATrue: Begin
                           MaxX        := Width Div 9;
                           MaxY        := Height Div 16;
                           CharBufSize := MaxX * MaxY * 2;
                           GetMem(CharBuf,CharBufSize);
                           FillChar(CharBuf^,CharBufSize,#0);
                           ScreenPtr := CharBuf;
                           WindMin := 0;
                           WindMax := (MaxX - 1) + ((MaxY - 1) Shl 8);
                           VESAInc := 64 Div VESAModeInfo.WindowGranularity;
                         End;
    End; { Case }
    CurrentMode := Mode;
    SetCursor(6,7);
    ShowCursor(True);
  End;
End; { TScreen.SetMode }


Procedure TScreen.ShowCursor(Show: Boolean);
Var Regs: Registers;
Begin
  Case DisplayType Of
    dtText,dtVESAText:
    Begin
      Regs.AH := 1;
      Regs.AL := 3;{Mem[Seg0040:$49];}
      Regs.CH := CursorStart;
      Regs.CL := CursorEnd;
      If Not Show Then Regs.CH := $20 + Regs.CH;
      Intr($10,Regs);
    End;
    dtVESA256,dtVESATrue:
    Begin
      If Show Then VESAShowCursor;
      If (Not Show) And CursorOn Then VESAHideCursor;
    End;
  End; { Case }
  CursorOn := Show;
End; { TScreen.ShowCursor }


Procedure TScreen.SetCursor(MinC,MaxC: Byte);
Begin
  CursorStart := MinC;
  CursorEnd   := MaxC;
  ShowCursor(CursorOn);
End; { TScreen.SetCursor }


Procedure TScreen.ClrScr;
Var
  St : String;
  I  : Integer;

Begin
  Case DisplayType Of
    dtText,dtVESAText: Begin
                         Crt.ClrScr;
                         WhereX := Crt.WhereX;
                         WhereY := Crt.WhereY;
                       End;
 dtVESA256,dtVESATrue: Begin
                         St := '';
                         While Length(St) < Lo(WindMax) - Lo(WindMin) + 1 Do St := St + ' ';
                         For I := Hi(WindMin) + 1 To Hi(WindMax) + 1 Do
                          FastWrite(Lo(WindMin) + 1,I,TextAttr,St);
                         ShowCursor(CursorOn);
                       End;
  End; { Case }
End; { TScreen.ClrScr }


Procedure TScreen.ClrEol;
Var St: String;
Begin
  Case DisplayType Of
    dtText,dtVESAText: Crt.ClrEol;
 dtVESA256,dtVESATrue: Begin
                         St := '';
                         While Length(St) < MaxX - WhereX + 1 Do St := St + ' ';
                         FastWrite(WhereX + Lo(WindMin),WhereY + Hi(WindMin),TextAttr,St);
                         ShowCursor(CursorOn);
                       End;
  End; { Case }
End; { TScreen.ClrScr }


Procedure TScreen.TextColor(Color: Byte);
Begin
  Crt.TextColor(Color);
  TextAttr := Crt.TextAttr;
  Case DisplayType Of
    dtVESA256,dtVESATrue: TextAttr := (TextAttr And $F0) + (Color And $F);
  End; { Case }
End; { TScreen.TextColor }


Procedure TScreen.TextBackground(Color: Byte);
Begin
  If Color > 7
   Then Crt.TextColor((Crt.TextAttr And $F) + Blink)
   Else Crt.TextColor(Crt.TextAttr And $F);
  Crt.TextBackground(Color And 7);
  TextAttr := Crt.TextAttr;
  Case DisplayType Of
    dtVESA256,dtVESATrue: TextAttr := (TextAttr And $F) + (Color Shl 4);
  End; { Case }
End; { TScreen.TextBackground }


Procedure TScreen.SetTextAttr(Attr: Byte);
Begin
  Crt.TextAttr := Attr;
  Case DisplayType Of
    dtVESA256,dtVESATrue: TextAttr := Attr;
  End; { Case }
End; { TScreen.SetTextAttr }


Procedure TScreen.GotoXY(X,Y: Byte);
Begin
  Crt.GotoXY(X,Y);
  Case DisplayType Of
    dtText,dtVESAText: Begin
                         WhereX := Crt.WhereX;
                         WhereY := Crt.WhereY;
                       End;
 dtVESA256,dtVESATrue: Begin
                         VESAHideCursor;
                         WhereX := X;
                         WhereY := Y;
                         ShowCursor(CursorOn);
                       End;
  End; { Case }
End; { TScreen.GotoXY }

Procedure TScreen.VESAScrollWindowUp;
Var I,J: Integer;
Begin
  VESAHideCursor;
  J := Lo(WindMax) - Lo(WindMin) + 1;
  For I := Hi(WindMin) To Hi(WindMax) - 1 Do
  Begin
    VESAMove(Ptr(Seg(CharBuf^),Ofs(CharBuf^) + ((I + 1) * MaxX + Lo(WindMin)) * 2),
     (I * MaxX + Lo(WindMin)) * 2,J * 2);
  End; { For I }
  WhereY := Hi(WindMax) - Hi(WindMin) + 1;
  ClrEol;
End; { TScreen.VESAScrollWindowUp }

Procedure TScreen.Write(St: String);
Var Cursor: Boolean;
Begin
  Case DisplayType Of
    dtText,dtVESAText: Begin
                         System.Write(St);
                         WhereX := Crt.WhereX;
                         WhereY := Crt.WhereY;
                       End;
 dtVESA256,dtVESATrue: Begin
                         Cursor := CursorOn;
                         ShowCursor(False);
                         If St <> '' Then
                          FastWrite(WhereX + Lo(WindMin),WhereY + Hi(WindMin),TextAttr,St);
                         Inc(WhereX,Length(St));
                         While WhereX > (Lo(WindMax) - Lo(WindMin) + 1) Do
                         Begin
                           Dec(WhereX,Lo(WindMax) - Lo(WindMin) + 1);
                           Inc(WhereY);
                         End; { While }
                         If WhereY > (Hi(WindMax) - Hi(WindMin) + 1) Then
                          VESAScrollWindowUp;
                         ShowCursor(Cursor);
                       End;
  End; { Case }
End; { TScreen.Write }

Procedure TScreen.WriteLn(St: String);
Var
  I,J    : Integer;
  Cursor : Boolean;

Begin
  Case DisplayType Of
    dtText,dtVESAText: Begin
                         System.WriteLn(St);
                         WhereX := Crt.WhereX;
                         WhereY := Crt.WhereY;
                       End;
 dtVESA256,dtVESATrue: Begin
                         Cursor := CursorOn;
                         ShowCursor(False);
                         If St <> '' Then
                          FastWrite(WhereX + Lo(WindMin),WhereY + Hi(WindMin),TextAttr,St);
                         WhereX := 1;
                         Inc(WhereY);
                         If WhereY > (Hi(WindMax) - Hi(WindMin) + 1) Then
                          VESAScrollWindowUp;
                         ShowCursor(Cursor);
                       End;
  End; { Case }
End; { TScreen.WriteLn }

Procedure TScreen.FastMove(Source: Pointer; Addr,NumBytes: Word);
Begin
  Case DisplayType Of
    dtText,dtVESAText: Move(Source^,Mem[$B800:Addr],NumBytes);
 dtVESA256,dtVESATrue: Begin
                         VESAHideCursor;
                         VESAMove(Source,Addr,NumBytes);
                         ShowCursor(CursorOn);
                       End;
  End; { Case }
End; { TScreen.FastMove }

Procedure TScreen.Window(X1,Y1,X2,Y2: Byte);
Begin
  Case DisplayType Of
    dtText,dtVESAText: Begin
                         Crt.WindMin := (X1 - 1) + ((Y1 - 1) Shl 8);
                         Crt.WindMax := (X2 - 1) + ((Y2 - 1) Shl 8);
                         WindMin     := Crt.WindMin;
                         WindMax     := Crt.WindMax;
                         WhereX      := Crt.WhereX;
                         WhereY      := Crt.WhereY;
                       End;
 dtVESA256,dtVESATrue: Begin
                         Dec(Y1);
                         Dec(Y2);
                         WindMin := (X1 - 1) + Word(Y1) Shl 8;
                         WindMax := (X2 - 1) + Word(Y2) Shl 8;
                         WhereX  := 1;
                         WhereY  := 1;
                       End;
  End; { Case }
End; { TScreen.Window }

{ ------------------------------ }

Procedure WriteAt(X,Y: Integer; Fore,Back: Byte; _St: String);
Begin
  If _St = '' Then Exit;
  Inc(X,Lo(Screen.WindMin));
  Inc(Y,Hi(Screen.WindMin));
  Screen.FastWrite(X,Y,Fore + (Back Shl 4),_St);
End; { WriteAt }


Procedure WriteAtAttr(X,Y: Integer; Attr: Byte; _St: String);
Begin
  If _St = '' Then Exit;
  Inc(X,Lo(Screen.WindMin));
  Inc(Y,Hi(Screen.WindMin));
  Screen.FastWrite(X,Y,Attr,_St);
End; { WriteAtAttr }

{------------------------------- }
{ DSelection                     }
{------------------------------- }

Constructor DSelection.Init(_X,_Y,_Width: Integer;
                            Const _SelStr: ScreenStr;
                            _HotChar: Byte;
                            _Protected: Boolean;
                            _Checked: Boolean;
                            _WindowAttr: Byte;
                            _HotCharAttr: Byte;
                            _SelAttr: Byte;
                            _SelHotCharAttr: Byte;
                            _ProtectedAttr: Byte;
                            _SelProtAttr: Byte);
Begin
  Inherited Init;
  X              := _X;
  Y              := _Y;
  Width          := _Width;
  SelStr         := _SelStr;
  HotChar        := _HotChar;
  Protected      := _Protected;
  Checked        := _Checked;
  WindowAttr     := _WindowAttr;
  HotCharAttr    := _HotCharAttr;
  SelAttr        := _SelAttr;
  SelHotCharAttr := _SelHotCharAttr;
  ProtectedAttr  := _ProtectedAttr;
  SelProtAttr    := _SelProtAttr;
  Asterisk       := False;
  Bar            := -1;
End; { DSelection.Init }

Destructor DSelection.Done;
Begin
  Inherited Done;
End; { DSelection.Done }

Procedure DSelection.Show(Selected: Boolean; Top: Integer);
Var St: ScreenStr;
Begin
  St := SelStr;
  If Checked Then
  Begin
    If St = '' Then St := ' ';
    St[1] := '';
  End;
  If Asterisk Then
  Begin
    If St = '' Then St := '  ';
    St[2] := '*';
  End;
  If Bar >= 0 Then
  Begin
    If St = '' Then St := '  ';
    Case (Bar And 3) Of
      0: St[2] := '-';
      1: St[2] := '/';
      2: St[2] := '';
      3: St[2] := '\';
    End; { Case }
  End;
  While Length(St) < Width Do St := St + ' ';

  { Display the entry }

  If (Y - Top > 0) And (Y - Top < Hi(Screen.WindMax) - Hi(Screen.WindMin) + 2) Then
  Begin
    If Not Protected Then
    Begin
      If Not Selected Then
      Begin
        WriteAtAttr(X,Y - Top,WindowAttr,St);
        If HotChar > 0 Then
         WriteAtAttr(X + HotChar - 1,Y - Top,HotCharAttr,Copy(St,HotChar,1));
      End
      Else
      Begin
        WriteAtAttr(X,Y - Top,SelAttr,St);
        If HotChar > 0 Then
         WriteAtAttr(X + HotChar - 1,Y - Top,SelHotCharAttr,Copy(St,HotChar,1));
      End;
    End
    Else
    Begin
      If Not Selected Then
      Begin
        WriteAtAttr(X,Y - Top,ProtectedAttr,St);
        If HotChar > 0 Then
         WriteAtAttr(X + HotChar - 1,Y - Top,HotCharAttr,Copy(St,HotChar,1));
      End
      Else
      Begin
        WriteAtAttr(X,Y - Top,SelProtAttr,St);
        If HotChar > 0 Then
         WriteAtAttr(X + HotChar - 1,Y - Top,SelHotCharAttr,Copy(St,HotChar,1));
      End
    End;
  End;
End; { DSelection.Show }

{------------------------------- }
{ DWindow                        }
{------------------------------- }

Constructor DWindow.Init(_X1,_Y1,_X2,_Y2: Integer;
                        WFore,WBack,BFore,BBack,CFore,CBack: Byte;
                        BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
Var Swap: Integer;
Begin
  Inherited Init;
  If _X2 < _X1 Then
  Begin
    Swap := _X2;
    _X2  := _X1;
    _X1  := Swap;
  End;
  If _Y2 < _Y1 Then
  Begin
    Swap := _Y2;
    _Y2  := _Y1;
    _Y1  := Swap;
  End;
  OX1         := _X1;
  OY1         := _Y1;
  OX2         := _X2;
  OY2         := _Y2;
  X1          := _X1;
  Y1          := _Y1;
  X2          := _X2;
  Y2          := _Y2;
  If X1 < 1         Then X1 := 1;
  If Y1 < 1         Then Y1 := 1;
  If X2 >= MaxX - 1 Then X2 := MaxX - 1;
  If Y2 >= MaxY - 1 Then Y2 := MaxY - 1;
  WindowAttr  := (WFore And $F) + (WBack Shl 4);
  BorderAttr  := (BFore And $F) + (BBack Shl 4);
  CaptionAttr := (CFore And $F) + (CBack Shl 4);
  CaptionTop  := CapTop;
  CaptionBot  := CapBot;
  Border      := BType;
  Shadow      := Shad;
  WindowPtr   := Nil;
  Visible     := False;
  If Shadow
   Then Size := (Y2 - Y1 + 2) * (X2 - X1 + 2) * 2
   Else Size := (Y2 - Y1 + 1) * (X2 - X1 + 1) * 2;
End; { DWindow.Init }


Constructor DWindow.InitWH(Width,Height: Integer;
                           WFore,WBack,BFore,BBack,CFore,CBack: Byte;
                           BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
Begin
  Inherited Init;
  X1 := (MaxX - Width) Div 2;
  Y1 := (MaxY - Height) Div 2;
  X2 := ((MaxX + Width) Div 2) - 1;
  Y2 := ((MaxY + Height) Div 2) - 1;
  OX1         := X1;
  OY1         := Y1;
  OX2         := X2;
  OY2         := Y2;
  If X1 < 1         Then X1 := 1;
  If Y1 < 1         Then Y1 := 1;
  If X2 >= MaxX - 1 Then X2 := MaxX - 1;
  If Y2 >= MaxY - 1 Then Y2 := MaxY - 1;
  WindowAttr  := (WFore And $F) + (WBack Shl 4);
  BorderAttr  := (BFore And $F) + (BBack Shl 4);
  CaptionAttr := (CFore And $F) + (CBack Shl 4);
  CaptionTop  := CapTop;
  CaptionBot  := CapBot;
  Border      := BType;
  Shadow      := Shad;
  WindowPtr   := Nil;
  Visible     := False;
  If Shadow
   Then Size := (Y2 - Y1 + 2) * (X2 - X1 + 2) * 2
   Else Size := (Y2 - Y1 + 1) * (X2 - X1 + 1) * 2;
End; { DWindow.InitWH }


Procedure DWindow.WriteFrame(Var R: Rect);
Var I: Integer;
Begin
  If R.Active Then
  Begin
    WriteAt(R.X1,R.Y1,R.AFore,R.ABack,R.AUL);
    WriteAt(R.X2,R.Y1,R.AFore,R.ABack,R.AUR);
    WriteAt(R.X1,R.Y2,R.AFore,R.ABack,R.ABL);
    WriteAt(R.X2,R.Y2,R.AFore,R.ABack,R.ABR);
    For I := R.X1 + 1 To R.X2 - 1 Do
    Begin
      WriteAt(I,R.Y1,R.AFore,R.ABack,R.AHZ);
      WriteAt(I,R.Y2,R.AFore,R.ABack,R.AHZ);
    End; { For I }
    For I := R.Y1 + 1 To R.Y2 - 1 Do
    Begin
      WriteAt(R.X1,I,R.AFore,R.ABack,R.AVT);
      WriteAt(R.X2,I,R.AFore,R.ABack,R.AVT);
    End; { For I }
  End
  Else
  Begin
    WriteAt(R.X1,R.Y1,R.PFore,R.PBack,R.PUL);
    WriteAt(R.X2,R.Y1,R.PFore,R.PBack,R.PUR);
    WriteAt(R.X1,R.Y2,R.PFore,R.PBack,R.PBL);
    WriteAt(R.X2,R.Y2,R.PFore,R.PBack,R.PBR);
    For I := R.X1 + 1 To R.X2 - 1 Do
    Begin
      WriteAt(I,R.Y1,R.PFore,R.PBack,R.PHZ);
      WriteAt(I,R.Y2,R.PFore,R.PBack,R.PHZ);
    End; { For I }
    For I := R.Y1 + 1 To R.Y2 - 1 Do
    Begin
      WriteAt(R.X1,I,R.PFore,R.PBack,R.PVT);
      WriteAt(R.X2,I,R.PFore,R.PBack,R.PVT);
    End; { For I }
  End;
End; { DWindow.WriteFrame }


Procedure DWindow.WriteArrows(Up,Down: Boolean);
Var
  UpSt,DnSt : String[1];
  WMin,WMax : Word;

Begin
  WMin := Screen.WindMin;
  WMax := Screen.WindMax;
  Screen.Window(1,1,MaxX,MaxY);
  If Border = DoubleBorder Then
  Begin
    If Up   Then UpSt := #24 Else UpSt := WindowRect.AVT;
    If Down Then DnSt := #25 Else DnSt := WindowRect.AVT;
  End
  Else
  Begin
    If Up   Then UpSt := #24 Else UpSt := WindowRect.PVT;
    If Down Then DnSt := #25 Else DnSt := WindowRect.PVT;
  End;
  If Up
   Then WriteAt(X2,Y1 + 1,BorderAttr Shr 4 ,BorderAttr And $F,UpSt)
   Else WriteAt(X2,Y1 + 1,BorderAttr And $F,BorderAttr Shr 4 ,UpSt);
  If Down
   Then WriteAt(X2,Y2 - 1,BorderAttr Shr 4 ,BorderAttr And $F,DnSt)
   Else WriteAt(X2,Y2 - 1,BorderAttr And $F,BorderAttr Shr 4 ,DnSt);
  Screen.Window(Lo(WMin) + 1,Hi(WMin) + 1,Lo(WMax) + 1,Hi(WMax) + 1);
End; { DWindow.WriteArrows }


Procedure DWindow.ShadeAt(X,Y: Integer; Len: Byte);
Begin
  Inc(X,Lo(Screen.WindMin));
  Inc(Y,Hi(Screen.WindMin));
  Screen.FastShade(X,Y,Len);
End; { DWindow.ShadeAt }


Procedure DWindow.Display;
Var
  I : Integer;
  P : Pointer;

Begin
  If Visible Then Exit;
  P := Screen.ScreenPtr;
  GetMem(WindowPtr,Size);
  If Shadow Then
  Begin
    For I := Y1 To Y2 + 1 Do
    Begin
      Move(Mem[Seg(P^):Ofs(P^) + ((I - 1) * MaxX + X1 - 1) * 2],
           Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 2) * 2],
           (X2 - X1 + 2) * 2);
    End; { For I }
  End
  Else
  Begin
    For I := Y1 To Y2 Do
    Begin
      Move(Mem[Seg(P^):Ofs(P^) + ((I - 1) * MaxX + X1 - 1) * 2],
           Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 1) * 2],
           (X2 - X1 + 1) * 2);
    End; { For I }
  End;
  LastWindMin := Screen.WindMin;
  LastWindMax := Screen.WindMax;
  Screen.Window(1,1,MaxX,MaxY);
  If Shadow Then
  Begin
    For I := Y1 + 1 To Y2 Do ShadeAt(X2 + 1,I,1);
    ShadeAt(X1 + 1,Y2 + 1,X2 - X1 + 1);
  End;
  Screen.SetTextAttr(BorderAttr);
  WindowRect.X1 := X1;
  WindowRect.Y1 := Y1;
  WindowRect.X2 := X2;
  WindowRect.Y2 := Y2;
  WindowRect.AFore := BorderAttr And $F;
  WindowRect.ABack := BorderAttr Shr 4;
  WindowRect.PFore := BorderAttr And $F;
  WindowRect.PBack := BorderAttr Shr 4;
  Case Border Of
    SingleBorder: WindowRect.Active := False;
    DoubleBorder: WindowRect.Active := True;
  End; { Case }
  WriteFrame(WindowRect);
  Screen.Window(X1,Y1,X2,Y2);
  If CaptionTop <> '' Then
  Begin
    Screen.SetTextAttr(CaptionAttr);
    WriteAtAttr(((X2 - X1 + 1) Div 2) + 1 - ((Length(CaptionTop) Div 2)),1,
     CaptionAttr,CaptionTop);
  End;
  If CaptionBot <> '' Then
  Begin
    Screen.SetTextAttr(CaptionAttr);
    WriteAtAttr(((X2 - X1 + 1) Div 2) + 1 - ((Length(CaptionBot) Div 2)),(Y2 - Y1) + 1,
     CaptionAttr,CaptionBot);
  End;
  Screen.SetTextAttr(WindowAttr);
  Screen.Window(X1 + 1,Y1 + 1,X2 - 1,Y2 - 1);
  Screen.ClrScr;
  Visible := True;
End; { DWindow.Display }


Procedure DWindow.Remove;
Var I: Integer;
Begin
  If (WindowPtr = Nil) Or Not Visible Then Exit;
  Screen.Window(Lo(LastWindMin) + 1,Hi(LastWindMin) + 1,
                Lo(LastWindMax) + 1,Hi(LastWindMax) + 1);
  If Shadow Then
  Begin
    For I := Y1 To Y2 + 1 Do
    Begin
      Screen.FastMove(Ptr(Seg(WindowPtr^),Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 2) * 2),
       ((I - 1) * MaxX + X1 - 1) * 2,(X2 - X1 + 2) * 2);
{
      Move(Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 2) * 2],
           Mem[$B800:((I - 1) * MaxX + X1 - 1) * 2],
           (X2 - X1 + 2) * 2);
}
    End; { For I }
  End
  Else
  Begin
    For I := Y1 To Y2 Do
    Begin
      Screen.FastMove(Ptr(Seg(WindowPtr^),Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 1) * 2),
           ((I - 1) * MaxX + X1 - 1) * 2,(X2 - X1 + 1) * 2);
{
      Move(Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 1) * 2],
           Mem[$B800:((I - 1) * MaxX + X1 - 1) * 2],
           (X2 - X1 + 1) * 2);
}
    End; { For I }
  End;
  FreeMem(WindowPtr,Size);
  Visible := False;
  WindowPtr := Nil;
End; { DWindow.Remove }


Destructor DWindow.Done;
Begin
  If Visible Then Remove;
  Inherited Done;
End; { DWindow.Done }

Procedure DWindow.MoveWindow(DeltaX,DeltaY: Integer);
Begin
  Inc(OX1,DeltaX);
  Inc(OY1,DeltaY);
  Inc(OX2,DeltaX);
  Inc(OY2,DeltaY);
  X1 := OX1;
  Y1 := OY1;
  X2 := OX2;
  Y2 := OY2;
  If X1 < 1         Then X1 := 1;
  If Y1 < 1         Then Y1 := 1;
  If X2 >= MaxX - 1 Then X2 := MaxX - 1;
  If Y2 >= MaxY - 1 Then Y2 := MaxY - 1;
  If Visible Then Screen.Window(X1 + 1,Y1 + 1,X2 - 1,Y2 - 1);
End; { DWindow.MoveWindow }

{------------------------------- }
{ DMenu                          }
{------------------------------- }

Constructor DMenu.Init(_X1,_Y1,_X2,_Y2: Integer;
                       WFore,WBack,BFore,BBack,CFore,CBack,HCFore,HCBack,
                       SFore,SBack,SHCFore,SHCBack,PFore,PBack,SPFore,SPBack: Byte;
                       BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
Begin
  Inherited Init(_X1,_Y1,_X2,_Y2,WFore,WBack,BFore,BBack,CFore,CBack,BType,CapTop,CapBot,Shad);
  HotCharAttr    := (HCFore  And $F) + (HCBack  Shl 4);
  SelAttr        := (SFore   And $F) + (SBack   Shl 4);
  SelHotCharAttr := (SHCFore And $F) + (SHCBack Shl 4);
  ProtectedAttr  := (PFore   And $F) + (PBack   Shl 4);
  SelProtAttr    := (SPFore  And $F) + (SPBack  Shl 4);
  CurSel         := 0;
  OldSel         := 0;
  SelList        := New(PCollection,Init(10,10));
  Callback       := Nil;
  Top            := 0;
  MaxSelY        := 0;
End; { DMenu.Init }


Constructor DMenu.InitWH(Width,Height: Integer;
                         WFore,WBack,BFore,BBack,CFore,CBack,HCFore,HCBack,
                         SFore,SBack,SHCFore,SHCBack,PFore,PBack,SPFore,SPBack: Byte;
                         BType: BorderType; CapTop,CapBot: ScreenStr; Shad: Boolean);
Var _X1,_Y1,_X2,_Y2: Integer;
Begin
  _X1 := (MaxX - Width) Div 2;
  _Y1 := (MaxY - Height) Div 2;
  _X2 := ((MaxX + Width) Div 2) - 1;
  _Y2 := ((MaxY + Height) Div 2) - 1;
  Inherited Init(_X1,_Y1,_X2,_Y2,WFore,WBack,BFore,BBack,CFore,CBack,BType,CapTop,CapBot,Shad);
  HotCharAttr    := (HCFore  And $F) + (HCBack  Shl 4);
  SelAttr        := (SFore   And $F) + (SBack   Shl 4);
  SelHotCharAttr := (SHCFore And $F) + (SHCBack Shl 4);
  ProtectedAttr  := (PFore   And $F) + (PBack   Shl 4);
  SelProtAttr    := (SPFore  And $F) + (SPBack  Shl 4);
  CurSel         := 0;
  OldSel         := 0;
  SelList        := New(PCollection,Init(10,10));
  Callback       := Nil;
  Top            := 0;
  MaxSelY        := 0;
End; { DMenu.InitWH }


Procedure DMenu.AddSelection(_X,_Y,SelWidth: Integer; HotCh: Byte;
                             Protect,Checked: Boolean; St: ScreenStr);
Begin
  SelList^.Insert(New(PSelection,Init(_X,_Y,SelWidth,St,HotCh,Protect,Checked,
                  WindowAttr,HotCharAttr,SelAttr,SelHotCharAttr,ProtectedAttr,
                  SelProtAttr)));
  If _Y > MaxSelY Then MaxSelY := _Y;
  CurSel := 0;
End; { DMenu.AddSelection }


Function DMenu.GetSelection(Sel: Integer): PSelection;
Begin
  If (Sel >= 1) And (Sel <= SelList^.Count)
   Then GetSelection := SelList^.At(Sel - 1)
   Else GetSelection := Nil;
End; { DMenu.GetSelection }


Procedure DMenu.SetSelection(Sel: Integer);
Var SelWork: PSelection;
Begin
  If (Sel <> OldSel) And (SelList^.Count > 0) Then
  Begin
    CurSel  := Sel;
    If (OldSel > 0) And (OldSel <= SelList^.Count) Then
    Begin
      SelWork := SelList^.At(OldSel - 1);
      SelWork^.Show(False,Top);
    End;
    OldSel := CurSel;
    If (OldSel > 0) And (OldSel <= SelList^.Count) Then
    Begin
      SelWork := SelList^.At(OldSel - 1);
      SelWork^.Show(True,Top);
    End;
  End;
End; { DMenu.SetSelection }


Procedure DMenu.Display;
Var
  SelWork : PSelection;
  I       : Integer;

Begin
  Inherited Display;
  I := 1;
  While I <= SelList^.Count Do
  Begin
    SelWork := SelList^.At(I - 1);
    SelWork^.Show(I = CurSel,Top);
    Inc(I);
  End; { While }
  If CurSel <= 0 Then CurSel := 1;
  SetSelection(CurSel);
  WriteArrows(Top > 0,Top < MaxSelY - (Y2 - Y1 - 1));
End; { DMenu.Display }


Procedure DMenu.DisplaySelection(Sel: Integer);
Var SelWork: PSelection;
Begin
  If (Sel >= 1) And (Sel <= SelList^.Count) Then
  Begin
    SelWork := SelList^.At(Sel - 1);
    SelWork^.Show(Sel = CurSel,Top);
  End;
End; { DMenu.DisplaySelection }


Destructor DMenu.Done;
Begin
  Dispose(SelList,Done);
  CurSel := 0;
  Inherited Done;
End; { DMenu.Done }


Procedure DMenu.CheckTop;
Var Sel: PSelection;
Begin
  Sel := GetSelection(CurSel);
  If Sel^.Y - Top > Y2 - Y1 - 1 Then
  Begin
    Top := Sel^.Y - (Y2 - Y1 - 1);
    Display;
  End;
  If Sel^.Y - Top < 1 Then
  Begin
    Top := Sel^.Y - 1;
    Display;
  End;
End; { DMenu.CheckTop }


Procedure DMenu.Execute;
Var
  Finished  : Boolean;
  Key       : Char;
  ScanCode  : Char;
  SelWork   : PSelection;
  Sel       : Integer;
  FoundChar : Boolean;
  Found     : Boolean;
  Y         : Integer;
  I         : Integer;

Begin
  If SelList^.Count < 1 Then Exit;
  If Not Visible Then Display;
  Finished := False;
  While Not Finished Do
  Begin
    If @Callback <> Nil Then Callback;
    If KeyPressed Then
    Begin
      Key := UpCase(ReadKey);
      Case Key Of
        Key_Esc:
        Begin
          SetSelection(0);
          Finished := True;
        End;
        Key_Enter:
        Begin
          SelWork := SelList^.At(CurSel - 1);
          If Not SelWork^.Protected Then Finished := True;
        End;
        Key_Null:
        Begin
          ScanCode := ReadKey;
          Case ScanCode Of
            Key_Up: Begin
                      Sel := CurSel - 1;
                      While (Sel <> CurSel) And GetSelection(Sel)^.Protected Do
                      Begin
                        Dec(Sel);
                        If Sel < 1 Then Sel := SelList^.Count;
                      End; { While }
                      SetSelection(Sel);
                      CheckTop;
                    End;
          Key_Down: Begin
                      Sel := CurSel + 1;
                      While (Sel <> CurSel) And GetSelection(Sel)^.Protected Do
                      Begin
                        Inc(Sel);
                        If Sel > SelList^.Count Then Sel := 1;
                      End; { While }
                      SetSelection(Sel);
                      CheckTop;
                    End;
          Key_Left: Begin
                      Sel   := CurSel;
                      Y     := GetSelection(CurSel)^.Y;
                      Found := False;
                      Repeat
                        Dec(Sel);
                        If Sel > 0 Then
                        Begin
                          SelWork := GetSelection(Sel);
                          If (SelWork^.Y = Y) And Not SelWork^.Protected Then Found := True;
                        End;
                      Until (Sel < 1) Or Found;
                      If Found Then
                      Begin
                        SetSelection(Sel);
                        CheckTop;
                      End;
                    End;
         Key_Right: Begin
                      Sel   := CurSel;
                      Y     := GetSelection(CurSel)^.Y;
                      Found := False;
                      Repeat
                        Inc(Sel);
                        If Sel <= SelList^.Count Then
                        Begin
                          SelWork := GetSelection(Sel);
                          If (SelWork^.Y = Y) And Not SelWork^.Protected Then Found := True;
                        End;
                      Until (Sel > SelList^.Count) Or Found;
                      If Found Then
                      Begin
                        SetSelection(Sel);
                        CheckTop;
                      End;
                    End;
          Key_Home: Begin
                      Sel := 1;
                      While (Sel < SelList^.Count) And GetSelection(Sel)^.Protected Do Inc(Sel);
                      SetSelection(Sel);
                      CheckTop;
                    End;
           Key_End: Begin
                      Sel := SelList^.Count;
                      While (Sel > 1) And GetSelection(Sel)^.Protected Do Dec(Sel);
                      SetSelection(Sel);
                      CheckTop;
                    End;
          End; { Case }
        End;
        Else
        Begin
          FoundChar := False;
          I         := CurSel;
          Repeat
            SelWork := SelList^.At(I - 1);
            If (SelWork^.HotChar > 0) And
               (Key = UpCase(SelWork^.SelStr[SelWork^.HotChar])) And
               Not SelWork^.Protected Then FoundChar := True
            Else
            Begin
              Inc(I);
              If I > SelList^.Count Then I := 1;
            End;
          Until (I = CurSel) Or FoundChar;
          If FoundChar Then
          Begin
            SetSelection(I);
            Finished := True;
          End;
        End;
      End; { Case }
    End;
  End; { While }
  Remove;
End; { DMenu.Execute }


Procedure DMenu.Remove;
Begin
  OldSel := 0;
  Inherited Remove;
End; { DMenu.Remove }

{$F+}
Procedure NewExitProc;
Begin
  ExitProc := OldExitProc;
  Screen.Done;
End; { NewExitProc }
{$F-}

{------------------------------- }
{ Unit setup                     }
{------------------------------- }

Begin
  OldExitProc := ExitProc;
  ExitProc    := @NewExitProc;
  Screen.Init;
End.
